/* dopplerlidar.h - functions for reading doppler lidar data */

#include <stdio.h>
#include <string.h>

#define MAX_GATES 400
/* Maximum raw gates depends on output format */
/*   Early versions - same number of gates as .hpl data */
/*   Later version  - sampled at number of points per gate * number of gates */
/*                    typical oversampling is 6-10 points per gate, adjust if more! */
/*   Note that there are currently 3 lags per sample, again adjust if this changes.. */
#define MAX_RAW_GATES 4000
/* Maximum number of beams expected for calculating winds from PPI scans */
#define MAX_BEAMS 24
#define PLANCK    6.6260693e-34
#define DEG2RAD   0.01745329251994
#define C_LIGHT   299792458 
#define PI_OVER_180 0.01745329
#define PI        3.1415926535897
#define HNU       1.28e-19
#define NCFILEWRITE 0
#define NCFILEREAD  1
#define NCFILECO    1    
#define NCFILECROSS 2

extern float missing_value;

/*** STRUCTURES ***/
typedef struct {
  unsigned short int hdr_lines;
  float gate_length;
  unsigned short int number_measurements;
  unsigned short int system_id;
  unsigned int number_gates_m1;
  unsigned int number_gates_m2;
  unsigned int number_pulses_m1;
  unsigned int number_pulses_m2;
  unsigned short int number_lags;
  unsigned short int gate_length_pts;
  unsigned int number_raw_samples_gate;
  unsigned int focus_range_m1;
  unsigned int focus_range_m2;
  unsigned short int num_integrations;
  unsigned short int num_rhis_in_scan;
  unsigned short int num_rays;
  unsigned short int scantype;
  unsigned short int scantypereq;
  signed short int year, month, day;
  signed short int hour, minute, second;
  float v_resolution;
  char filetype;
  char is_pitch_and_roll;
  unsigned int iarg;
} DopplerLidarHeader;

typedef struct {
  float signal[MAX_GATES];
  float beta_raw[MAX_GATES];
  float velocity_raw[MAX_GATES];
  float beta[MAX_GATES];
  float velocity[MAX_GATES];
  float beta_fractional_error[MAX_GATES];
  float velocity_error[MAX_GATES];
  float azimuth[MAX_GATES];
  float elevation[MAX_GATES];
  float pitch;
  float roll;
  float mean_noise;
  float std_noise;
  float time;
} DopplerLidarData;

typedef struct {
  double acf_raw[6*MAX_RAW_GATES];
} DopplerLidarRawData;

typedef struct {
  float acf_real[3*MAX_RAW_GATES];
  float acf_imag[3*MAX_RAW_GATES];
  float signal[MAX_RAW_GATES];
  float v_pp[MAX_RAW_GATES];
  float sigma_v_pp[MAX_RAW_GATES];
  float skew_v_pp[MAX_RAW_GATES];
  float kurt_v_pp[MAX_RAW_GATES];
  float mean_noise;
  float std_noise;
  unsigned short int numrays[MAX_RAW_GATES];
} DopplerLidarACFData;

typedef struct {
  float mean_beta[MAX_GATES];
  float mean_v[MAX_GATES];
  float sigma_v[MAX_GATES];
  float skew_v[MAX_GATES];
  float kurt_v[MAX_GATES];
  float numrays[MAX_GATES];
} DopplerLidarAvData;

typedef struct {
  float range[MAX_GATES];
  float height[MAX_GATES];
  float uwind[MAX_GATES];
  float vwind[MAX_GATES];
  float wwind[MAX_GATES];
  float speed[MAX_GATES];
  float direction[MAX_GATES];
  float uwind_error[MAX_GATES];
  float vwind_error[MAX_GATES];
  float wwind_error[MAX_GATES];
  float speed_error[MAX_GATES];
  float direction_error[MAX_GATES];
  float mean_snr[MAX_GATES];
  float noise_threshold;
  float azimuthal_correction;
  char  applyazimuthalcorrection; 
} DopplerLidarWindData;

typedef struct {
  float beta[MAX_GATES];
  float beta_cr[MAX_GATES];
  float velocity[MAX_GATES];
  float velocity_cr[MAX_GATES];
  float depolarization[MAX_GATES];
  float differential_v[MAX_GATES];
  float depolarization_error[MAX_GATES];
  float differential_v_error[MAX_GATES];
  float time;
} DopplerLidarDepolData;


typedef struct {  
  signed short int day, month, year;
  signed short int hour, minute, second;
} DopplerLidarTime;

typedef struct {
  float wavelength;
  float beam_energy;
  float lens_diameter;
  float divergence;
  float prf;
  float pulse_duration;
  float bandwidth;
  float nyquist;
  float elevation;
  float noise_threshold;
  float cleanv_threshold;
  float numseconds;
  float calfactor;
  float focus;
  float nominal_noise_floor;/*raw*/
  float nominal_noise_std;/*raw*/
  float acf_real_offset_applied;/*raw*/
  float acf_imag_offset_applied;/*raw*/
  float v_offset_applied; /*raw*/
  float missing_value;
  unsigned short int numrays;
  char  applycalfactor;
  char  applyfocuscorrection;
  char  cleanrawrayflag;
  char  calculate_epsilon;
} DopplerLidarSpec;

typedef struct {  
  float beta[2];
  float velocity[2]; 
  float beta_error[2];
  float velocity_error[2];
  float depolarization[2];
  float differential_v[2];
  float depolarization_error[2];
  float differential_v_error[2];
  float sigma_v[2];
  float skew_v[2];
  float kurt_v[2];
  float num[2];
  float uwind[2];
  float vwind[2];
  float wwind[2];
  float speed[2];
  float dir[2];
  float uwind_error[2];
  float vwind_error[2];
  float wwind_error[2];
  float speed_error[2];
  float dir_error[2];
} DopplerLidarPlotRange;


/*** PROTOTYPES ***/

/* File operations */
int doppler_lidar_read_hdr(FILE *fp, DopplerLidarHeader *header);
int doppler_lidar_read_profile(FILE *fp, DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarTime *current_time);

int doppler_lidar_open_hpl(DopplerLidarHeader *header, int argc, char **argv, char *hpldir);
int doppler_lidar_read_raw_profile(FILE *fp, DopplerLidarRawData *rawprofile, DopplerLidarACFData *acfprofile, DopplerLidarHeader *header, int argc, char **argv, char *hpldir);


/* Data manipulation operations */
int doppler_lidar_simple_clean_ray(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value);
int doppler_lidar_clean_ray(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value);
int doppler_lidar_clean_velocity(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value);
int doppler_lidar_correct_focus(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value);
int doppler_lidar_uncertainty(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value);

void doppler_lidar_init_average_profile(DopplerLidarAvData *avprofile, int ngates);
void doppler_lidar_average_profile(DopplerLidarData *profile, DopplerLidarAvData *avprofile, int ngates);
void doppler_lidar_calculate_moments(DopplerLidarAvData *avprofile, int ngates, int average_nrays, float missing_value);
void doppler_lidar_wind_profile(DopplerLidarData *DBSprofile, DopplerLidarWindData *windprofile, int ngates, float convention, float missing_value);
void doppler_lidar_vad_profile_TL(DopplerLidarData *VADprofile, DopplerLidarWindData *windprofile, int ngates, int nrays_per_scan, float convention, float missing_value);
void doppler_lidar_vad_profile_EP(DopplerLidarData *VADprofile, DopplerLidarWindData *windprofile, int ngates, int nrays_per_scan, float convention, float missing_value);
void doppler_lidar_depol_profile(DopplerLidarData *rawprofile, DopplerLidarDepolData *depolprofile, int ngates, float noise_threshold, float missing_value);

int doppler_lidar_simple_clean_raw_ray(DopplerLidarACFData *acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec);
int doppler_lidar_calculate_pulsepair_v(DopplerLidarACFData *acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec);
int doppler_lidar_single_raw_profile(FILE *fp, DopplerLidarRawData *rawprofile, DopplerLidarACFData *acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec, int argc, char **argv, char *hpldir);
int doppler_lidar_average_raw_profile(FILE *fp, DopplerLidarRawData *rawprofile, DopplerLidarACFData *acfprofile, DopplerLidarACFData *ave_acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec, int argc, char **argv, char *hpldir);
int doppler_lidar_average_raw_profile_grid(FILE *fp, DopplerLidarRawData *rawprofile, DopplerLidarACFData *acfprofile, DopplerLidarACFData *ave_acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec, int argc, char **argv, char *hpldir);

/* Time/Date operations */
void doppler_lidar_check_date(DopplerLidarHeader *header, DopplerLidarTime *current_time);

/* NetCDF File operations */
void doppler_lidar_write_global_attributes(int ncid, DopplerLidarTime *current_time, char *location, char *user, char *lscantype, char *institution, short int system_id);
void doppler_lidar_write_instrument_attributes(int ncid, int bs_dimids[3], int *numpulsesm1id, int *numpulsesm2id, int *numlagsid, int *numptsid, int *lensid, int *divid, int *focusid, int *energyid, int *waveid, int *prfid, int *pulselengthid, int *noiseid, int *bandid);
void doppler_lidar_write_location_attributes(int ncid, int bs_dimids[3], int *latid, int *lonid, int *altid);
void doppler_lidar_write_coordvar_attributes(int ncid, int bs_dimids[3], int *time_dim, int *timeid, int *range_dim, int *rangeid, int ngates, int *lag_dim, int *lagid, int nlags, char outputraw);
void doppler_lidar_write_height_attributes(int ncid, int *range_dim, int *heightid);
void doppler_lidar_write_az_el_attributes(int ncid, int bs_dimids[3], int elazid[4], unsigned short int scantypereq, unsigned short int scantype, char is_pitch_and_roll);
void doppler_lidar_write_standardvar_attributes(DopplerLidarHeader *header, DopplerLidarSpec *spec, int ncid, int bs_dimids[3], int *mnid, int *bsid, int *vid, int *rangeid, int *bsrawid, int *vrawid, int *sigid, float missing_value, DopplerLidarPlotRange *plot_range);
void doppler_lidar_write_errorvar_attributes(DopplerLidarHeader *header, DopplerLidarSpec *spec, int ncid, int bs_dimids[3], int *bserrid, int *verrid, float missing_value, DopplerLidarPlotRange *plot_range);
void doppler_lidar_write_rawvar_attributes(DopplerLidarHeader *header, DopplerLidarSpec *spec, int ncid, int bs_dimids[3], int *arid, int *aiid, int *siid, int *vid, float missing_value, char outputraw, DopplerLidarPlotRange *plot_range);
void doppler_lidar_write_averagedrawvar_attributes(int ncid, int bs_dimids[3], int *svid, int *nrid, float missing_value, DopplerLidarPlotRange *plot_range);
void doppler_lidar_write_averagedvar_attributes(int ncid, int bs_dimids[3], int *mbetaid, int *mvid, int *svid, int *nrid, float missing_value, DopplerLidarPlotRange *plot_range);
void doppler_lidar_write_highermomentvar_attributes(int ncid, int bs_dimids[3], int *skvid, int *kvid, float missing_value, DopplerLidarPlotRange *plot_range, char ispulsepair);
void doppler_lidar_write_windvar_attributes(int ncid, int bs_dimids[3], int *uwindid, int *vwindid, int *wwindid, int *speedid, int *dirid, float convention, float azimuthalcorrection, float missing_value, DopplerLidarPlotRange *plot_range);
void doppler_lidar_write_snr_attributes(int ncid, int bs_dimids[3], int *meansnrid, float missing_value);
void doppler_lidar_write_winderrorvar_attributes(int ncid, int bs_dimids[3], int *uerrid, int *verrid, int *werrid, int *speederrid, int *direrrid, float missing_value, DopplerLidarPlotRange *plot_range);
void doppler_lidar_write_depolvar_attributes(int ncid, int bs_dimids[3], int paramid[6], float missing_value, DopplerLidarPlotRange *plot_range);

void doppler_lidar_copy_global_attributes(int ncid[3], DopplerLidarTime *current_time, char *user, char *location);
void doppler_lidar_copy_scalar_varatts(int ncid[3]);
void doppler_lidar_copy_range_var(int ncid[3], int rangeid[3], int ngates);
void doppler_lidar_copy_range_var_wind(int ncid[2], DopplerLidarWindData *windprofile, int rangeid[2], int ngates);
void doppler_lidar_copy_profile(int ncid, DopplerLidarData *profile, int ngates, int index);

void doppler_lidar_populate_timestruct_nc(int ncid, int timeid, DopplerLidarTime *current_time);


