/* 
Convert Doppler Lidar ASCII data to netCDF 

History
-------
2006-09-25 eoc started
2007-10-03 eoc revision for variable file formats
2008-01-10 eoc revision for higher moment generation in separate files
2008-07-03 eoc add capability for depolarisation
2010-04-11 eoc calibration factor can be applied 
2010-09-15 eoc take vertical stare only from concatenated wind profiles
2011-11-10 eoc create horizontal wind vector and wind speed/direction files 
2014-04-03 eoc update doppler_read.c to read new file format
2018-03-08 eoc update doppler_read.c to read new file format


Author - Ewan O'Connor
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netcdf.h>
#include <ctype.h>
#include <math.h> 

#include "dopplerlidar.h"
#include "nctools.h"

#define MAX_STRING_LENGTH 512
#define MAX_LINE_LEN 2048


int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}


void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.txt  [file2.txt ...]\n"
	  "Options:\n"
	  "       -filedate d        only read data from this date\n"
	  " -noisethreshold d        use noise threshold to clean data\n"
	  "        -avetime d        average onto a regular d-second grid \n"
	  "        -averays d        average d rays (always integer)\n"
	  "         -cleanv f        clean velocity below beta threshold \n"
	  "        -noclean          don't remove speckle\n"
	  "      -elevation f        elevation of instrument (from horizontal)\n"
	  "       -altitude f        altitude (in m) of instrument above sea level\n"
	  "       -location s f f f  location of the lidar (e.g. Chilbolton), followed by\n"
	  "                             its latitude (N), longitude (E) and altitude (m)\n"
	  "           -site s        use known lat, lon and alt for site s - currently\n"
	  "                             understood sites are Chilbolton\n"
          "    -institution s        Instrument owner (short acronym). Will override global\n"
          "                             attributes for source and title with: [Institution] 1.5 micron\n"
          "                             Doppler lidar operating at [location]. Use, for instance,\n"
          "                             when instrument is on a campaign.\n" 
	  "         -prefix dir      write the files in specified directory\n"
	  "           -user s        name and email address of the user\n"
	  "    -addcrosspol          attempt to include depolarisation channel if available\n"
	  "    -crosspoldir dir      find crosspolar files in specified directory\n"
	  "      -calfactor f        multiply beta by calibration factor f\n"
          "     -applyfocus          apply focus correction to beta using value in header\n"
          "                             if focus not infinity (default)\n"
          " -dontapplyfocus          don't apply focus correction using value in header\n"
          "                             (use beta in file, which assumes focus is infinity)\n"
          "   -correctfocus f        recalculate beta from signal using focus f\n"
          "                             (overrides -[dont]applyfocus)\n"
          " -correctnyquist f        apply correct nyquist velocity (default is 19.56 m s-1)\n"
	  "       -vertical          select vertical data only from wind or stare profiles\n"
	  "        -verbose          write more\n",
	  argv[0]);
  exit(0);
}

int main(int argc, char **argv){
  FILE *fp = NULL;
  char *filename = '\0';
  char *prefix = ".";
  char *crosspoldir = "./cross/";
  DopplerLidarHeader header;
  DopplerLidarData profile;
  DopplerLidarSpec spec;
  float latitude = 51.1445;
  float longitude = 358.563;
  float altitude = 90;
  char *location = NULL;
  char *llocation = NULL;
  char *scantype = NULL;
  char *lscantype = NULL;
  char *user = NULL;
  char *institution = NULL;
  DopplerLidarPlotRange plot_range = {{1e-7,    1e-4},
                                      { -2.0,    2.0},
                                      {  0.001,  1.0},
                                      {  0.001, 10.0},
                                      {  0.0,    0.5},
                                      { -1.0,    1.0},
                                      {  0.001,  1.0},
                                      {  0.001, 10.0},
                                      {  0.001,  1.0},
                                      { -2.0,    2.0},
                                      { -2.0,    2.0},
                                      {  0.0,    1.0},
                                      {-15.0,   15.0},
                                      {-15.0,   15.0},
                                      { -2.0,    2.0},
                                      {  0.0,   15.0},
                                      {  0.0,  360.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    2.0},
                                      {  0.0,   10.0}};

  /* Need to keep previous time to check when midnight occurs */
  DopplerLidarTime time, last_time = {0, 0, 0, 0, 0, 0};
  DopplerLidarTime *current_time = &time; /* initialise current */
  DopplerLidarTime *prev_time = &last_time;    /* and next block */

  int iarg = 0; 
  char ncfilename[MAX_STRING_LENGTH];
  unsigned short int fday=0, fmonth=0, fyear=0;
  unsigned short int fhour=0, fmin=0, fsec=0;


  float time_increment = 0;

  int  ncid;                         /* netCDF ID */
  int  time_dim, range_dim, lag_dim; /* dimension IDs */
  int  timeid, rangeid, lagid;       /* variable ID's */
  int  latid = 0, lonid = 0, altid = 0;	
  int  bsid, bsrawid, vid, vrawid, sigid, bserrid, verrid;	
  int  numpulsesm1id, numpulsesm2id, numlagsid, numptsid;
  int  lensid, divid, focusid;
  int  energyid, waveid, prfid, pulselengthid, noiseid, mnid, bandid, azelid[4];
  int  bs_dimids[3];                 /* variable shape */
  float missing_value=-999;

  /* counters */
  int index = 0; /* Index of the current profile in the input file */

  /* flags */
  char newfile = 1;
  char filestatus = 1;
  char filecrossstatus = 1;
  char fileopen = 0;
  char ncfileopen = 0;
  char noclean = 0;
  char avetime = 0;
  char averays = 0;
  char cleanv = 0;
  char verbose = 0;
  char onlyverticalstare = 0;
  char addcrosspol = 0;
  char outputraw = 0;
  char ispulsepair = 0;

  /* Initialise array for spec */
  spec.wavelength = 0.0000015;
  spec.beam_energy = 0.00001;
  spec.lens_diameter = 0.060;
  spec.divergence = 0.000033;
  spec.prf = 15000;
  spec.pulse_duration = 0.0000002;
  spec.bandwidth = 5e7;
  spec.nyquist = 19.56;
  spec.elevation = 90;
  spec.noise_threshold = 1.0009;
  spec.cleanv_threshold = 0;
  spec.numseconds = 0;
  spec.numrays = 1;
  spec.calfactor = 1;
  spec.focus = -1;
  spec.applycalfactor = 0;
  spec.applyfocuscorrection = 0;

  /* Initialise array for header */
  header.hdr_lines = 12;
  /*  header.date_stamp = NULL;*/
  header.gate_length = 0;
  header.number_measurements = 0;
  header.system_id = 0;  
  header.number_gates_m1 = 0;
  header.number_gates_m2 = 0;
  header.number_pulses_m1 = 0;
  header.number_pulses_m2 = 0;
  header.number_lags = 0;
  header.gate_length_pts = 0;
  header.number_raw_samples_gate = 1;
  header.focus_range_m1 = 0;
  header.focus_range_m2 = 0;
  header.num_integrations = 0;
  header.num_rhis_in_scan = 0;
  header.num_rays = 0;
  header.scantype = 0;
  header.scantypereq = 0;
  header.year = 0;
  header.month = 0; 
  header.day = 0;
  header.hour = 0; 
  header.minute = 0;
  header.second = 0;
  header.v_resolution = 0;
  header.is_pitch_and_roll = 0;

  /* Process command-line arguments */
  if (argc < 2)
    {
      Usage(argv);
      exit(0);
    }

  while (++iarg < argc) {

    if (argv[iarg][0] != '-') {
      /* Found a lidar filename.*/
      filename = argv[iarg];
      break;
    }
    else if (strcmp(argv[iarg],"-filedate") == 0 && iarg+1<argc) {
      if (sscanf(argv[++iarg],"%04hd%02hd%02hd",
		 &fyear, &fmonth, &fday) < 3) {
	fprintf(stderr, "Failed to read date: should be in the form YYYYMMDD");
      }
    }

    else if (strcmp(argv[iarg],"-starttime") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%02hd%02hd%02hd",
		 &fhour, &fmin, &fsec) < 3) {
	fprintf(stderr, "Failed to read start time: should be in the form HHMMSS");
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",
		 &time_increment) < 1) {
	fprintf(stderr, "Failed to read time increment as float");
      }
    }
    else if (strcmp(argv[iarg],"-noclean") == 0 ) {
      noclean = 1;
    }
    else if (strcmp(argv[iarg],"-noisethreshold") == 0 && iarg+1<argc) {
      spec.noise_threshold = atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-calfactor") == 0 && iarg+1<argc) {
      spec.calfactor = atof(argv[++iarg]);
      if (spec.calfactor <= 0.0) {
	fprintf(stderr, "Multiplicative calibration factor cannot be negative or zero");
      }
      spec.applycalfactor = 1;
    }
    else if (strcmp(argv[iarg],"-applyfocus") == 0) {
      spec.applyfocuscorrection = 1;
    }
   /* Only apply if -correctfocus has not been supplied */
    else if (strcmp(argv[iarg],"-dontapplyfocus") == 0 && spec.focus < 0) {
      spec.applyfocuscorrection = 0;
    }
    else if (strcmp(argv[iarg],"-correctfocus") == 0 && iarg+1<argc) {
      spec.focus = atof(argv[++iarg]);
      spec.applyfocuscorrection = 1;
    }
    else if (strcmp(argv[iarg],"-correctnyquist") == 0 && iarg+1<argc) {
      spec.nyquist = atof(argv[++iarg]);
    }

    /*    else if (strcmp(argv[header.iarg], "-avetime") == 0) {
      if (argc > header.iarg+1) {
	spec.numseconds = atof(argv[++header.iarg]);
	avetime = 1;
      }
      else {
	fprintf(stderr,"Averaging time not specified");
	exit(1);
      }
    }
    else if (strcmp(argv[header.iarg], "-averays") == 0) {
      if (argc > header.iarg+1) {
	spec.numrays = (unsigned short) atoi(argv[++header.iarg]);
	plot_range_num[1] = (float) spec.numrays;
	if (spec.numrays > 1)
	  averays = 1;
      }
      else {
	fprintf(stderr,"Number of rays to average not specified");
	exit(1);
      }
      }*/
    else if (strcmp(argv[iarg], "-cleanv") == 0) {
      if (argc > iarg+1) {
	spec.cleanv_threshold = atof(argv[++iarg]);
	cleanv = 1;
      }
      else {
	fprintf(stderr,"Threshold for velocity not specified");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg],"-prefix") == 0) {
      if (argc > iarg+1) {
	prefix = argv[++iarg];
      }
      else {
	fprintf(stderr,"Prefix not specified");
	exit(1);
      }
    } 
    else if (strcmp(argv[iarg], "-verbose") == 0) {
      verbose = 1;
    }   
    else if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	user = argv[++iarg];
      }
      else {
	fprintf(stderr,"Error: user not specified\n");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg], "-elevation") == 0) {
      if (argc > iarg+1) {
	spec.elevation = atof(argv[++iarg]);
      }
      else {
	fprintf(stderr,"Elevation not specified");
	exit(1);
      }
    } 
    else if (strcmp(argv[iarg], "-altitude") == 0) {
      if (argc > iarg+1) {
	altitude = atof(argv[++iarg]);
      }
      else {
	fprintf(stderr,"Altitude not specified");
	exit(1);
      }
    }   
    else if (strcmp(argv[iarg],"-location") == 0) {
      if (argc > iarg+4) {
	location = argv[++iarg];
	latitude = atof(argv[++iarg]);
	longitude = atof(argv[++iarg]);
	altitude = atof(argv[++iarg]);
      }
      else {
	fprintf(stderr,"Not all location arguments were provided");
      }
    }
    else if (strcmp(argv[iarg], "-site") == 0) {
      if (argc > iarg+1) {
	++iarg;
	if (strcasecmp(argv[iarg], "Chilbolton") == 0) {
	  location = "Chilbolton";
	  latitude = 51.1445;
	  longitude = 358.563;
	  altitude = 90;

	}
	else if (strcasecmp(argv[iarg], "Salford") == 0) {
	  location = "Salford";
	  latitude = 0.0;
	  longitude = 0.0;
	  altitude = 0;

	}
	else {
	  fprintf(stderr,"Site unknown: I only know about Chilbolton and Salford");
	}
      }
    }

    else if (strcmp(argv[iarg],"-institution") == 0) {
      if (argc > iarg+1) {
        institution = argv[++iarg];
      }
      else {
        fprintf(stderr,"Error: user not specified\n");
        exit(1);
      }
    }

    else if (strcmp(argv[iarg], "-addcrosspol") == 0) {
      addcrosspol = 1;
    }   

    else if (strcmp(argv[iarg],"-crosspoldir") == 0) {
      if (argc > iarg+1) {
	crosspoldir = argv[++iarg];
      }
      else {
	fprintf(stderr,"Crosspolar directory not specified");
	exit(1);
      }
    } 
    else if (strcmp(argv[iarg], "-vertical") == 0) {
      header.scantypereq = 1;
      onlyverticalstare = 1;	
    }  

    else {
      fprintf(stderr, "Error: `%s' not understood or not accompanied "
	      "by enough subsequent arguments\n", argv[iarg]);
      Usage(argv);
      exit(1);
    }
  }

  /* cycle through filenames in argv list */
  while (iarg < argc ) 
    {
      size_t tmp_index;
      size_t start[2], count[2];
      float tmp_float;
      short tmp_short;
      signed int tmp_int;
      short j;
      DopplerLidarTime *tmpptr;

      if (!fileopen) 
	{
	  if ((fp = fopen(argv[iarg], "r")) == NULL) 
	    {
	      fprintf(stderr,"can't open input file\n");
	      exit(1);
	    }
	  if (verbose)
	    fprintf(stderr, "Reading file %s\n", argv[iarg]);
	  fileopen = 1;
	  
	  if (doppler_lidar_read_hdr(fp, &header)) 
	    {
	      fprintf(stderr,"Error reading header for %s\n", argv[iarg]);
	      exit(1);
	    }

	  if (header.scantype == 2) {
	    scantype = "FIX";
	    lscantype = "tilt";
	    spec.elevation = 86;
//	    isverticalstare = 1;
            header.scantypereq = 1; 	
	  }
	  else if (header.scantype == 3) {
	    scantype = "RHI";
	    lscantype = "rhi";
	  }
	  else if (header.scantype == 4) {
	    scantype = "VAD";
	    lscantype = "vad";
	  }
	  else if (header.scantype == 5) {
	    scantype = "WIND";
	    lscantype = "wind";
	  }
	  else if (header.scantype == 6) {
	    scantype = "USER";
	    lscantype = "user";
	  }
	  else {
	    scantype = "FIX";
	    lscantype = "fix";
	  }

	  /* For Stare scantype - is it vertical? */
          /* Allow user to request only vertical data from any scan type*/
          /* Following should no longer be needed! */
/*	  if ((header.scantype > 2 && header.scantypereq == 0) || (header.scantype == 1 && (profile.elevation[0] < 89.9 || profile.elevation[0] > 90.01)))
	    onlyverticalstare = 0;*/	
	}

      filestatus = doppler_lidar_read_profile(fp, &profile, &header, current_time);

      if (filestatus > 0)
	{
	  if (verbose)
	    fprintf(stderr, "Problem reading file, skipping rest of file.\n");
	  continue;
	}
       else if (filestatus < 0)
	{
	  fclose(fp);
	  fileopen = 0;
	  ++iarg;           
	}
      else
	{
      
	  if (!location) {
	    location = "Chilbolton";
	    llocation = "chilbolton";
	  }
	  else {
	    char *ch = strdup(location);
	    llocation = ch;
	    while (*ch) {
              if ( *ch == ' ' )
	        *ch = '-';
              else 
                *ch = tolower(*ch);
	      ++ch;
	    }
	  }
	  
	  doppler_lidar_check_date(&header, current_time);
	  if (spec.applyfocuscorrection)
	    doppler_lidar_correct_focus(&profile, &header, &spec, missing_value);
	  doppler_lidar_simple_clean_ray(&profile, &header, &spec, missing_value);

	  /* Calculate uncertainty estimates (both beta and velocity) */
	  doppler_lidar_uncertainty(&profile, &header, &spec, missing_value);
	  
	  /*      if (cleanv)
		  doppler_lidar_clean_ray(&profile, &header, &spec, missing_value);
		  doppler_lidar_clean_velocity(&profile, &header, &spec, missing_value);*/
	  
	  if (ncfileopen && (prev_time->year < current_time->year 
			     || prev_time->month < current_time->month 
			     || prev_time->day < current_time->day))
	    {
	      newfile = 1;
	      prev_time->year = current_time->year;
	      prev_time->month = current_time->month;
	      prev_time->day = current_time->day;
	      
	      /* Close netCDF file */
	      handle_error(nc_close(ncid));
	      index = 0;
	      ncfileopen = 0;
	    }
	  
	  if (newfile) {
	    
	    float tmp[header.number_gates_m1];

	    /* Open netCDF file for writing */		
	    if (header.scantype < 2 || header.scantypereq == 1) 
	      sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd_%s_halo-doppler-lidar.nc", prefix, current_time->year, current_time->month, current_time->day, llocation);	
	    else if (header.scantype == 2 || header.scantype == 5 )
	      sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd_%s_halo-doppler-lidar_%s.nc", prefix, current_time->year, current_time->month, current_time->day, llocation, lscantype);	   
	    else if (header.scantype == 4 || header.scantype == 6 )
	      sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd_%2.2hd%2.2hd%2.2hd_%s_halo-doppler-lidar_%s.nc", prefix, current_time->year, current_time->month, current_time->day, current_time->hour, current_time->minute, current_time->second, llocation, lscantype);	
	    else 	
	      sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd%2.2hd_%s_halo-doppler-lidar_%s.nc", prefix, current_time->year, current_time->month, current_time->day, current_time->hour, llocation, lscantype);

	    ncfilename[MAX_STRING_LENGTH-1] = '\0';	
	    
	    handle_error(nc_create(ncfilename, NC_CLOBBER, &ncid));
	    fprintf(stderr,"Writing %s\n",ncfilename);
	    fileopen = 1;
	    
	    /* Add some global attributes */
            tmp_int = (signed int) header.system_id;
	    doppler_lidar_write_global_attributes(ncid, current_time, location, user, scantype, institution, header.system_id);

	    /* Add a `command_line' attribute */
	    handle_error(nct_add_command_line(ncid, argc, argv));
	    
	    /* Add some scalar variables */
	    doppler_lidar_write_instrument_attributes(ncid, bs_dimids, &numpulsesm1id, &numpulsesm2id, &numlagsid, &numptsid, &lensid, &divid, &focusid, &energyid, &waveid, &prfid, &pulselengthid, &noiseid, &bandid);

	    if (spec.applyfocuscorrection)
	      nct_append_attribute(ncid, focusid, "comments", "This variable has been modified from the original value held in the raw .hpl files.",""); 

	    /* Add location variables */
	    doppler_lidar_write_location_attributes(ncid, bs_dimids, &latid, &lonid, &altid);
	    
	    /* Define dimensions and add coordinate variables */
	    doppler_lidar_write_coordvar_attributes(ncid, bs_dimids, &time_dim, &timeid, &range_dim, &rangeid, header.number_gates_m1, &lag_dim, &lagid, header.number_lags, outputraw);

	    /* Add azimuth and elevation (pitch and roll) variables */ 
//	    if (header.scantypereq == 0) 
//	      doppler_lidar_write_az_el_attributes(ncid, bs_dimids, &elid, &azid, &piid, &roid, header.scantypereq, header.scantype, header.is_pitch_and_roll);
//	    else
	      doppler_lidar_write_az_el_attributes(ncid, bs_dimids, azelid, header.scantypereq, header.scantype, header.is_pitch_and_roll);

	    /* Add standard variables */ 
	    doppler_lidar_write_standardvar_attributes(&header, &spec, ncid, bs_dimids, &mnid, &bsid, &vid, &rangeid, &bsrawid, &vrawid, &sigid, missing_value, &plot_range);

	    /* Add error variables */ 
	    doppler_lidar_write_errorvar_attributes(&header, &spec, ncid, bs_dimids, &bserrid, &verrid, missing_value, &plot_range);

	    /* End define mode */
	    handle_error(nc_enddef(ncid));

	    /* write scalar values into netCDF variables */
	    tmp_index = 0;

	    handle_error(nc_put_var1_float(ncid, latid, &tmp_index, &latitude));
	    handle_error(nc_put_var1_float(ncid, lonid, &tmp_index, &longitude));
	    handle_error(nc_put_var1_float(ncid, altid, &tmp_index, &altitude));
	    tmp_int = (signed int) header.number_pulses_m1;
	    handle_error(nc_put_var1_int(ncid, numpulsesm1id, &tmp_index, &tmp_int));
	    tmp_int = (signed int) header.number_pulses_m2;
	    handle_error(nc_put_var1_int(ncid, numpulsesm2id, &tmp_index, &tmp_int));
	    tmp_short = (short int) header.number_lags;
	    handle_error(nc_put_var1_short(ncid, numlagsid, &tmp_index, &tmp_short));
            tmp_short = (short int) header.gate_length_pts;
            handle_error(nc_put_var1_short(ncid, numptsid, &tmp_index, &tmp_short));
	    handle_error(nc_put_var1_float(ncid, lensid, &tmp_index, &(spec.lens_diameter)));
	    handle_error(nc_put_var1_float(ncid, divid, &tmp_index, &(spec.divergence)));
	    if (spec.applyfocuscorrection)
	      tmp_short = (short int) spec.focus;
	    else
	      tmp_short = (short int) header.focus_range_m1;

	    handle_error(nc_put_var1_short(ncid, focusid, &tmp_index, &tmp_short));
	    handle_error(nc_put_var1_float(ncid, energyid, &tmp_index, &(spec.beam_energy)));
	    handle_error(nc_put_var1_float(ncid, waveid, &tmp_index, &(spec.wavelength)));
	    handle_error(nc_put_var1_float(ncid, prfid, &tmp_index, &(spec.prf)));
	    handle_error(nc_put_var1_float(ncid, pulselengthid, &tmp_index, &(spec.pulse_duration)));
	    handle_error(nc_put_var1_float(ncid, bandid, &tmp_index, &(spec.bandwidth)));
	    handle_error(nc_put_var1_float(ncid, noiseid, &tmp_index, &(spec.noise_threshold)));

//	    if (header.scantype < 3) {	  
//	      if (isverticalstare) {
//		handle_error(nc_put_var1_float(ncid, elid, &tmp_index, &(spec.elevation)));       
//	      }
//	      else {
//		handle_error(nc_put_var1_float(ncid, elid, &tmp_index, &(profile.elevation[0])));  
//		handle_error(nc_put_var1_float(ncid, azid, &tmp_index, &(profile.azimuth[0])));  
//	      } 
//	    }
//	    else if (header.scantypereq == 1) {
//	      handle_error(nc_put_var1_float(ncid, elid, &tmp_index, &(profile.elevation[0])));  
//	    }

            if (header.scantypereq == 1)
              handle_error(nc_put_var1_float(ncid, azelid[0], &tmp_index, &(profile.elevation[0])));  
            
	    /* write vector values into netCDF variables */
	    for (j = 0; j < header.number_gates_m1 ; j++) 
	      {
		tmp[j] = (header.gate_length * (float) j * sin( spec.elevation / 180 * PI)) + (header.gate_length / 2);
	      }
	    handle_error(nc_put_var_float(ncid, rangeid, tmp));

	    newfile = 0;
	  }
	
	  /* Add a new profile of data */  
	  tmp_float = ((float) current_time->hour) + ((float) current_time->minute) / 60.0 + (((float) current_time->second)) / 3600.0;

	  tmp_index = index;
	  handle_error(nc_put_var1_float(ncid, timeid, &tmp_index, &tmp_float));
	  handle_error(nc_put_var1_float(ncid, mnid, &tmp_index, &(profile.mean_noise)));
      
	  start[0] = index;
	  start[1] = 0;
	  count[0] = 1;
	  count[1] = header.number_gates_m1;
      
	  handle_error(nc_put_vara_float(ncid, bsid, start,
					 count, profile.beta));
	  handle_error(nc_put_vara_float(ncid, vid, start,
					 count, profile.velocity));
	  handle_error(nc_put_vara_float(ncid, bsrawid, start,
					 count, profile.beta_raw));
	  handle_error(nc_put_vara_float(ncid, sigid, start,
					 count, profile.signal));
	  handle_error(nc_put_vara_float(ncid, vrawid, start,
					 count, profile.velocity_raw));
	  handle_error(nc_put_vara_float(ncid, bserrid, start,
					 count, profile.beta_fractional_error));
	  handle_error(nc_put_vara_float(ncid, verrid, start,
					 count, profile.velocity_error));

	  if (header.scantype == 5 || header.scantype == 6) {
	    handle_error(nc_put_var1_float(ncid, azelid[0], &tmp_index, &(profile.elevation[0])));
	    handle_error(nc_put_var1_float(ncid, azelid[1], &tmp_index, &(profile.azimuth[0])));
	  }
//	  else if (header.scantype > 2 && header.scantypereq != 1) {
          else if (header.scantypereq != 1) {

	    handle_error(nc_put_vara_float(ncid, azelid[0], start,
					   count, profile.elevation));
	    handle_error(nc_put_vara_float(ncid, azelid[1], start,
					   count, profile.azimuth));
	  }

          if (header.is_pitch_and_roll) {
            handle_error(nc_put_var1_float(ncid, azelid[2], &tmp_index, &(profile.pitch)));
            handle_error(nc_put_var1_float(ncid, azelid[3], &tmp_index, &(profile.roll)));
          }
	  
	  ++index;  
      
	  tmpptr=current_time;    /* current points to current block */
	  current_time=prev_time; /* current points to prev block */
	  prev_time=tmpptr;       /* prev points to current block which is now available for new block */
	  	
	} 
    }
  
  if (!newfile) 
    {
      /* Close netCDF file */
      handle_error(nc_close(ncid));
    }
  
  fclose(fp);
  exit(0); 
}
