/* 
Generate averages and higher moments from Doppler Lidar netCDF data

History
-------
2008-01-25 eoc started

Author - Ewan O'Connor
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netcdf.h>
#include <ctype.h>
#include <math.h> 

#include "dopplerlidar.h"
#include "nctools.h"

#define MAX_STRING_LENGTH 512
#define MAX_LINE_LEN 2048
//#define NCFILEREAD 0
//#define NCFILEWRITE 1

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s (%s)\n", \
	    __LINE__, nc_strerror(ncstatus), s); exit(-1); \
} 

static
char *
seek_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  if (nc_inq_atttype(ncid, varid, name, &type) != NC_NOERR
      || type != NC_CHAR) {
    return NULL;
  }

  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = malloc((length+1) * sizeof(char));
  if (!value) {
    fprintf(stderr, "Out of memory\n");
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}

static
short
get_short_attribute(int ncid, int varid, char *name)
{
  size_t length;
  short value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    fprintf(stderr, "Error: %s attribute is not of type short\n", name);
    exit(1);
  }
  if (length > 1) {
    fprintf(stderr, "Error: %s attribute has more than 1 element\n", name);
    exit(1);
  }
  handle_error(nc_get_att_short(ncid, varid, name, &value));
  return value;
}


void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.txt  [file2.txt ...]\n"
	  "Options:\n"
	  "        -avetime d        average onto a regular d-second grid \n"
	  "        -averays d        average d rays (always integer)\n"
	  "         -prefix dir      write the files in specified directory\n"
	  "           -user s        name and email address of the user\n"
	  "        -verbose          write more\n",
	  argv[0]);
  exit(0);
}

int main(int argc, char **argv){
  char *filename = '\0';
  char *prefix = ".";
  DopplerLidarData profile;
  DopplerLidarAvData avprofile;

  char *llocation = NULL;
  char *user = NULL;
  DopplerLidarPlotRange plot_range = {{1e-7,    1e-4},
                                      { -2.0,    2.0},
                                      {  0.001,  1.0},
                                      {  0.001, 10.0},
                                      {  0.0,    0.5},
                                      { -1.0,    1.0},
                                      {  0.001,  1.0},
                                      {  0.001, 10.0},
                                      {  0.001,  1.0},
                                      { -2.0,    2.0},
                                      { -2.0,    2.0},
                                      {  0.0,    1.0},
                                      {-15.0,   15.0},
                                      {-15.0,   15.0},
                                      { -2.0,    2.0},
                                      {  0.0,   15.0},
                                      {  0.0,  360.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    2.0},
                                      {  0.0,   10.0}};
  DopplerLidarTime time  = {0, 0, 0, 0, 0, 0};
  DopplerLidarTime *current_time = &time; /* initialise current */

  int   iarg = 0; 
  char  ncfilename[MAX_STRING_LENGTH];
  int   average_nrays = 1;
  float numseconds = 0;

  int   ncid[2];                        /* netCDF ID */
  int   time_dim[2], range_dim[2];      /* dimension IDs */
  int   timeid[2], rangeid[2], elid[2]; /* variable ID's */
  int   mbetaid, mvid, svid, nrid, skvid, kvid;
  int   bs_dimids[3];                       /* variable shape */
  size_t   nrays, ngates;
  int   lag_dim, lagid; /* Dummy IDs */
  int   nlags = 0; /* dummy var*/
  float missing_value=-999;

  /* flags */
  char  avetime = 0;
  char  averays = 0;
  char  verbose = 0;
  char  outputraw = 0;
  char  ispulsepair = 0;

  /* Process command-line arguments */
  if (argc < 2)
    {
      Usage(argv);
      exit(0);
    }

  while (++iarg < argc) {

    if (argv[iarg][0] != '-') {
      /* Found a lidar filename.*/
      filename = argv[iarg];
      break;
    }
    else if (strcmp(argv[iarg], "-avetime") == 0) {
      if (argc > iarg+1) {
	numseconds = atof(argv[++iarg]);
	avetime = 1;
      }
      else {
	fprintf(stderr,"Averaging time not specified");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg], "-averays") == 0) 
      {
	if (argc > iarg+1) 
	  {
	    average_nrays = atoi(argv[++iarg]);
	    plot_range.num[1] = (float) average_nrays;
	    if (average_nrays > 1)
	      averays = 1;
	  }
	else 
	  {
	    fprintf(stderr,"Number of rays to average not specified");
	    exit(1);
	  }
      }
    else if (strcmp(argv[iarg],"-prefix") == 0) {
      if (argc > iarg+1) {
	prefix = argv[++iarg];
      }
      else {
	fprintf(stderr,"Prefix not specified");
	exit(1);
      }
    } 
    else if (strcmp(argv[iarg], "-verbose") == 0) {
      verbose = 1;
    }   
    else if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	user = argv[++iarg];
      }
      else {
	fprintf(stderr,"Error: user not specified\n");
	exit(1);
      }
    }
    else {
      fprintf(stderr, "Error: `%s' not understood or not accompanied "
	      "by enough subsequent arguments\n", argv[iarg]);
      Usage(argv);
      exit(1);
    }
  }

  /* cycle through filenames in argv list */
  while (iarg < argc ) 
    {
      size_t tmp_index;
      size_t start[2], count[2];
      float tmp_float;
      short j;
      int nelev;
      float start_time = 0, end_time = 0;

      /* counters */
      int index = 0; /* Index of the current profile in the input file */

      /* Open file and get dimension sizes */
      handle_error(nc_open(argv[iarg], NC_NOWRITE, &ncid[NCFILEREAD]));
      handle_error(nc_inq_dimid(ncid[NCFILEREAD], "time", &time_dim[NCFILEREAD]));
      handle_error(nc_inq_dimlen(ncid[NCFILEREAD], time_dim[NCFILEREAD], &nrays));
      handle_error(nc_inq_dimid(ncid[NCFILEREAD], "range", &range_dim[NCFILEREAD]));
      handle_error(nc_inq_dimlen(ncid[NCFILEREAD], range_dim[NCFILEREAD], &ngates));
      
      /* Check file type, we only want to average fix(es), also known as stare(s) */
      char *scantype = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "scantype");
      handle_error(nc_inq_varid(ncid[NCFILEREAD], "elevation", &elid[NCFILEREAD]));
      handle_error(nc_inq_varndims(ncid[NCFILEREAD], elid[NCFILEREAD], &nelev));     

      /* Should also check whether elevation is 90! */
      if (scantype == NULL ||(strncmp(scantype,"FIX",3) == 0 && nelev == 0))
	{
	  /* Populate time struct from ncfile*/
	  current_time->year = get_short_attribute(ncid[NCFILEREAD], NC_GLOBAL, "year");
	  current_time->month = get_short_attribute(ncid[NCFILEREAD], NC_GLOBAL, "month");
	  current_time->day = get_short_attribute(ncid[NCFILEREAD], NC_GLOBAL, "day");

	  char *location = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "location");
	  if (location == NULL)
	    {
	      char *name = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "source");
	      /* If first character is a space! */
	      if (strspn(name," "))
		name++;
	      
	      location = strdup(name);
	      location[strcspn(name," ")]='\0';
	    }

	  char *ch = strdup(location);
	  llocation = ch;
	  while (*ch) {
	    *ch = tolower(*ch);
	    ++ch;
	  }


	  /* Open netCDF file for writing */		
	  sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd_%s_halo-doppler-lidar_averaged.nc", prefix, current_time->year, current_time->month, current_time->day, llocation);	
	  ncfilename[MAX_STRING_LENGTH-1] = '\0';      
	    
	  handle_error(nc_create(ncfilename, NC_CLOBBER, &ncid[NCFILEWRITE]));
	  fprintf(stderr,"Writing %s\n", ncfilename);

	  /* Copy global attributes */
	  doppler_lidar_copy_global_attributes(ncid, current_time, user, location);

	  /* Add a `command_line' attribute */
	  handle_error(nct_add_command_line(ncid[NCFILEWRITE], argc, argv));
	    
	  /* Copy some scalar variables */
	  doppler_lidar_copy_scalar_varatts(ncid);

	  /* Define dimensions and add coordinate variables */
	  handle_error(nc_inq_varid(ncid[NCFILEREAD], "time", &timeid[NCFILEREAD]));
	  doppler_lidar_write_coordvar_attributes(ncid[NCFILEWRITE], bs_dimids, &time_dim[NCFILEWRITE], &timeid[NCFILEWRITE], &range_dim[NCFILEWRITE], &rangeid[NCFILEWRITE], ngates, &lag_dim, &lagid, nlags, outputraw);

	  /* Add averaged variables */ 
	  doppler_lidar_write_averagedvar_attributes(ncid[NCFILEWRITE], bs_dimids, &mbetaid, &mvid, &svid, &nrid, missing_value, &plot_range);
	  
	  /* Add higher moment variables */ 
	  doppler_lidar_write_highermomentvar_attributes(ncid[NCFILEWRITE], bs_dimids, &skvid, &kvid, missing_value, &plot_range, ispulsepair);
	  
	  /* End define mode */
	  handle_error(nc_enddef(ncid[NCFILEWRITE]));
	    
	  /* Copy range variable */
	  doppler_lidar_copy_range_var(ncid, rangeid, ngates);

	  /* Read and process variables one profile at a time */
	  while ((index + 1) * average_nrays < nrays)
	    {
	      /* Read new profiles of data*/
	      for (j=0; j < average_nrays; j++) 
		{	      
		  tmp_index = (index*average_nrays)+j;
		  /* Get time of profile */
		  handle_error(nc_get_var1_float(ncid[NCFILEREAD], timeid[NCFILEREAD], &tmp_index, &end_time));

		  if (j == 0) 
		    {
		      start_time = end_time; /* Set start time here */
		      doppler_lidar_init_average_profile(&avprofile, ngates);
		    }
		  
		  doppler_lidar_copy_profile(ncid[NCFILEREAD],&profile, ngates, (index*average_nrays)+j);
		  doppler_lidar_average_profile(&profile, &avprofile, ngates);
		}
	      doppler_lidar_calculate_moments(&avprofile, ngates, average_nrays, missing_value);
	      
	      /* Write new profile of data */  
	      tmp_float = start_time + (end_time - start_time) / 2;
	      
	      tmp_index = index;
	      handle_error(nc_put_var1_float(ncid[NCFILEWRITE], timeid[NCFILEWRITE], &tmp_index, &tmp_float));

	      start[0] = index;
	      start[1] = 0;
	      count[0] = 1;
	      count[1] = ngates;
	      
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], mbetaid, start,
					     count, avprofile.mean_beta));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], mvid, start,
					     count, avprofile.mean_v));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], svid, start,
					     count, avprofile.sigma_v));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], skvid, start,
					     count, avprofile.skew_v));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], kvid, start,
					     count, avprofile.kurt_v));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], nrid, start,
					     count, avprofile.numrays));
	      
	      ++index;  
	    }
	  /* Close netCDF file */
	  handle_error(nc_close(ncid[NCFILEWRITE]));
	}
      /* Close netCDF file */
      handle_error(nc_close(ncid[NCFILEREAD]));
      ++iarg;
    }
  exit(0); 
}
