/* doppler_write.c - functions for writing doppler lidar data in netCDF */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netcdf.h>
#include <ctype.h>
#include <math.h> 

#include "dopplerlidar.h"
#include "nctools.h"

#define MAX_STRING_LENGTH 512
#define NUMLINES 20


int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s (%s)\n", \
	    __LINE__, nc_strerror(ncstatus), s); exit(-1); \
} 

static
void
add_string_attribute(int ncid, int fieldid, char *attname, char *value)
{
  if (value) {
    /* The +1 ensures that \0 is included at the end of each attribute
       in the netcdf file - is this necessary? */
    handle_error(nc_put_att_text(ncid, fieldid, attname,
				 strlen(value), value));
  }
}

static
void
add_attributes(int ncid, int fieldid, char *units, char *units_html, char *long_name, char *standard_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (units_html) {
    handle_error(nc_put_att_text(ncid, fieldid, "units_html",
				 strlen(units_html), units_html));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
  if (standard_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "standard_name",
				 strlen(standard_name), standard_name));
  }

}

static
void
add_missing_value(int ncid, int fieldid, float value)
{
  handle_error(nc_put_att_float(ncid, fieldid,
				"missing_value",
				NC_FLOAT, 1, &value));

  handle_error(nc_put_att_float(ncid, fieldid,
				"_FillValue",
				NC_FLOAT, 1, &value));
}

static
void
add_missing_value_short(int ncid, int fieldid, short value)
{
  handle_error(nc_put_att_short(ncid, fieldid,
				"missing_value",
				NC_SHORT, 1, &value));
  handle_error(nc_put_att_short(ncid, fieldid,
				"_FillValue",
				NC_SHORT, 1, &value));
}

static
char *
seek_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  if (nc_inq_atttype(ncid, varid, name, &type) != NC_NOERR
      || type != NC_CHAR) {
    return NULL;
  }

  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = malloc((length+1) * sizeof(char));
  if (!value) {
    fprintf(stderr, "Out of memory\n");
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}

static
short
get_short_attribute(int ncid, int varid, char *name)
{
  size_t length;
  short value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    fprintf(stderr, "Error: %s attribute is not of type short\n", name);
    exit(1);
  }
  if (length > 1) {
    fprintf(stderr, "Error: %s attribute has more than 1 element\n", name);
    exit(1);
  }
  handle_error(nc_get_att_short(ncid, varid, name, &value));
  return value;
}

void
doppler_lidar_populate_timestruct_nc(int ncid, int timeid, DopplerLidarTime *current_time)
{
  size_t tmp_index = 0;
  float  decimal_time;
  int ttid;

  /* Populate time struct from ncfile*/

  /* These come from global attributes */
  current_time->year  = get_short_attribute(ncid, NC_GLOBAL, "year");
  current_time->month = get_short_attribute(ncid, NC_GLOBAL, "month");
  current_time->day   = get_short_attribute(ncid, NC_GLOBAL, "day");

  /* These come from first value in variable 'time' */
  handle_error(nc_inq_varid(ncid, "time", &ttid));
  handle_error(nc_get_var1_float(ncid, ttid, &tmp_index, &decimal_time));

  current_time->hour   = (short) floor(decimal_time);
  current_time->minute = (short) floor((decimal_time - floor(decimal_time)) * 60) ;
  current_time->second = (short) (((decimal_time - floor(decimal_time)) * 3600) - floor((decimal_time - floor(decimal_time)) * 60) *60 );

}



void
doppler_lidar_copy_global_attributes(int ncid[3], DopplerLidarTime *current_time, char *user, char *location)
{  
  char  *tmp_name = '\0';
  short tmp_short = 0;
  
  /* Copy some global attributes */
  tmp_name = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "location");
  if (tmp_name != NULL)
    {
      handle_error(nc_put_att_text(ncid[NCFILEWRITE], NC_GLOBAL, "location", strlen(tmp_name), tmp_name));
    }
  else
    {
      handle_error(nc_put_att_text(ncid[NCFILEWRITE], NC_GLOBAL, "location", strlen(location), location));
    }
  tmp_name = '\0';	
  tmp_name = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "system");
  if (tmp_name != NULL)
    handle_error(nc_put_att_text(ncid[NCFILEWRITE], NC_GLOBAL, "system", strlen(tmp_name), tmp_name));
  tmp_name = '\0';	
  tmp_name = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "source");
  if (tmp_name != NULL)
    handle_error(nc_put_att_text(ncid[NCFILEWRITE], NC_GLOBAL, "source", strlen(tmp_name), tmp_name));
  tmp_name = '\0';	
  tmp_name = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "title");
  if (tmp_name != NULL)
    handle_error(nc_put_att_text(ncid[NCFILEWRITE], NC_GLOBAL, "title", strlen(tmp_name), tmp_name));

  handle_error(nc_put_att_short(ncid[NCFILEWRITE], NC_GLOBAL, "day", NC_SHORT, 1, &current_time->day));
  handle_error(nc_put_att_short(ncid[NCFILEWRITE], NC_GLOBAL, "month", NC_SHORT, 1, &current_time->month));
  handle_error(nc_put_att_short(ncid[NCFILEWRITE], NC_GLOBAL, "year", NC_SHORT, 1, &current_time->year));

  tmp_name = '\0';	
  tmp_name = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "scantype");
  if (tmp_name != NULL)
    handle_error(nc_put_att_text(ncid[NCFILEWRITE], NC_GLOBAL, "scantype", strlen(tmp_name), tmp_name));

  tmp_name = '\0';	
  tmp_name = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "scantype");
  if (tmp_name != NULL)
    handle_error(nc_put_att_text(ncid[NCFILEWRITE], NC_GLOBAL, "scantype", strlen(tmp_name), tmp_name));


  /* Does global attribute systemID exist?  */  
  if (nc_get_att_short(ncid[NCFILEREAD], NC_GLOBAL, "systemID", &tmp_short) == NC_NOERR)
    handle_error(nc_put_att_short(ncid[NCFILEWRITE], NC_GLOBAL, "systemID", NC_SHORT, 1,  &tmp_short));

  /* Add a `history' attribute */
  handle_error(nct_add_history(ncid[NCFILEWRITE], "Averaged data generated from (original resolution) netCDF data", user));
}

void
doppler_lidar_copy_scalar_varatts(int ncid[3])
{
  int j;
  short k, l = 0;
  char tmpname[NC_MAX_NAME+1];

  char *name = tmpname;
  nc_type var_type; 
  int ndims, nvars;
  int dimids[3];
  int natts;
  int varoutid[30];

  char   tmp_char;
  short  tmp_short;
  int    tmp_int;
  float  tmp_float;
  double tmp_double;
  
  handle_error(nc_inq_nvars(ncid[NCFILEREAD], &nvars));
  for (j = 0; j < nvars; j++) 
    {
      handle_error(nc_inq_var(ncid[NCFILEREAD], j, name, &var_type, &ndims, dimids, &natts));
      
      /* Only copying scalar variables */
      if (ndims == 0)
	{
	  handle_error(nc_def_var(ncid[NCFILEWRITE], name, var_type, 0, dimids, &varoutid[l]));

	  /* For each attribute */
	  for (k = 0; k < natts; k++) 
	    {
	      tmpname[0] = '\0';
	      /* Get attribute name (not interested in type or length) */
	      handle_error(nc_inq_attname(ncid[NCFILEREAD], j, k, name));
	      handle_error(nc_copy_att(ncid[NCFILEREAD], j, name, ncid[NCFILEWRITE], varoutid[l]));
	    }
	  /* End define mode */
	  handle_error(nc_enddef(ncid[NCFILEWRITE]));	    
	  
	  /* Write scalar variable */
	  if (var_type == NC_CHAR || var_type == NC_BYTE)
	    {
	      handle_error(nc_get_var_text(ncid[NCFILEREAD], j, &tmp_char));
	      handle_error(nc_put_var_text(ncid[NCFILEWRITE], varoutid[l], &tmp_char));
	    }
	  else if (var_type == NC_SHORT)
	    {
	      handle_error(nc_get_var_short(ncid[NCFILEREAD], j, &tmp_short));
	      handle_error(nc_put_var_short(ncid[NCFILEWRITE], varoutid[l], &tmp_short));
	    }
	  else if (var_type == NC_INT)
	    {
	      handle_error(nc_get_var_int(ncid[NCFILEREAD], j, &tmp_int));
	      handle_error(nc_put_var_int(ncid[NCFILEWRITE], varoutid[l], &tmp_int));
	    }
	  else if (var_type == NC_FLOAT)
	    {
	      handle_error(nc_get_var_float(ncid[NCFILEREAD], j, &tmp_float));
	      handle_error(nc_put_var_float(ncid[NCFILEWRITE], varoutid[l], &tmp_float));
	    }
	  else if (var_type == NC_DOUBLE)
	    {
	      handle_error(nc_get_var_double(ncid[NCFILEREAD], j, &tmp_double));
	      handle_error(nc_put_var_double(ncid[NCFILEWRITE], varoutid[l], &tmp_double));
	    }
	  /* Return to define mode */
	  handle_error(nc_redef(ncid[NCFILEWRITE]));
	  l++; /*increment variable id counter*/
	}
    }
}

void 
doppler_lidar_copy_range_var(int ncid[3], int rangeid[3], int ngates)
{
  float tmp_floatarray[ngates];
  
  handle_error(nc_inq_varid(ncid[NCFILEREAD], "range", &rangeid[NCFILEREAD]));
  handle_error(nc_get_var_float(ncid[NCFILEREAD], rangeid[NCFILEREAD], tmp_floatarray));
  handle_error(nc_put_var_float(ncid[NCFILEWRITE], rangeid[NCFILEWRITE], tmp_floatarray));
}

void
doppler_lidar_copy_range_var_wind(int ncid[2], DopplerLidarWindData *windprofile, int rangeid[2], int ngates)
{
  size_t start[2], count[2];

  start[0] = 0;
  count[0] = ngates;

  handle_error(nc_inq_varid(ncid[NCFILEREAD], "range", &rangeid[NCFILEREAD]));
  handle_error(nc_get_vara_float(ncid[NCFILEREAD], rangeid[NCFILEREAD], start, count, windprofile->range));
  handle_error(nc_put_vara_float(ncid[NCFILEWRITE], rangeid[NCFILEWRITE], start, count, windprofile->range));

}


void 
doppler_lidar_copy_profile(int ncid, DopplerLidarData *profile, int ngates, int index)
{
  size_t start[2], count[2];
  int paramid[7], azelid[4];
  int ndims, j;

  start[0] = index;
  start[1] = 0;
  count[0] = 1;
  count[1] = ngates;

  handle_error(nc_inq_varid(ncid, "beta", &paramid[0]));
  handle_error(nc_get_vara_float(ncid, paramid[0], start, count, profile->beta));
  handle_error(nc_inq_varid(ncid, "v", &paramid[1]));
  handle_error(nc_get_vara_float(ncid, paramid[1], start, count, profile->velocity));
  handle_error(nc_inq_varid(ncid, "signal", &paramid[2]));
  handle_error(nc_get_vara_float(ncid, paramid[2], start, count, profile->signal));
  handle_error(nc_inq_varid(ncid, "beta_raw", &paramid[3]));
  handle_error(nc_get_vara_float(ncid, paramid[3], start, count, profile->beta_raw));
  handle_error(nc_inq_varid(ncid, "v_raw", &paramid[4]));
  handle_error(nc_get_vara_float(ncid, paramid[4], start, count, profile->velocity_raw));



  /* Include these variables, if they exist, else set to NaN */
  if (nc_inq_varid(ncid, "beta_error", &paramid[5]) == NC_NOERR) {
    handle_error(nc_get_vara_float(ncid, paramid[5], start, count, profile->beta_fractional_error));
  }
  else {
    for (j = 0; j < ngates; j++)
      profile->beta_fractional_error[j] = NAN;
  }

  if (nc_inq_varid(ncid, "v_error", &paramid[6]) == NC_NOERR) {
    handle_error(nc_get_vara_float(ncid, paramid[6], start, count, profile->velocity_error));
  }
  else {
    for (j = 0; j < ngates; j++)
      profile->velocity_error[j] = NAN;
  }


  /* Check dimensions of azimuth and elevation - if constant */
  /* modify retrieval as appropriate */
    
  handle_error(nc_inq_varid(ncid, "elevation", &azelid[0]));
  handle_error(nc_inq_varndims(ncid, azelid[0], &ndims)); 
  if (ndims == 0) {
    handle_error(nc_get_var_float(ncid, azelid[0], profile->elevation));
  }
  else {
    handle_error(nc_get_var1_float(ncid, azelid[0], start, profile->elevation));
  }


  handle_error(nc_inq_varid(ncid, "azimuth", &azelid[1]));
  handle_error(nc_inq_varndims(ncid, azelid[1], &ndims));  
  if (ndims == 0) {
    handle_error(nc_get_var_float(ncid, azelid[1], profile->azimuth));
  }
  else {
    handle_error(nc_get_var1_float(ncid, azelid[1], start, profile->azimuth));
  }

  /* Include pitch and roll, if they exist */
  if (nc_inq_varid(ncid, "pitch", &azelid[2]) == NC_NOERR) {
    handle_error(nc_inq_varid(ncid, "roll", &azelid[3]));
    handle_error(nc_get_var1_float(ncid, azelid[2], start, &(profile->pitch)));
    handle_error(nc_get_var1_float(ncid, azelid[3], start, &(profile->roll)));
  }



}

