/* 
Generate depolarization from Doppler Lidar netCDF data

History
-------
2012-08-29 eoc started

Author - Ewan O'Connor
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netcdf.h>
#include <ctype.h>
#include <math.h> 

#include "dopplerlidar.h"
#include "nctools.h"

#define MAX_STRING_LENGTH 512
#define MAX_LINE_LEN 2048
#define MAX_NUM_PROFILES 86400

//#define NCFILEREAD  1
//#define NCFILEWRITE 0

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s (%s)\n", \
	    __LINE__, nc_strerror(ncstatus), s); exit(-1); \
} 

static
char *
seek_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  if (nc_inq_atttype(ncid, varid, name, &type) != NC_NOERR
      || type != NC_CHAR) {
    return NULL;
  }

  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = malloc((length+1) * sizeof(char));
  if (!value) {
    fprintf(stderr, "Out of memory\n");
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}

static
short
get_short_attribute(int ncid, int varid, char *name)
{
  size_t length;
  short value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    fprintf(stderr, "Error: %s attribute is not of type short\n", name);
    exit(1);
  }
  if (length > 1) {
    fprintf(stderr, "Error: %s attribute has more than 1 element\n", name);
    exit(1);
  }
  handle_error(nc_get_att_short(ncid, varid, name, &value));
  return value;
}


void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.nc  [file2.nc ...]\n"
	  "Options:\n"
          "    -crosspoldir dir      find crosspolar files in specified directory (assumes same\n"
          "                            naming convention as co-polar files)\n"
          "   -crosspolfile s        use specified crosspolar file (full or relative path),\n"
          "                            will only process first file in argument list\n" 
	  "   -correctdepol f        scale depolarization by factor f (i.e. if co and cross channels\n"
          "                            appear to have different calibration constants)\n"
	  " -noisethreshold d        use signal noise threshold to clean data (default = 1.018)\n"
	  "           -site s        override site name (default is to use location given in file)\n"
	  "         -prefix dir      write the files in specified directory\n"
	  "           -user s        name and email address of the user\n"
	  "        -verbose          write more\n",
	  argv[0]);
  exit(0);
}

int main(int argc, char **argv){
  char *filename = '\0';
  char *prefix = ".";
  char *crosspoldir = "./cross/";
  char *crosspolfile = NULL;

  DopplerLidarData profile[2];
  DopplerLidarData *rawprofile = profile;
  DopplerLidarDepolData depolprofile;

  char *location = NULL;
  char *llocation = NULL;
  char *user = NULL;

  DopplerLidarPlotRange plot_range = {{1e-7,    1e-4},
				      { -2.0,    2.0},
				      {  0.001,  1.0},
				      {  0.001, 10.0},
                                      {  0.0,    0.5},
                                      { -0.2,    0.2},
                                      {  0.001,  1.0},
                                      {  0.001, 10.0},
				      {  0.001,  1.0},
				      { -2.0,    2.0},
				      { -2.0,    2.0},
				      {  0.0,    1.0},
				      {-15.0,   15.0},
				      {-15.0,   15.0},
				      { -2.0,    2.0},
				      {  0.0,   15.0},
				      {  0.0,  360.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    2.0},
                                      {  0.0,   10.0}};  
  DopplerLidarTime time  = {0, 0, 0, 0, 0, 0};
  DopplerLidarTime *current_time = &time; /* initialise current */

  int   iarg = 0; 
  char  ncfilename[MAX_STRING_LENGTH];

  int   ncid[3];                            /* netCDF ID */
  int   time_dim[3], range_dim[3];          /* dimension IDs */
  int   timeid[3], rangeid[3], elid[3], azid[3];     /* variable IDs */
  int   lag_dim, lagid; /* Dummy IDs */
  int   nlags = 0; /* dummy var*/
 // int   heightid;
  int   paramid[6];                         /* parameter IDs */
  int   bs_dimids[2];                       /* variable shape */
  size_t   nrays[2], ngates[2];
  float missing_value=-999;
  /*float noise_threshold = 0;*/
  float noise_threshold = 1.018;
  float depol_correction = 1;

  float time_co[MAX_NUM_PROFILES], time_cr[MAX_NUM_PROFILES];
  float minimum_time_diff = 24; /* Time is in decimal hours*/
  float minimum_dt = 24;        /* Time is in decimal hours*/
  /* flags */
  char  correctdepol = 0;
  char  verbose = 0;
  char  location_override = 0;
  char  crosspolfile_specified = 0;
  char  outputraw = 0;

  /* Process command-line arguments */
  if (argc < 2)
    {
      Usage(argv);
      exit(0);
    }

  while (++iarg < argc) {

    if (argv[iarg][0] != '-') {
      /* Found a lidar filename.*/
      filename = argv[iarg];
      break;
    }    
    else if (strcmp(argv[iarg],"-crosspoldir") == 0) {
      if (argc > iarg+1) {
        crosspoldir = argv[++iarg];
      }
      else {
        fprintf(stderr,"Crosspolar directory not specified");
        exit(1);
      }
    }    
    else if (strcmp(argv[iarg],"-crosspolfile") == 0) {
      if (argc > iarg+1) {
        crosspolfile = argv[++iarg];
        crosspolfile_specified = 1;
    }
      else {
        fprintf(stderr,"Crosspolar file not specified");
        exit(1);
      }
    }
    else if (strcmp(argv[iarg], "-correctdepol") == 0) {
      if (argc > iarg+1) {
	depol_correction = atof(argv[++iarg]);
	correctdepol = 1;
      }
      else {
	fprintf(stderr,"Depolarization correction factor not specified");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg],"-noisethreshold") == 0) {
      if (argc > iarg+1) {
	noise_threshold = atof(argv[++iarg]); 
      }
      else {
	fprintf(stderr,"Noise threshold not specified");
	exit(1);
      }
    }

    else if (strcmp(argv[iarg],"-location") == 0) {
      if (argc > iarg+1) {
	location = argv[++iarg];
	location_override = 1;
      }
      else {
	fprintf(stderr,"Location not specified");
      }
    }
    else if (strcmp(argv[iarg],"-prefix") == 0) {
      if (argc > iarg+1) {
	prefix = argv[++iarg];
      }
      else {
	fprintf(stderr,"Prefix not specified");
	exit(1);
      }
    } 
    else if (strcmp(argv[iarg], "-verbose") == 0) {
      verbose = 1;
    }   
    else if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	user = argv[++iarg];
      }
      else {
	fprintf(stderr,"Error: user not specified\n");
	exit(1);
      }
    }
    else {
      fprintf(stderr, "Error: `%s' not understood or not accompanied "
	      "by enough subsequent arguments\n", argv[iarg]);
      Usage(argv);
      exit(1);
    }
  }

  /* cycle through filenames in argv list */
  while (iarg < argc ) 
    {
      size_t tmp_index;
      size_t start[2], count[2];
      float  tmp_float;
      int    j;
      int    nelev, naz;

      /* counters */
      int index_co = 0; /* Index of the current profile in the co-polar file */
      int index_cr = 0; /* Index of the current profile in the cross-polar file */
      int index    = 0; /* Index of the current profile in the output file */

      /* Open co file and get dimension sizes */
      handle_error(nc_open(argv[iarg], NC_NOWRITE, &ncid[NCFILECO]));
      handle_error(nc_inq_dimid(ncid[NCFILECO], "time", &time_dim[NCFILECO]));
      handle_error(nc_inq_dimlen(ncid[NCFILECO], time_dim[NCFILECO], &nrays[NCFILECO]));
      handle_error(nc_inq_dimid(ncid[NCFILECO], "range", &range_dim[NCFILECO]));
      handle_error(nc_inq_dimlen(ncid[NCFILECO], range_dim[NCFILECO], &ngates[NCFILECO]));


      /* Generate cross-polar filename to correspond to co-polar */
      if (!crosspolfile_specified) {
        char crosspolfilename[MAX_STRING_LENGTH];
        char *dot,  *slash;
        
        /* First generate crosspolar filename from copolar filename. */
        if (strlen(crosspoldir) == 1 && crosspoldir[0] == '.')
          strcpy(crosspolfilename,argv[iarg]);
        else {
          /* crosspolar files are in different directory */
          strncpy(crosspolfilename,crosspoldir,MAX_STRING_LENGTH);
          if (crosspolfilename[strlen(crosspolfilename)-1] == '/')
            crosspolfilename[strlen(crosspolfilename)-1] = '\0';
          slash = strrchr(argv[iarg], '/');
          if (!(slash == NULL)) 
            strcpy(crosspolfilename + strlen(crosspolfilename), slash);
          else 
            return 1;
        }

        /* cross filename is similar to co filename */
        /* may be cross.nc or cross_uc.nc */
        
        /* first search for co.nc */ 
        dot = strstr(crosspolfilename, "co.nc");
        if (dot == NULL) {
          /* Search for co_uc.nc */  
	  dot = strstr(crosspolfilename, "co_uc.nc");
          if (dot == NULL) 
            return 1;
          else 
            strncpy(dot,"cross_uc.nc",11);
        }
        else 
          strncpy(dot,"cross.nc",8);
          
          crosspolfilename[strlen(crosspolfilename)] = '\0';

        /* Should now have sensible crosspolar filename, assuming that file exists! */
        printf("Corresponding cross-polar file is %s \n",crosspolfilename );
        crosspolfile = &crosspolfilename[0];
      }    

      /* Open cross file and get dimension sizes */
      handle_error(nc_open(crosspolfile, NC_NOWRITE, &ncid[NCFILECROSS]));
      handle_error(nc_inq_dimid(ncid[NCFILECROSS], "time", &time_dim[NCFILECROSS]));
      handle_error(nc_inq_dimlen(ncid[NCFILECROSS], time_dim[NCFILECROSS], &nrays[NCFILECROSS]));
      handle_error(nc_inq_dimid(ncid[NCFILECROSS], "range", &range_dim[NCFILECROSS]));
      handle_error(nc_inq_dimlen(ncid[NCFILECROSS], range_dim[NCFILECROSS], &ngates[NCFILECROSS]));

      /* Check file type, we only want to read stare file(s) */
      char *scantype_co = seek_char_attribute(ncid[NCFILECO], NC_GLOBAL, "scantype");
      char *scantype_cr = seek_char_attribute(ncid[NCFILECROSS], NC_GLOBAL, "scantype");

      /* Return errors if files are not suitable */
      if (strncmp(scantype_co,"FIX",3) != 0 || strncmp(scantype_cr,"FIX",3) != 0) {
        fprintf(stderr, "Input files are not Stare files.\n");
        exit(1);
      }
      if (nrays[NCFILECO] < 20 || nrays[NCFILECROSS] < 20) {
        fprintf(stderr, "Not enough data to warrant file creation.\n");
        exit(1);
      }
      if (ngates[NCFILECO] != ngates[NCFILECROSS]) {
        fprintf(stderr, "Number of gates in co-polar and cross-polar files do not match.\n");
        exit(1);
      }

      if (strncmp(scantype_co,"FIX",3) == 0 && strncmp(scantype_cr,"FIX",3) == 0 && ngates[NCFILECO] == ngates[NCFILECROSS])
	{
          handle_error(nc_inq_varid(ncid[NCFILECO], "elevation", &elid[NCFILECO]));
          handle_error(nc_inq_varid(ncid[NCFILECO], "azimuth", &azid[NCFILECO]));
          handle_error(nc_inq_varndims(ncid[NCFILECO], elid[NCFILECO], &nelev));
          handle_error(nc_inq_varndims(ncid[NCFILECO], azid[NCFILECO], &naz));

	  /* Populate time struct from ncfile*/
	  current_time->year = get_short_attribute(ncid[NCFILECO], NC_GLOBAL, "year");
	  current_time->month = get_short_attribute(ncid[NCFILECO], NC_GLOBAL, "month");
	  current_time->day = get_short_attribute(ncid[NCFILECO], NC_GLOBAL, "day");

	  if (!location_override)
	    {
	      location = seek_char_attribute(ncid[NCFILECO], NC_GLOBAL, "location");
	      if (location == NULL)
		{
		  char *name = seek_char_attribute(ncid[NCFILECO], NC_GLOBAL, "source");
		  /* If first character is a space! */
		  if (strspn(name," "))
		    name++;
		  
		  location = strdup(name);
		  location[strcspn(name," ")]='\0';
		}
	    }

	  char *ch = strdup(location);
	  llocation = ch;
	  while (*ch) {
              if ( *ch == ' ' )
                *ch = '-';
              else 
	        *ch = tolower(*ch);
	    ++ch;
	  }


	  /* Open netCDF file for writing */		
	  sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd_%s_halo-doppler-lidar_depol.nc", prefix, current_time->year, current_time->month, current_time->day, llocation);	
	  ncfilename[MAX_STRING_LENGTH-1] = '\0';      
	    
	  handle_error(nc_create(ncfilename, NC_CLOBBER, &ncid[NCFILEWRITE]));
	  fprintf(stderr,"Writing %s\n", ncfilename);

	  /* Copy global attributes */
	  doppler_lidar_copy_global_attributes(ncid, current_time, user, location);

	  /* Add a `command_line' attribute */
	  handle_error(nct_add_command_line(ncid[NCFILEWRITE], argc, argv));
	    
	  /* Copy some scalar variable attributes */
	  doppler_lidar_copy_scalar_varatts(ncid);

	  /* Add dummy elevation variable (no azimuth) attribute */
//	  doppler_lidar_write_az_el_attributes(ncid[NCFILEWRITE], bs_dimids, &elid[NCFILEWRITE], &azid[NCFILEWRITE], 1, 0);

	  /*  Define dimensions and add coordinate variables */
	  handle_error(nc_inq_varid(ncid[NCFILECO], "time", &timeid[NCFILECO]));
          handle_error(nc_inq_varid(ncid[NCFILECROSS], "time", &timeid[NCFILECROSS]));
	  doppler_lidar_write_coordvar_attributes(ncid[NCFILEWRITE], bs_dimids, &time_dim[NCFILEWRITE], &timeid[NCFILEWRITE], &range_dim[NCFILEWRITE], &rangeid[NCFILEWRITE], ngates[NCFILECO], &lag_dim, &lagid, nlags, outputraw);
	  
	  /* Add depol variables */ 
	  doppler_lidar_write_depolvar_attributes(ncid[NCFILEWRITE], bs_dimids, paramid, missing_value, &plot_range);

	  /* End define mode */
	  handle_error(nc_enddef(ncid[NCFILEWRITE]));
	    
//	  /* Add dummy elevation variable */
//	  tmp_float = 90;
//	  handle_error(nc_put_var_float(ncid[NCFILEWRITE], elid[NCFILEWRITE], &tmp_float));  

	  /* Copy range variable */
	  doppler_lidar_copy_range_var(ncid, rangeid, ngates[NCFILECO]);

	  /* Read and process variables from nrays_depol co and cross profiles, may have */
          /* to skip some profiles to pair up co/cross in time. */
	  /* Default is one pair of profiles */
          /* Base all suitable profiles on what is available in cross file, not co! There are additional */
          /* vertical stares in co with no corresponding cross */

          /* Read time vectors */
     	  handle_error(nc_get_var_float(ncid[NCFILECO], timeid[NCFILECO], time_co));
          handle_error(nc_get_var_float(ncid[NCFILECROSS], timeid[NCFILECROSS], time_cr));

          /* Generate suitable minimum time between profile pairs, assuming not set*/
          for (j = 0; j < (int) nrays[NCFILECROSS]-1; j++) {
            if (fabs(time_cr[j+1] - time_cr[j]) < minimum_time_diff)
              minimum_time_diff = fabs(time_cr[j+1] - time_cr[j]);
          }
          minimum_time_diff = 3 * minimum_time_diff;

          while (index_cr < (int) nrays[NCFILECROSS])
	    {
	      /* time_cr[index_cr] is current time from cross profile */
              /* Find closest profile time in co profile */
              for (j = 0; j < (int) nrays[NCFILECO]; j++) {
                if ( fabs(time_cr[index_cr] - time_co[j]) < minimum_time_diff ) {   
                  minimum_dt = fabs(time_cr[index_cr] - time_co[j]);
                  index_co = j;
                }
              }
             
              /* Check that time difference for this pair is not too large */
              if (minimum_dt > minimum_time_diff) {
                ++index_cr;
                continue;
              }
              /* Found suitable pair of profiles */

              /* Read new profiles of data*/
              doppler_lidar_copy_profile(ncid[NCFILECO],&rawprofile[0], ngates[NCFILECO], index_co);
              doppler_lidar_copy_profile(ncid[NCFILECROSS],&rawprofile[1], ngates[NCFILECROSS], index_cr);
              /* calculate depolarization */
  	      doppler_lidar_depol_profile(rawprofile, &depolprofile, ngates[NCFILECO], noise_threshold, missing_value);
	      
	      /* Write new profile of data */  
	      tmp_float = time_cr[index_cr] + (time_co[index_co] - time_cr[index_cr]) / 2;
	      
	      tmp_index = index;
	      handle_error(nc_put_var1_float(ncid[NCFILEWRITE], timeid[NCFILEWRITE], &tmp_index, &tmp_float));

	      start[0] = index;
	      start[1] = 0;
	      count[0] = 1;
	      count[1] = ngates[NCFILECO];
	      
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], paramid[0], start,
					     count, depolprofile.beta));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], paramid[1], start,
					     count, depolprofile.velocity));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], paramid[2], start,
					     count, depolprofile.depolarization));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], paramid[3], start,
					     count, depolprofile.depolarization_error));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], paramid[4], start,
					     count, depolprofile.differential_v));
              handle_error(nc_put_vara_float(ncid[NCFILEWRITE], paramid[5], start,
                                             count, depolprofile.differential_v_error));

	      ++index;
              ++index_cr;
	    }
	  /* Close netCDF file */
	  handle_error(nc_close(ncid[NCFILEWRITE]));
	}
      /* Close netCDF file */
      handle_error(nc_close(ncid[NCFILECO]));
      handle_error(nc_close(ncid[NCFILECROSS]));
      ++iarg;
    }
  exit(0); 
}
