/* doppler_write.c - functions for writing doppler lidar data in netCDF */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netcdf.h>
#include <ctype.h>
#include <math.h> 
#include <time.h>

#include "dopplerlidar.h"
#include "nctools.h"
#include "svd.c"
#include "transpose_matrix.c"
#include "multiply_matrices.c"
#include "dsvd.c"


#define MAX_STRING_LENGTH 512
#define NUMLINES 20

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s (%s)\n", \
	    __LINE__, nc_strerror(ncstatus), s); exit(-1); \
} 

static
void
add_string_attribute(int ncid, int fieldid, char *attname, char *value)
{
  if (value) {
    /* The +1 ensures that \0 is included at the end of each attribute
       in the netcdf file - is this necessary? */
    handle_error(nc_put_att_text(ncid, fieldid, attname,
				 strlen(value), value));
  }
}

static
void
add_attributes(int ncid, int fieldid, char *units, char *units_html, char *long_name, char *standard_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (units_html) {
    handle_error(nc_put_att_text(ncid, fieldid, "units_html",
				 strlen(units_html), units_html));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
  if (standard_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "standard_name",
				 strlen(standard_name), standard_name));
  }

}

static
void
add_missing_value(int ncid, int fieldid, float value)
{
  handle_error(nc_put_att_float(ncid, fieldid,
				"missing_value",
				NC_FLOAT, 1, &value));

  handle_error(nc_put_att_float(ncid, fieldid,
				"_FillValue",
				NC_FLOAT, 1, &value));
}

static
void
add_missing_value_short(int ncid, int fieldid, short value)
{
  handle_error(nc_put_att_short(ncid, fieldid,
				"missing_value",
				NC_SHORT, 1, &value));
  handle_error(nc_put_att_short(ncid, fieldid,
				"_FillValue",
				NC_SHORT, 1, &value));
}


static
char *
seek_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  if (nc_inq_atttype(ncid, varid, name, &type) != NC_NOERR
      || type != NC_CHAR) {
    return NULL;
  }

  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = malloc((length+1) * sizeof(char));
  if (!value) {
    fprintf(stderr, "Out of memory\n");
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}

double **
dalloc2d(unsigned nrows, unsigned ncols)   
/* Allocates space for 'nrows' by 'ncols' matrix, sets elements to zero */
/* and returns double pointer to that matrix */   
{   
//  long size;   
  double **ppd;   
  unsigned i,j;   

    
  /* ppd is a pointer to pointer to double */   
  /* it points to the first element in an array of pointers to double */                    
  ppd = (double **) malloc ((size_t) nrows * sizeof(double *));   
                               
  if(ppd == NULL) {   
    fprintf(stderr,"Memory allocation error for matrix initialisation!");   
    exit(1);   
  }
  /* ppd[0] is a pointer to double */   
  /* points to the first double in the memory we have allocated */   
                                                                             
  ppd[0] = (double *) malloc ((size_t) nrows * ncols * sizeof(double));   

  if(ppd[0] == NULL) {   
    fprintf(stderr,"Memory allocation error for matrix initialisation!");   
    exit(1);
  }

  for (i=1; i<nrows; i++)   
    ppd[i] = ppd[0] + i * ncols;   

  for (i=0; i<nrows; i++)      /* fill the array in a portable way */   
    for (j=0; j<ncols; j++)   
       ppd[i][j] = 0.0;   
                                                                                                                                  
//  dclear2d(ppd, nrows, ncols);   
  return ppd;                 /* return the pointer */
}

double *   
dalloc1d(unsigned size)   
/* Allocates space for a 'nrows' by 'ncols' matrix and returns the  
 *  *  double pointer to that matrix */   
{   
  double *pd;   
  unsigned j;
 
  pd = (double *) malloc ((size_t) size * sizeof(double));   
  if(pd == NULL) {   
    fprintf(stderr,"Memory allocation error for 1D matrix initialisation!");   
    exit(1);   
  }   
   
//    dclear1d( pd, size);   
  for (j=0; j<size; j++)  /* fill the vector in a portable way */   
    pd[j] = 0.0;
     
    return pd;                  /* return the pointer */   
} 

void       
dfree1d(double * pd) 
{   
    free (pd);   
}   
      
void   
dfree2d(double **ppd)   
/* releases the memory allocated to a dynamic array that was created  
 *  * by dalloc2d.  This function correctly deallocates all memory. */   
{   
    /* DO NOT CHANGE THE ORDER OF THESE STATEMENTS! */   
    free(ppd[0]);               /* free data memory */   
    free(ppd);                  /* free array of pointers */   
}   

void 
doppler_lidar_init_average_profile(DopplerLidarAvData *avprofile, int ngates)
{
  int j;
 
  /* Set all summations to zero */
  for (j=0; j < ngates; j++) {
    avprofile->mean_beta[j] = 0;
    avprofile->mean_v[j]    = 0;
    avprofile->sigma_v[j]   = 0;
    avprofile->skew_v[j]    = 0;
    avprofile->kurt_v[j]    = 0;
    avprofile->numrays[j]   = 0;
  }
}

void
doppler_lidar_average_profile(DopplerLidarData *profile, DopplerLidarAvData *avprofile, int ngates)
{
  int j;
  
  /* Perform summations for meteorological targets only and keep a record of how many summed */
  for (j=0; j < ngates; j++) {	      
    if (profile->beta[j] > 0) {
      avprofile->mean_beta[j]  += profile->beta[j];
      avprofile->mean_v[j]  += profile->velocity[j];
      avprofile->sigma_v[j] += (profile->velocity[j] * profile->velocity[j]);	  
      avprofile->skew_v[j]  += (profile->velocity[j] * profile->velocity[j] * profile->velocity[j]);
      avprofile->kurt_v[j]  += (profile->velocity[j] * profile->velocity[j] * profile->velocity[j] * profile->velocity[j]);	
      avprofile->numrays[j]++;
    }
  }
}

void
doppler_lidar_calculate_moments(DopplerLidarAvData *avprofile, int ngates, int average_nrays, float missing_value)
{
  int j;

  /* Calculate the mean (1st moment) */  
  for (j=0; j < ngates; j++) {	
    avprofile->mean_v[j]    = avprofile->mean_v[j]    / avprofile->numrays[j];
    avprofile->mean_beta[j] = avprofile->mean_beta[j] / avprofile->numrays[j];
  }

  /* sigma_v, skew_v, kurt_v currently point to the 2nd, 3rd, 4th moments */
  /* Remember to calculate higher moments first before they are overwritten! */
  /* All moments are only calculated from non-noise targets */
  for (j=0; j < ngates; j++) {	   
    if (avprofile->numrays[j] > average_nrays / 4) {
      avprofile->kurt_v[j] = ((avprofile->kurt_v[j] / avprofile->numrays[j] - 
				   (4 * avprofile->mean_v[j] * avprofile->skew_v[j] / avprofile->numrays[j]) +
				   (6 * avprofile->mean_v[j] * avprofile->mean_v[j] * avprofile->sigma_v[j] / avprofile->numrays[j]) -
				   (3 * avprofile->mean_v[j] * avprofile->mean_v[j] * avprofile->mean_v[j] * avprofile->mean_v[j])) / 
				  pow(((avprofile->sigma_v[j] / avprofile->numrays[j]) 
				       - (avprofile->mean_v[j] * avprofile->mean_v[j])),2))-3;
      
      avprofile->skew_v[j] = (avprofile->skew_v[j] / avprofile->numrays[j] - 
				  (3 * avprofile->mean_v[j] * avprofile->sigma_v[j] / avprofile->numrays[j]) + 
				  (2 * avprofile->mean_v[j] * avprofile->mean_v[j] * avprofile->mean_v[j])) / 
	pow(((avprofile->sigma_v[j] / avprofile->numrays[j]) 
	     - (avprofile->mean_v[j] * avprofile->mean_v[j])),1.5);
      
      avprofile->sigma_v[j] = sqrt((avprofile->sigma_v[j] / avprofile->numrays[j]) 
				       - (avprofile->mean_v[j] * avprofile->mean_v[j]));
      
    }
    else {
      avprofile->mean_beta[j] = missing_value;
      avprofile->mean_v[j]    = missing_value;
      avprofile->sigma_v[j]   = missing_value;
      avprofile->skew_v[j]    = missing_value;
      avprofile->kurt_v[j]    = missing_value;
      /*      avprofile->numrays[j]    = missing_value;*/
    }
  }
}

void
doppler_lidar_wind_profile(DopplerLidarData *DBSprofile, DopplerLidarWindData *windprofile, int ngates, float convention, float missing_value)
{
  int j;
  int interp_index = 0; /* index for range vs height in interpolation */
  double theta[3],phi[3];
  double phi_correction = 0; 

  /* If DBS type is 3 beam */
  /* [0] refers to vertical beam */
  /* [1] refers to North beam */
  /* [2] refers to East beam */

  /* Use fmod to keep 0<=phi<360 */

  for (j=0; j < 3; j++) {
    theta[j] = (double) DBSprofile[j].elevation[0] * PI /180;
    phi[j]   = fmod( (double) DBSprofile[j].azimuth[0] * PI /180, 2*PI);
  }

  /* Check that we have orthogonal oblique beams (within 3 degrees) */
  if (fabs(atan2(sin(phi[2]- phi[1]), cos(phi[2]- phi[1]))-PI/2) > 0.05) 
    fprintf(stderr,"Warning: oblique beams are not orthogonal\n");

  /* First calculate height from (range * sin elevation) using North beam (2nd) elevation angle */
  for (j=0; j < ngates; j++)
    windprofile->height[j] = windprofile->range[j] * sin(theta[1]);

  for (j=0; j < ngates; j++) {
    /* Check interpolation index for range relative to height */
    if (j > 1 && windprofile->height[j] < windprofile->range[interp_index-1])
      interp_index--;

    /* Ignore first three gates for now. There is signal there though..*/
    /* should put a better test on SNR here*/
//    if (j > 2 && DBSprofile[0].signal[j] > noise_threshold && DBSprofile[1].signal[j] > noise_threshold && DBSprofile[2].signal[j] > noise_threshold && DBSprofile[0].velocity[j] > missing_value + 1 && DBSprofile[1].velocity[j] > missing_value + 1 && DBSprofile[2].velocity[j] > missing_value + 1 ) {

    if (j > 2 && DBSprofile[0].signal[j] > windprofile->noise_threshold && DBSprofile[1].signal[j] > windprofile->noise_threshold && DBSprofile[2].signal[j] > windprofile->noise_threshold) {

      /* Interpolate vertical wind to match off-zenith beams in terms of height (rather than range) */
      windprofile->wwind[j]  =  DBSprofile[0].velocity_raw[interp_index-1] +
                                  ( ((windprofile->height[j] - windprofile->range[interp_index-1]) / (windprofile->range[interp_index] - windprofile->range[interp_index-1])) 
                                  * (DBSprofile[0].velocity_raw[interp_index] - DBSprofile[0].velocity_raw[interp_index-1]));

//      windprofile->wwind[j]  = DBSprofile[0].velocity[j];


      /* Remember that theta (elevation) is the angle subtended from the ground, not zenith!! */
      windprofile->vwind[j]  = (DBSprofile[1].velocity_raw[j] - (DBSprofile[0].velocity_raw[j] * sin(theta[1]))) / cos(theta[1]);
      windprofile->uwind[j]  = (DBSprofile[2].velocity_raw[j] - (DBSprofile[0].velocity_raw[j] * sin(theta[2]))) / cos(theta[2]);

      windprofile->speed[j] = sqrt(windprofile->uwind[j]*windprofile->uwind[j] + windprofile->vwind[j]*windprofile->vwind[j]);
      windprofile->direction[j] = 180/PI * atan2(windprofile->uwind[j],windprofile->vwind[j]) + convention;

      /* Calculate errors in all components */
      windprofile->vwind_error[j]     = sqrt(DBSprofile[1].velocity_error[j]*DBSprofile[1].velocity_error[j] *
                                             ((1 + (DBSprofile[0].velocity_error[j] / DBSprofile[1].velocity_error[j]) 
                                                    * (DBSprofile[0].velocity_error[j] / DBSprofile[1].velocity_error[j])
                                               * sin(theta[1]) * sin(theta[1]) ) / 
                                              ( cos(theta[1]) * cos(theta[1]) ) )
                                            );

      windprofile->uwind_error[j]     = sqrt(DBSprofile[2].velocity_error[j]*DBSprofile[2].velocity_error[j] *
                                             ((1 + (DBSprofile[0].velocity_error[j] / DBSprofile[2].velocity_error[j])
                                                    * (DBSprofile[0].velocity_error[j] / DBSprofile[2].velocity_error[j]) 
                                               * sin(theta[2]) * sin(theta[2]) ) /
                                              ( cos(theta[1]) * cos(theta[1]) ) )
                                             );

      windprofile->wwind_error[j]     = DBSprofile[0].velocity_error[interp_index-1] +
                                  ( ((windprofile->height[j] - windprofile->range[interp_index-1]) / (windprofile->range[interp_index] - windprofile->range[interp_index-1]))
                                  * (DBSprofile[0].velocity_error[interp_index] - DBSprofile[0].velocity_error[interp_index-1]));
                                  
      windprofile->speed_error[j]     = sqrt(windprofile->uwind_error[j] * windprofile->uwind_error[j] + windprofile->vwind_error[j] * windprofile->vwind_error[j]);
      windprofile->direction_error[j] = sqrt(windprofile->vwind[j] * windprofile->vwind[j] * windprofile->uwind_error[j] * windprofile->uwind_error[j]  
                                             + windprofile->uwind[j] * windprofile->uwind[j] * windprofile->vwind_error[j] * windprofile->vwind_error[j]) 
                                        / (windprofile->speed[j] * windprofile->speed[j]) * 180 / PI;



      /* Correct wind direction */
      /* Are oblique beams at 0 and 90 degrees - test is for 1 degree in radians */
      if (abs(atan2(sin(phi[1]), cos(phi[1]))) > 0.018) {
        phi_correction = phi[1] * 180 / PI;
      }

      phi_correction = phi_correction + windprofile->azimuthal_correction;

      /* If there is an azimuthal correction to apply */
      if (abs(phi_correction) > 0.1) {
        /* Correct wind direction */
        windprofile->direction[j] = windprofile->direction[j] + phi_correction;

        /* Correct u and v */
        windprofile->vwind[j] = windprofile->speed[j] * cos(windprofile->direction[j] * PI /180); 
        windprofile->uwind[j] = windprofile->speed[j] * sin(windprofile->direction[j] * PI /180);

      }

      /* Check wind direction */
      if (windprofile->direction[j] < 0)
        windprofile->direction[j] = windprofile->direction[j]+360;
      if (windprofile->direction[j] >= 360)
        windprofile->direction[j] = windprofile->direction[j]-360;


    }
    else {
      windprofile->uwind[j]     = missing_value;
      windprofile->vwind[j]     = missing_value;
      windprofile->wwind[j]     = missing_value;
      windprofile->speed[j]     = missing_value;
      windprofile->direction[j] = missing_value;

      windprofile->uwind_error[j]     = missing_value;
      windprofile->vwind_error[j]     = missing_value;
      windprofile->wwind_error[j]     = missing_value;
      windprofile->speed_error[j]     = missing_value;
      windprofile->direction_error[j] = missing_value;
    }
    
    interp_index++;
  }
}


void
doppler_lidar_vad_profile_TL(DopplerLidarData *VADprofile, DopplerLidarWindData *windprofile, int ngates, int nrays_per_scan, float convention, float missing_value)
{
  int j,k;
//  double theta[MAX_BEAMS],phi[MAX_BEAMS];
  double theta[24],phi[24];
  double phi_correction = 0;
  double prefactor[3]; 
  double windcomponents[3];

  # define M 24
  # define N 3

  /* Set azimuth pointing correction */
  phi_correction = phi_correction + (double) windprofile->azimuthal_correction;

  /* Use fmod to keep 0<=phi<360 */
  /* note that theta is elevation angle from zenith, not horizontal */
  for (j=0; j < nrays_per_scan; j++) {
    theta[j] = (90 - (double) VADprofile[j].elevation[0]) * PI /180;
    phi[j]   = fmod( ( (double) VADprofile[j].azimuth[0] + phi_correction )* PI /180, 2*PI);
  }

  /* First calculate height from (range * sin elevation [use cos(elevation from zenith)]) using 1st beam elevation angle */
  for (j=0; j < ngates; j++)
    windprofile->height[j] = windprofile->range[j] * cos(theta[0]);

  /* Calculate the prefactor for u, v, and w */
  /* assuming theta is constant */
  prefactor[0] = 2 / (nrays_per_scan * sin(theta[0]));
  prefactor[1] = 2 / (nrays_per_scan * sin(theta[0]));
  prefactor[2] = 1 / (nrays_per_scan * cos(theta[0]));

  /* Calculate wind components for each range */
  for (k=0; k < ngates; k++) {
    /* Initialise summations for each range */
    windcomponents[0] = 0;
    windcomponents[1] = 0;
    windcomponents[2] = 0;

    for (j=0; j < nrays_per_scan; j++) {
      /* Summations for each range */
      windcomponents[0] += sin(phi[j]) * (double) VADprofile[j].velocity_raw[k];
      windcomponents[1] += cos(phi[j]) * (double) VADprofile[j].velocity_raw[k];
      windcomponents[2] +=               (double) VADprofile[j].velocity_raw[k];
    }

    windprofile->uwind[k]     = (float) ( prefactor[0] * windcomponents[0] );
    windprofile->vwind[k]     = (float) ( prefactor[1] * windcomponents[1] );
    windprofile->wwind[k]     = (float) ( prefactor[2] * windcomponents[2] );
    windprofile->speed[k]     = sqrt(windprofile->uwind[k]*windprofile->uwind[k] + windprofile->vwind[k]*windprofile->vwind[k]);
    windprofile->direction[k] = 180/PI * atan2(windprofile->uwind[k],windprofile->vwind[k]) + convention;

    /* Deterministic error */
    /* Calculated using departure from mean U */
    /* Initialise summations for each range */
    windcomponents[0] = 0;
    windcomponents[1] = 0;
    windcomponents[2] = 0;

    for (j=0; j < nrays_per_scan; j++) {
      /* radial velocity */
      /* Summations for each range */
      windcomponents[0] += sin(phi[j]) * abs( (cos(phi[j] - (double) atan2(windprofile->uwind[k],windprofile->vwind[k]) ) * (double) windprofile->speed[k] * cos(theta[j]) ) - (double) VADprofile[j].velocity_raw[k]);
      windcomponents[1] += cos(phi[j]) * abs( (cos(phi[j] - (double) atan2(windprofile->uwind[k],windprofile->vwind[k]) ) * (double) windprofile->speed[k] * cos(theta[j]) ) - (double) VADprofile[j].velocity_raw[k]);
      windcomponents[2] +=               abs( (cos(phi[j] - (double) atan2(windprofile->uwind[k],windprofile->vwind[k]) ) * (double) windprofile->speed[k] * cos(theta[j]) ) - (double) VADprofile[j].velocity_raw[k]);
    }    

    windprofile->uwind_error[k]     = (float) ( prefactor[0] * windcomponents[0] );
    windprofile->vwind_error[k]     = (float) ( prefactor[1] * windcomponents[1] );
    windprofile->wwind_error[k]     = (float) ( prefactor[2] * windcomponents[2] );
    windprofile->speed_error[k]     = sqrt(windprofile->uwind_error[k] * windprofile->uwind_error[k] + windprofile->vwind_error[k] * windprofile->vwind_error[k]);
    windprofile->direction_error[k] = sqrt(windprofile->vwind[k] * windprofile->vwind[k] * windprofile->uwind_error[k] * windprofile->uwind_error[k]
                                           + windprofile->uwind[k] * windprofile->uwind[k] * windprofile->vwind_error[k] * windprofile->vwind_error[k])
                                        / (windprofile->speed[k] * windprofile->speed[k]) * 180 / PI;

    /* Calculate mean SNR for each range */

    /* Initialise summations for each range */
    windcomponents[0] = 0;

    for (j=0; j < nrays_per_scan; j++) {
      /* Summations for each range */
      windcomponents[0] += (double) (VADprofile[j].signal[k] -1);
    }
    windprofile->mean_snr[k] = (float) (10 * log10( fabs(windcomponents[0] / nrays_per_scan)));

  }

  /* Apply stochastic error model? */
}

void
doppler_lidar_vad_profile_EP(DopplerLidarData *VADprofile, DopplerLidarWindData *windprofile, int ngates, int nrays_per_scan, float convention, float missing_value)
{
  int j,k,l;
//  double theta[MAX_BEAMS],phi[MAX_BEAMS];
  double theta[24],phi[24];
  double phi_correction = 0;

  # define M 24
  # define N 3
  double   A[M][N];
  double   At[N][M];
  double   Astar[N][M];
  double   U[M][N];
  double   Ut[N][M];
  double   V[N][N];
  double   D[M][N] = { };
  double   singular_values[N];
  double   dummy_array[N];
  double   D_U[M][N];
  double   D_V[N][N];
  double   D_D[N];
  double   Dstar[N][M];
  double   tolerance = 0;

  double   C1[N][M];
  double   C2[N][M];
  double   C3[N][N];
  double   rv[M][1];
  double   windcomponents[N][1];
  double   Z[M][N];
  double   CN[M][N];
  double   varCovar[M][M] = {};

  double  **a = NULL;
  double  **v = NULL;
  double  *d = NULL;
  double w[N];

  int  non_convergence;

  /* Use fmod to keep 0<=phi<360 */
  /* note that theta is elevation angle from zenith, not horizontal */ 
  for (j=0; j < nrays_per_scan; j++) {
    theta[j] = (90 - (double) VADprofile[j].elevation[0]) * PI /180;
    phi[j]   = fmod( (double) VADprofile[j].azimuth[0] * PI /180, 2*PI);
  }

  /* First calculate height from (range * sin elevation [use cos(elevation from zenith)]) using 1st beam elevation angle */
  for (j=0; j < ngates; j++)
    windprofile->height[j] = windprofile->range[j] * cos(theta[0]);


  /* Calculate the unit vectors for the scan pattern */
  for (j=0; j < nrays_per_scan; j++) {
    /* Unit vectors for each ray*/
    A[j][0] = sin(phi[j]) * sin(theta[j]);
    A[j][1] = cos(phi[j]) * sin(theta[j]);
    A[j][2] = cos(theta[j]);
  }


      printf("A :");
      for (k=0; k < 24; k++) {
        for (l=0; l < 3; l++)
          printf("%f ",A[k][l]);


        printf("\n");

        }
        printf("\n");


      //
      ////     dummy_array = (double*) malloc(N * sizeof(double));                    //
      ////     if (dummy_array == NULL) {printf(" No memory available\n"); exit(0); }

  non_convergence = Singular_Value_Decomposition((double*) A, M, N, (double*) U, singular_values, (double*) V, (double*) dummy_array);
  if (non_convergence < 0) printf(" Failed to converge (A) \n");


      printf("U :");
      for (k=0; k < 24; k++) {
        for (l=0; l < 24; l++)
          printf("%f ",U[k][l]);


        printf("\n");

        }
        printf("\n");

      printf("V :");
      for (k=0; k < 3; k++) {
        for (l=0; l < 3; l++)
          printf("%f ",V[k][l]);


        printf("\n");

        }
        printf("\n");




  /* Fill D */
  D[0][0] = singular_values[0];
  D[1][1] = singular_values[1];
  D[2][2] = singular_values[2];

printf("Singular values: %f %f %f\n\n",singular_values[0],singular_values[1],singular_values[2]);
// TEST 
// A=UDV'
  Transpose_Matrix(&D_V[0][0], &V[0][0], N, N);
  Multiply_Matrices(&Z[0][0], &U[0][0], M, M, &D[0][0], N);

      printf("UD :");
      for (k=0; k < 24; k++) {
        for (l=0; l < 3; l++)
          printf("%f ",D_V[k][l]);


        printf("\n");

        }
        printf("\n");



  Multiply_Matrices(&CN[0][0], &Z[0][0], M, N, &D_V[0][0], N);

      printf("UDV' :");
      for (k=0; k < 24; k++) {
        for (l=0; l < 3; l++)
          printf("%f ",CN[k][l]);


        printf("\n");

        }
        printf("\n");

///TEST different routine
//void svdcmp(double **a, int m, int n, double w[], double **v)
// Pass in A,overwritten to be U
//
//prepare a from A
a = dalloc2d(M, N);
for (k=0; k < 24; k++) {
  for (l=0; l < 3; l++)
    a[k][l] = A[k][l];
}


    printf("a :");
      for (k=0; k < 24; k++) {
        for (l=0; l < 3; l++)
          printf("%f ",a[k][l]);


        printf("\n");

        }
        printf("\n");
v = dalloc2d(N, N);
    printf("Test1 :\n");

   dsvd(a ,M,N,w,v);
      printf("Test2 :\n");


    printf("u :");
      for (k=0; k < 24; k++) {
        for (l=0; l < 3; l++)
          printf("%f ",a[k][l]);


        printf("\n");

        }
        printf("\n");

    printf("v :");
      for (k=0; k < 3; k++) {
        for (l=0; l < 3; l++)
          printf("%f ",v[k][l]);


        printf("\n");

        }
        printf("\n");
    printf("D (w) :");
        for (l=0; l < 3; l++)
          printf("%f ",w[l]);

        printf("\n");
        printf("\n");




  /* Clear and refill D */
  for (k=0; k < 3; k++)
    for (l=0; l < 3; l++)
          D[k][l] = 0;

  D[0][0] = singular_values[0];
  D[1][1] = singular_values[1];
  D[2][2] = singular_values[2];

  /* refill V */
  for (k=0; k < 3; k++)
    for (l=0; l < 3; l++)
          V[k][l] = v[k][l];


// TEST 
// A=UDV'
   Transpose_Matrix(&D_V[0][0], &V[0][0], N, N);
   Multiply_Matrices(&Z[0][0], &U[0][0], M, M, &D[0][0], N);

      printf("UD :");
      for (k=0; k < 24; k++) {
        for (l=0; l < 3; l++)
          printf("%f ",D_V[k][l]);


        printf("\n");

        }
        printf("\n");



  Multiply_Matrices(&CN[0][0], &Z[0][0], M, N, &D_V[0][0], N);

      printf("UDV' :");
      for (k=0; k < 24; k++) {
        for (l=0; l < 3; l++)
          printf("%f ",CN[k][l]);


        printf("\n");

        }
        printf("\n");


dfree2d(a);
dfree2d(v);

//
//
//
//
//
//



  /* First get svd(D)*/
  non_convergence = Singular_Value_Decomposition((double*) D, M, N, (double*) D_U, D_D, (double*) D_V, (double*) dummy_array);
  if (non_convergence < 0) printf(" Failed to converge (D)\n");
  
  /* Then get pinv(D) */
  Singular_Value_Decomposition_Inverse((double*) D_U, D_D, (double*) D_V, tolerance, M, N, (double*) Dstar);

      printf("pinv(D) :");
      for (k=0; k < 3; k++) {
        for (l=0; l < 24; l++)
          printf("%f ",Dstar[k][l]);


        printf("\n");

        }
        printf("\n");



  /* Calculate wind components for each range */
  for (j=0; j < ngates; j++) {
    // Formally: windcomponents = V * pinv(D) * transpose(U) * rv;
    Transpose_Matrix(&Ut[0][0], &U[0][0], M, N);

    for (k=0; k < nrays_per_scan; k++) 
      rv[k][0] = (double) VADprofile[k].velocity_raw[j];

    // Interim stage 1 //
    Multiply_Matrices(&C1[0][0], &V[0][0], N, N, &Dstar[0][0], M);
    // Interim stage 2 //
    Multiply_Matrices(&C2[0][0], &C1[0][0], N, M, &Ut[0][0], M);
    // Final stage to get windcomponents//
    Multiply_Matrices(&windcomponents[0][0], &C2[0][0], N, M, &rv[0][0], 1);

    windprofile->uwind[j]     = (float) windcomponents[0][0];
    windprofile->vwind[j]     = (float) windcomponents[1][0];
    windprofile->wwind[j]     = (float) windcomponents[2][0];
    windprofile->speed[j]     = sqrt(windprofile->uwind[j]*windprofile->uwind[j] + windprofile->vwind[j]*windprofile->vwind[j]);
    windprofile->direction[j] = 180/PI * atan2(windprofile->uwind[j],windprofile->vwind[j]) + convention;

    /* Calculate errors in all components */
    Transpose_Matrix(&At[0][0], &A[0][0], M, N);
    Multiply_Matrices(&C3[0][0], &At[0][0], N, M, &A[0][0], N);
    C3[0][0] = 1 / sqrt(C3[0][0]);
    C3[1][1] = 1 / sqrt(C3[1][1]);
    C3[2][2] = 1 / sqrt(C3[2][2]);
    Multiply_Matrices(&Z[0][0], &A[0][0], M, N, &C3[0][0], N);

    non_convergence = Singular_Value_Decomposition((double*) Z, M, N, (double*) U, singular_values, (double*) V, (double*) dummy_array);
    if (non_convergence < 0) printf(" Failed to converge (A) \n");

    //i            CN = max(svd(Z))/min(svd(Z));
    // need to do thresholding to get sensible answer here
    //

    /* Instrumental errors */
    /* Initialize variance-covariance matrix (eye matrix) */
    for (k=0; k < nrays_per_scan; k++)
      varCovar[k][k] = (double) VADprofile[k].velocity_error[j];

    /* Get pinv(A) */
    Singular_Value_Decomposition_Inverse((double*) U, singular_values, (double*) V, tolerance, M, N, (double*) Astar);
    Transpose_Matrix(&C1[0][0], &Astar[0][0], N, M);
    // Interim stage //
    Multiply_Matrices(&C2[0][0], &Astar[0][0], N, M, &varCovar[0][0], M);
    // C_v_v is C3 //
    Multiply_Matrices(&C3[0][0], &C2[0][0], N, N, &C1[0][0], N);

    windprofile->uwind_error[j]     = sqrt(C3[0][0]);
    windprofile->vwind_error[j]     = sqrt(C3[1][1]);
    windprofile->wwind_error[j]     = sqrt(C3[2][2]);
    // check these
    windprofile->speed_error[j]     = sqrt(windprofile->uwind_error[j] * windprofile->uwind_error[j] + windprofile->vwind_error[j] * windprofile->vwind_error[j]);
    windprofile->direction_error[j] = sqrt(windprofile->vwind[j] * windprofile->vwind[j] * windprofile->uwind_error[j] * windprofile->uwind_error[j]
                                           + windprofile->uwind[j] * windprofile->uwind[j] * windprofile->vwind_error[j] * windprofile->vwind_error[j])
                                        / (windprofile->speed[j] * windprofile->speed[j]) * 180 / PI;
//
//
//
//      //                                   


if (j==3){
      printf("C2 :");
      for (k=0; k < 3; k++) {
        for (l=0; l < 24; l++)
          printf("%f ",C2[k][l]);


        printf("\n");

        }
        printf("\n");

      printf("rv :");
      for (k=0; k < 24; k++) {
        for (l=0; l < 1; l++)
          printf("%f ",rv[k][l]);


        printf("\n");

        }
        printf("\n");

printf("CN: %f %f %f\n",singular_values[0],singular_values[1],singular_values[2]);

}


//      windprofile->vwind_error[j]     = sqrt(DBSprofile[1].velocity_error[j]*DBSprofile[1].velocity_error[j] *
//                                             ((1 + (DBSprofile[0].velocity_error[j] / DBSprofile[1].velocity_error[j])
//                                                    * (DBSprofile[0].velocity_error[j] / DBSprofile[1].velocity_error[j])
//                                               * sin(theta[1]) * sin(theta[1]) ) /
//                                              ( cos(theta[1]) * cos(theta[1]) ) )
//                                            );

//      windprofile->uwind_error[j]     = sqrt(DBSprofile[2].velocity_error[j]*DBSprofile[2].velocity_error[j] *
//                                             ((1 + (DBSprofile[0].velocity_error[j] / DBSprofile[2].velocity_error[j])
//                                                    * (DBSprofile[0].velocity_error[j] / DBSprofile[2].velocity_error[j])
//                                               * sin(theta[2]) * sin(theta[2]) ) /
//                                              ( cos(theta[1]) * cos(theta[1]) ) )
//                                             );

//      windprofile->wwind_error[j]     = DBSprofile[0].velocity_error[interp_index-1] +
//                                  ( ((windprofile->height[j] - windprofile->range[interp_index-1]) / (windprofile->range[interp_index] - windprofile->range[interp_index-1]))
//                                  * (DBSprofile[0].velocity_error[interp_index] - DBSprofile[0].velocity_error[interp_index-1]));

//      windprofile->speed_error[j]     = sqrt(windprofile->uwind_error[j] * windprofile->uwind_error[j] + windprofile->vwind_error[j] * windprofile->vwind_error[j]);
//      windprofile->direction_error[j] = sqrt(windprofile->vwind[j] * windprofile->vwind[j] * windprofile->uwind_error[j] * windprofile->uwind_error[j]
//                                             + windprofile->uwind[j] * windprofile->uwind[j] * windprofile->vwind_error[j] * windprofile->vwind_error[j])
//                                        / (windprofile->speed[j] * windprofile->speed[j]) * 180 / PI;



      /* Correct wind direction */
      /* Are oblique beams at 0 and 90 degrees - test is for 1 degree in radians */
//      if (abs(atan2(sin(phi[1]), cos(phi[1]))) > 0.018) {
//        phi_correction = phi[1] * 180 / PI;
//      }

//      phi_correction = phi_correction + windprofile->azimuthal_correction;

      /* If there is an azimuthal correction to apply */
//      if (abs(phi_correction) > 0.1) {
        /* Correct wind direction */
//        windprofile->direction[j] = windprofile->direction[j] + phi_correction;

        /* Correct u and v */
//        windprofile->vwind[j] = windprofile->speed[j] * cos(windprofile->direction[j] * PI /180);
//        windprofile->uwind[j] = windprofile->speed[j] * sin(windprofile->direction[j] * PI /180);

//      }

      /* Check wind direction */
//      if (windprofile->direction[j] < 0)
//        windprofile->direction[j] = windprofile->direction[j]+360;
//      if (windprofile->direction[j] >= 360)
//        windprofile->direction[j] = windprofile->direction[j]-360;


//    }
//    else {
//      windprofile->uwind[j]     = missing_value;
//      windprofile->vwind[j]     = missing_value;
//      windprofile->wwind[j]     = missing_value;
//      windprofile->speed[j]     = missing_value;
//      windprofile->direction[j] = missing_value;

//      windprofile->uwind_error[j]     = missing_value;
//      windprofile->vwind_error[j]     = missing_value;
//      windprofile->wwind_error[j]     = missing_value;
//      windprofile->speed_error[j]     = missing_value;
//      windprofile->direction_error[j] = missing_value;
//    }
  }

printf("Finished calculating winds\n");

printf("winds at 4: %f %f %f %f %f\n",windprofile->uwind[3],windprofile->vwind[3],windprofile->wwind[3],windprofile->speed[3], windprofile->direction[3]);
printf("winds at 14: %f %f %f %f %f\n",windprofile->uwind[13],windprofile->vwind[13],windprofile->wwind[13],windprofile->speed[13], windprofile->direction[13]);



}




int
doppler_lidar_correct_focus(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value)
{
  int j;

  float alpha = 0.01;
  float het_area, effective_area, T;
  float range, nt, het_cnr;

  het_area = PI * (0.7 * spec->lens_diameter / 2) * (0.7 * spec->lens_diameter / 2);

  for (j=0; j <  header->number_gates_m1; j++) {
    range = (header->gate_length * (float) j )  + (header->gate_length / 2);

    /* focus is at infinity */
    if (spec->focus < 0) 
      nt = het_area / spec->wavelength * (-1 / range); 
    else    
      nt = het_area / spec->wavelength * (1/spec->focus  - 1 / range); 

    effective_area = het_area / (1 + nt * nt);
    T = pow(10, (-1 * alpha * range / 5000));
    het_cnr = 0.7 * 0.4 * 0.6 * effective_area * C_LIGHT * spec->beam_energy * T / (2 * range * range * HNU * spec->bandwidth);

    profile->beta_raw[j] = (profile->signal[j] - 1) / het_cnr;

  }


  return 0;
}


int
doppler_lidar_uncertainty(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value)
{
  int j;
  float deltav = 2;
  float ff, Np, alpha, signal;
  
  /*-- Calculate uncertainty in velocity estimates --*/

  /* deltav: typical signal spectral width - best value close to 2? */
  /* ff: ratio of signal spectral width to bandwidth */
  /*     can also be written as ratio of atmospheric spectral width
           to 2*nyquist velocity, approx 0.05 */

  //  ff = 0.05;  
  ff = deltav / (2 * spec->nyquist);

  for (j=0; j <  header->number_gates_m1; j++) {
    signal = profile->signal[j] - 1;
    if (signal < 0)
      signal = signal * -1;

   /* Np: accumulated photon count */
    Np = (float) header->gate_length_pts * (float) header->number_pulses_m1 * signal;

    /* alpha: ratio of photon count to speckle count */
    alpha =  signal / (sqrt(2 * PI) * ff);
    
    /* bandwidth here is in terms of Nyquist*2*/
    profile->velocity_error[j] = spec->nyquist * 2 * sqrt( sqrt(8) * ff * ff / (Np * alpha) * (pow(1 +  alpha / (sqrt(2 * PI)),2)));
    
    /*-- Calculate uncertainty in beta estimates --*/
    /* fractional error provided here */ 
    profile->beta_fractional_error[j] = (1 / sqrt(header->number_pulses_m1)) * (1 + ( 1 / signal));

  }

  return 0;
}


void
doppler_lidar_depol_profile(DopplerLidarData *rawprofile, DopplerLidarDepolData *depolprofile, int ngates, float noise_threshold, float missing_value)
{
  int j;

  /* Depol from pair of profiles */
  /* [0] refers to co-polar */
  /* [1] refers to cross-polar */

  for (j=0; j < ngates; j++) {
    depolprofile->beta[j] = rawprofile[0].beta[j];    
    depolprofile->velocity[j] = rawprofile[0].velocity[j];
 
    /* Ignore first three gates for now. There is signal there though..*/
    /* should put a better test on SNR here*/
    if (j > 2 && rawprofile[0].signal[j] > noise_threshold && rawprofile[1].signal[j] > noise_threshold && rawprofile[0].velocity[j] > missing_value + 1 && rawprofile[1].velocity[j] > missing_value + 1) {
      
      depolprofile->depolarization[j] = rawprofile[1].beta[j] / rawprofile[0].beta[j];
      depolprofile->differential_v[j] = rawprofile[1].velocity[j] - rawprofile[0].velocity[j];
 
      depolprofile->depolarization_error[j] = sqrt((rawprofile[0].beta_fractional_error[j] * rawprofile[0].beta_fractional_error[j]) + (rawprofile[1].beta_fractional_error[j] * rawprofile[1].beta_fractional_error[j]));

      depolprofile->differential_v_error[j] = sqrt((rawprofile[0].velocity_error[j] * rawprofile[0].velocity_error[j]) + (rawprofile[1].velocity_error[j] * rawprofile[1].velocity_error[j]));

    }
    else {
      depolprofile->depolarization[j]       = missing_value;
      depolprofile->differential_v[j]       = missing_value;
      depolprofile->depolarization_error[j] = missing_value;
      depolprofile->differential_v_error[j] = missing_value;

    }
  }
}

int
doppler_lidar_calculate_pulsepair_v(DopplerLidarACFData *acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec)
{
  int i;

  for (i=0; i < header->number_gates_m1 * header->number_raw_samples_gate; i++) {
    acfprofile->numrays[i] = 1;

//fprintf(stdout, "%f %f   %d    %f %f\n", acfprofile->acf_imag[(i * header->number_lags) + 1],acfprofile->acf_real[(i * header->number_lags) + 1], acfprofile->numrays[i], (spec->acf_imag_offset_applied * (float) acfprofile->numrays[i]),(spec->acf_real_offset_applied * (float) acfprofile->numrays[i]));
    acfprofile->v_pp[i] = atan2(acfprofile->acf_imag[(i * header->number_lags) + 1] + (spec->acf_imag_offset_applied * (float) acfprofile->numrays[i]), acfprofile->acf_real[(i * header->number_lags) + 1] + (spec->acf_real_offset_applied * (float) acfprofile->numrays[i])) * spec->nyquist / PI  + spec->v_offset_applied;

  }

  return 0;
}

int
doppler_lidar_average_raw_profile_grid(FILE *fp, DopplerLidarRawData *rawprofile, DopplerLidarACFData *acfprofile, DopplerLidarACFData *ave_acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec, int argc, char **argv, char *hpl)
{
  if (fread(rawprofile, sizeof(double) * 6 * header->number_gates_m1 * header->number_raw_samples_gate, 1, fp)) {
    int i;

    for (i=0; i < header->number_gates_m1 * header->number_lags * header->number_raw_samples_gate; i++) {
      acfprofile->acf_real[i] = (float) rawprofile->acf_raw[i*2];
      acfprofile->acf_imag[i] = (float) rawprofile->acf_raw[i*2+1];
    }

    for (i=0; i < header->number_gates_m1 * header->number_raw_samples_gate; i++) {
      acfprofile->v_pp[i] = atan2(acfprofile->acf_imag[(i * header->number_lags) + 1] - 79, acfprofile->acf_real[(i * header->number_lags) + 1] -40);
    }
    return 0;
  }
  else
    return -1;
}

int
doppler_lidar_simple_clean_raw_ray(DopplerLidarACFData *acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec)
{
  int i;

  for (i=0; i < header->number_gates_m1 * header->number_raw_samples_gate; i++) {

    acfprofile->signal[i] = acfprofile->acf_real[(i * header->number_lags)];
    /* Remove data below threshold in spec */
    if (acfprofile->signal[i] < spec->nominal_noise_floor + 5*spec->nominal_noise_std) {
      acfprofile->signal[i] = spec->missing_value;
      acfprofile->v_pp[i] = spec->missing_value;
    }
  }
  return 0;
}




