/* 
Convert Doppler Lidar ASCII data to netCDF 

History
-------
2006-09-25 eoc started
2006-10-03 eoc revision for variable file formats
2013-02-28 eoc major revision for inclusion in main trunk - include variable number of range gate

Author - Ewan O'Connor
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netcdf.h>
#include <ctype.h>
#include <math.h> 

#include "dopplerlidar.h"
#include "nctools.h"

#define MAX_STRING_LENGTH 512
#define MAX_LINE_LEN 8048


int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}

static
void
add_string_attribute(int ncid, int fieldid, char *attname, char *value)
{
  if (value) {
    /* The +1 ensures that \0 is included at the end of each attribute
       in the netcdf file - is this necessary? */
    handle_error(nc_put_att_text(ncid, fieldid, attname,
				 strlen(value), value));
  }
}

static
void
add_attributes(int ncid, int fieldid, char *units, char *units_html, char *long_name, char *standard_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (units_html) {
    handle_error(nc_put_att_text(ncid, fieldid, "units_html",
				 strlen(units_html), units_html));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
  if (standard_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "standard_name",
				 strlen(standard_name), standard_name));
  }

}


static
void
add_missing_value(int ncid, int fieldid, float value)
{
  handle_error(nc_put_att_float(ncid, fieldid,
				"missing_value",
				NC_FLOAT, 1, &value));

  handle_error(nc_put_att_float(ncid, fieldid,
				"_FillValue",
				NC_FLOAT, 1, &value));
}

static
void
add_missing_value_short(int ncid, int fieldid, short value)
{
  handle_error(nc_put_att_short(ncid, fieldid,
				"missing_value",
				NC_SHORT, 1, &value));
  handle_error(nc_put_att_short(ncid, fieldid,
				"_FillValue",
				NC_SHORT, 1, &value));
}

void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.txt  [file2.txt ...]\n"
	  "Options:\n"
	  "General data processing options (f=float d=integer s=string):\n"
	  "       -filedate d        only read data from this date\n"
	  " -noisethreshold d        use noise threshold when clean data\n"
	  "        -avetime d        average onto a regular d-second grid \n"
	  "        -averays d        average d rays (always integer)\n"
	  "         -cleanv f        clean velocity below beta threshold \n"
	  "      -outputraw          output acfs \n"
	  "  -highermoments          include higher moments (skew and kurtosis) \n"
	  "        -noclean          don't remove speckle\n"
	  "      -elevation f        elevation of instrument (from horizontal)\n"
	  "       -altitude f        Altitude (in m) of instrument above sea level\n"
	  "       -location s f f f  location of the lidar (e.g. Chilbolton), followed by\n"
	  "                             its latitude (N), longitude (E) and altitude (m)\n"
	  "           -site s        use known lat, lon and alt for site s - currently\n"
	  "                             understood sites are Chilbolton\n"
	  "         -prefix dir      write the files in specified directory\n"
	  "         -acfdir dir      read the raw acf files from specified directory (default ./)\n"
	  "         -hpldir dir      read the hpl files from specified directory (default ./ or acfdir)\n"
	  "                            instrument parameters are recorded in hpl files as\n"
	  "                            acf files contain the raw acfs only.\n"
	  "           -user s        name and email address of the user\n"
	  "        -verbose          write more\n",
	  argv[0]);
  exit(0);
}

int main(int argc, char **argv){
  FILE *fp_acf = NULL;
  char *filename = '\0';
  char *prefix = ".";
  char *acfdir = ".";
  char *hpldir = ".";
  DopplerLidarHeader header;
  /*  DopplerLidarData profile; */
  DopplerLidarSpec spec;
  DopplerLidarRawData rawprofile;
  DopplerLidarACFData acfprofile, rawacfprofile;
  float latitude = 51.1445;
  float longitude = 358.563;
  float altitude = 90;
  char *location = NULL;
  char *llocation = NULL;
  char *scantype = NULL;
  char *lscantype = NULL;
  char *user = NULL;
  char *institution = NULL;
  DopplerLidarPlotRange plot_range = {{1e-7,    1e-4},
                                      { -2.0,    2.0},
                                      {  0.001,  1.0},
                                      {  0.001, 10.0},
                                      {  0.0,    0.5},
                                      { -1.0,    1.0},
                                      {  0.001,  1.0},
                                      {  0.001, 10.0},
                                      {  0.001,  1.0},
                                      { -2.0,    2.0},
                                      { -2.0,    2.0},
                                      {  0.0,    1.0},
                                      {-15.0,   15.0},
                                      {-15.0,   15.0},
                                      { -2.0,    2.0},
                                      {  0.0,   15.0},
                                      {  0.0,  360.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    2.0},
                                      {  0.0,   10.0}};

  //float plot_range_signal[] = {990, 1500};
  float plot_range_num[] = {0.0, 1.0};

  /* Need to keep previous time to check when midnight occurs */
  DopplerLidarTime time, last_time = {0, 0, 0, 0, 0, 0};
  DopplerLidarTime *current_time = &time; /* initialise current */
  DopplerLidarTime *prev_time = &last_time;    /* and next block */

  char ncfilename[MAX_STRING_LENGTH];
  unsigned short int fday=0, fmonth=0, fyear=0;
  unsigned short int fhour=0, fmin=0, fsec=0;


  float time_increment = 0;

  int  ncid;                         /* netCDF ID */
  int  time_dim, range_dim, lag_dim; /* dimension IDs */
  int  timeid, rangeid;/* variable ID's */
  int  latid, lonid, altid;	
  int  siid, vid, nrid, svid, skvid, kvid;	
  int  numpulsesm1id, numpulsesm2id, numlagsid, numptsid;
  int  lensid, divid, focusid, lagid, arid, aiid;
  int  energyid, waveid, prfid, pulselengthid, elid, azid, noiseid, mnid, bandid;
  int  bs_dimids[3];                 /* variable shape */
  float missing_value=-999;

  /* counters */
  int index = 0; /* Index of the current profile in the input file */

  /* flags */
  char newfile = 1;
  char filestatus = 1;
  char filecrossstatus = 1;
  char fileopen = 0;
  char ncfileopen = 0;
  char noclean = 0;
  char cleanv = 0;
  char verbose = 0;
  char avetime = 0;
  char averays = 0;
  char outputraw = 0;  
  char highermoments = 0; 
  char isverticalstare = 1;
  char addcrosspol = 0;
  char ispulsepair = 1;

 /* Initialise array for spec */
  spec.wavelength = 0.0000015;
  spec.beam_energy = 0.00001;
  spec.lens_diameter = 0.060;
  spec.divergence = 0.000033;
  spec.prf = 15000;
  spec.pulse_duration = 0.0000002;
  spec.bandwidth = 5e7;
  spec.nyquist = 9.68;
  spec.elevation = 90;
  spec.noise_threshold = 1.0009;
  spec.cleanv_threshold = 0;
  spec.numseconds = 0;
  spec.calfactor = 1;
  spec.focus = -1;
  spec.nominal_noise_floor = 990;
  spec.nominal_noise_std = 6;
  spec.acf_real_offset_applied = -40;
  spec.acf_imag_offset_applied = -79;
  spec.v_offset_applied = -0.32;
  spec.missing_value = missing_value;
  spec.numrays = 1;
  spec.applycalfactor = 0;
  spec.applyfocuscorrection = 0;
  spec.cleanrawrayflag = 1;
  spec.calculate_epsilon = 0;



  /* Initialise array for header */
  header.hdr_lines = 12;
  /*  header.date_stamp = NULL;*/
  header.gate_length = 36;
  header.number_measurements = 0;
  header.system_id = 0;  
  header.number_gates_m1 = 276;
  header.number_gates_m2 = 0;
  header.number_pulses_m1 = 0;
  header.number_pulses_m2 = 0;
  header.number_lags = 3;
  header.gate_length_pts = 0;
  header.number_raw_samples_gate = 1;
  header.focus_range_m1 = 0;
  header.focus_range_m2 = 0;
  header.num_integrations = 0;
  header.num_rhis_in_scan = 0;
  header.num_rays = 0;
  header.scantype = 0;
  header.scantypereq = 0;
  header.year = 0;
  header.month = 0; 
  header.day = 0;
  header.hour = 0; 
  header.minute = 0;
  header.second = 0;
  header.v_resolution = 0;
  header.iarg = 0;

  /* Process command-line arguments */
  if (argc < 2)
    {
      Usage(argv);
      exit(0);
    }

  while (++header.iarg < argc) {

    if (argv[header.iarg][0] != '-') {
      /* Found a lidar filename.*/
      filename = argv[header.iarg];
      break;
    }
    else if (strcmp(argv[header.iarg],"-filedate") == 0 && header.iarg+1<argc) {
      if (sscanf(argv[++header.iarg],"%04hu%02hu%02hu",
		 &fyear, &fmonth, &fday) < 3) {
	fprintf(stderr, "Failed to read date: should be in the form YYYYMMDD");
      }
    }

    else if (strcmp(argv[header.iarg],"-starttime") == 0) {
      if (argc <= header.iarg+1 || sscanf(argv[++header.iarg],"%02hu%02hu%02hu",
		 &fhour, &fmin, &fsec) < 3) {
	fprintf(stderr, "Failed to read start time: should be in the form HHMMSS");
      }
      if (argc <= header.iarg+1 || sscanf(argv[++header.iarg],"%f",
		 &time_increment) < 1) {
	fprintf(stderr, "Failed to read time increment as float");
      }
    }
    else if (strcmp(argv[header.iarg],"-noclean") == 0 ) {
      spec.cleanrawrayflag = 0;
    }
    else if (strcmp(argv[header.iarg],"-outputraw") == 0 ) {
      outputraw = 1;
    }
    else if (strcmp(argv[header.iarg],"-highermoments") == 0 ) {
      highermoments = 1;
    }
    else if (strcmp(argv[header.iarg],"-noisethreshold") == 0 && header.iarg+1<argc) {
      spec.noise_threshold = atof(argv[++header.iarg]);
    }

    else if (strcmp(argv[header.iarg], "-avetime") == 0) {
      if (argc > header.iarg+1) {
	spec.numseconds = atof(argv[++header.iarg]);
	avetime = 1;
      }
      else {
	fprintf(stderr,"Averaging time not specified");
	exit(1);
      }
    }
    else if (strcmp(argv[header.iarg], "-averays") == 0) {
      if (argc > header.iarg+1) {
	spec.numrays = (unsigned short) atoi(argv[++header.iarg]);
	plot_range_num[1] = (float) spec.numrays;
	if (spec.numrays > 1)
	  averays = 1;
      }
      else {
	fprintf(stderr,"Number of rays to average not specified");
	exit(1);
      }
    }
    else if (strcmp(argv[header.iarg], "-cleanv") == 0) {
      if (argc > header.iarg+1) {
	spec.cleanv_threshold = atof(argv[++header.iarg]);
	cleanv = 1;
      }
      else {
	fprintf(stderr,"Elevation not specified");
	exit(1);
      }
    }
    else if (strcmp(argv[header.iarg],"-prefix") == 0) {
      if (argc > header.iarg+1) {
	prefix = argv[++header.iarg];
      }
      else {
	fprintf(stderr,"Prefix not specified");
	exit(1);
      }
    }   
    else if (strcmp(argv[header.iarg],"-acfdir") == 0) {
      if (argc > header.iarg+1) {
	acfdir = argv[++header.iarg];
      }
      else {
	fprintf(stderr,"acf directory not specified");
	exit(1);
      }
    } 
    else if (strcmp(argv[header.iarg],"-hpldir") == 0) {
      if (argc > header.iarg+1) {
	hpldir = argv[++header.iarg];
      }
      else {
	fprintf(stderr,"hpl directory not specified");
	exit(1);
      }
    }   
    else if (strcmp(argv[header.iarg], "-verbose") == 0) {
      verbose = 1;
    }   
    else if (strcmp(argv[header.iarg],"-user") == 0) {
      if (argc > header.iarg+1) {
	user = argv[++header.iarg];
      }
      else {
	fprintf(stderr,"Error: user not specified\n");
	exit(1);
      }
    }
    else if (strcmp(argv[header.iarg], "-elevation") == 0) {
      if (argc > header.iarg+1) {
	spec.elevation = atof(argv[++header.iarg]);
      }
      else {
	fprintf(stderr,"Elevation not specified");
	exit(1);
      }
    } 
    else if (strcmp(argv[header.iarg], "-altitude") == 0) {
      if (argc > header.iarg+1) {
	altitude = atof(argv[++header.iarg]);
      }
      else {
	fprintf(stderr,"Altitude not specified");
	exit(1);
      }
    }   
    else if (strcmp(argv[header.iarg],"-location") == 0) {
      if (argc > header.iarg+4) {
	location = argv[++header.iarg];
	latitude = atof(argv[++header.iarg]);
	longitude = atof(argv[++header.iarg]);
	altitude = atof(argv[++header.iarg]);
      }
      else {
	fprintf(stderr,"Not all location arguments were provided");
      }
    }
    else if (strcmp(argv[header.iarg], "-site") == 0) {
      if (argc > header.iarg+1) {
	++header.iarg;
	if (strcasecmp(argv[header.iarg], "Chilbolton") == 0) {
	  location = "Chilbolton";
	  latitude = 51.1445;
	  longitude = 358.563;
	  altitude = 90;

	}
	else {
	  fprintf(stderr,"Site unknown: I only know about Chilbolton");
	}
      }
    }
    else {
      fprintf(stderr, "Error: `%s' not understood or not accompanied "
	      "by enough subsequent arguments\n", argv[header.iarg]);
      Usage(argv);
      exit(1);
    }
  }

  /* cycle through filenames in argv list */
  while (header.iarg < argc ) 
    {
      size_t tmp_index;
      size_t start[3], count[3];
      char *tmp_name;
      float tmp_float;
      short tmp_short;
      short tmp_short_var[3];
      short j;
      int tmp_int;
      /*      DopplerLidarTime *tmpptr;*/

      if (!fileopen) 
	{
	  if ((fp_acf = fopen(argv[header.iarg], "r")) == NULL) 
	    {
	      fprintf(stderr,"can't open input file\n");
	      exit(1);
	    }
	  if (verbose)
	    fprintf(stderr, "Reading file %s\n", argv[header.iarg]);
	  fileopen = 1;
	  
	  if (doppler_lidar_open_hpl(&header, argc, argv, hpldir)) 
	    exit(1);

          if (header.scantype == 2) {
            scantype = "FIX";
            lscantype = "tilt";
            spec.elevation = 86;
            isverticalstare = 1;
          }
          else if (header.scantype == 3) {
            scantype = "RHI";
            lscantype = "rhi";
          }
          else if (header.scantype == 4) {
            scantype = "VAD";
            lscantype = "vad";
          }
          else if (header.scantype == 5) {
            scantype = "WIND";
            lscantype = "wind";
          }
          else if (header.scantype == 6) {
            scantype = "USER";
            lscantype = "user";
          }
          else {
            scantype = "FIX";
            lscantype = "fix";
          }

          /* Number of raw gates depends on output format */
          /*   Early versions - same number of gates as .hpl data */
          /*   Later version  - sampled at number of points per gate * number of gates */
          /*                    typical oversampling is 6-10 points per gate, adjust if more! */
          /*   Note that there are currently 3 lags per sample, again adjust if this changes.. */

          /* Test which format using system_id*/
          /* Not known when format changed, possibly when system_id was added to header */
          /* therefore use presence of this for now*/

          if (header.system_id > 0)
            header.number_raw_samples_gate = (int) header.gate_length_pts;

	}

      if (averays || spec.numrays > 1) 
	filestatus = doppler_lidar_average_raw_profile(fp_acf, &rawprofile, &rawacfprofile, &acfprofile, &header, &spec, argc, argv, hpldir);
      else
	filestatus = doppler_lidar_single_raw_profile(fp_acf, &rawprofile, &acfprofile, &header, &spec, argc, argv, hpldir);

      /* For Stare scantype - is it vertical? */
//      if ((header.scantype > 2 && header.scantypereq == 0) || (header.scantype == 1 && (rawprofile.elevation[0] < 89.9 || rawprofile.elevation[0] > 90.01)))
      if ((header.scantype > 2 && header.scantypereq == 0))
        isverticalstare = 0;

      if (filestatus > 0)
	{
	  if (verbose)
	    fprintf(stderr, "Problem reading file, skipping rest of file.\n");
	  continue;
	}
      else if (filestatus < 0)
	{
	  /*	  fclose(fp_acf);
		  fileopen = 0;*/
	  ++header.iarg;           
	  continue;
	}
      
      
      if (!location) {
	location = "Chilbolton";
	llocation = "chilbolton";
      }
      else {
	char *ch = strdup(location);
	llocation = ch;
	while (*ch) {
	  *ch = tolower(*ch);
	  ++ch;
	}
      }
      
      char *ch = strdup(argv[header.iarg]);
      /* search from the end for the start of the filename. */
      ch = strrchr(argv[header.iarg], '/');
      
      sscanf(ch,"/%2hu%2hu%2hu%*u.raw", &(current_time->day), &(current_time->month), &(current_time->year));
      current_time->year = current_time->year + 2000;

      
      if (ncfileopen && (prev_time->year < current_time->year 
			 || prev_time->month < current_time->month 
			 || prev_time->day < current_time->day))
	{
	  newfile = 1;
	  prev_time->year = current_time->year;
	  prev_time->month = current_time->month;
	  prev_time->day = current_time->day;
	  
	  /* Close netCDF file */
	  handle_error(nc_close(ncid));
	  index = 0;
	  ncfileopen = 0;
	}
      
      if (newfile) {
	
	float tmp[header.number_gates_m1 * header.number_raw_samples_gate];
	char tmp_string[MAX_STRING_LENGTH];

	/* Open netCDF file for writing */		
	sprintf(ncfilename,"%s/%4hu%2.2hu%2.2hu_%s_halo-doppler-lidar-acf.nc", prefix, current_time->year, current_time->month, current_time->day, llocation);
	ncfilename[MAX_STRING_LENGTH-1] = '\0';	
	
	handle_error(nc_create(ncfilename, NC_CLOBBER, &ncid));
	fprintf(stderr,"Writing %s\n",ncfilename);
	fileopen = 1;
      
	/* Add some global attributes */
        tmp_int = (signed int) header.system_id;
        doppler_lidar_write_global_attributes(ncid, current_time, location, user, scantype, institution, header.system_id);

	/* Add a `command_line' attribute */
	handle_error(nct_add_command_line(ncid, argc, argv));
	
        /* Add some scalar variables */
        doppler_lidar_write_instrument_attributes(ncid, bs_dimids, &numpulsesm1id, &numpulsesm2id, &numlagsid, &numptsid, &lensid, &divid, &focusid, &energyid, &waveid, &prfid, &pulselengthid, &noiseid, &bandid);

        if (spec.applyfocuscorrection)
          nct_append_attribute(ncid, focusid, "comments", "This variable has been modified from the original value held in the raw .hpl files.","");

        /* Add location variables */
        doppler_lidar_write_location_attributes(ncid, bs_dimids, &latid, &lonid, &altid);


        /* Define dimensions and add coordinate variables */
        doppler_lidar_write_coordvar_attributes(ncid, bs_dimids, &time_dim, &timeid, &range_dim, &rangeid, header.number_gates_m1 * header.number_raw_samples_gate, &lag_dim, &lagid, header.number_lags, 1);

	/* define  variables */
	bs_dimids[0] = time_dim;
	bs_dimids[1] = range_dim;
	bs_dimids[2] = lag_dim;


        /* Add raw variables */
        doppler_lidar_write_rawvar_attributes(&header, &spec, ncid, bs_dimids, &arid, &aiid, &siid, &vid, missing_value, outputraw, &plot_range);

	if (averays || avetime)
          doppler_lidar_write_averagedrawvar_attributes(ncid, bs_dimids, &svid, &nrid, missing_value, &plot_range);
	
	if (highermoments && (averays || avetime)) 
          doppler_lidar_write_highermomentvar_attributes(ncid, bs_dimids, &skvid, &kvid, missing_value, &plot_range, ispulsepair);
	
	/* End define mode */
	handle_error(nc_enddef(ncid));

	/* write scalar values into netCDF variables */
	tmp_index = 0;

	handle_error(nc_put_var1_float(ncid, latid, &tmp_index, &latitude));
	handle_error(nc_put_var1_float(ncid, lonid, &tmp_index, &longitude));
	handle_error(nc_put_var1_float(ncid, altid, &tmp_index, &altitude));
        tmp_int = (signed int) header.number_pulses_m1;
	handle_error(nc_put_var1_int(ncid, numpulsesm1id, &tmp_index, &tmp_int));
        tmp_int = (signed int) header.number_pulses_m2;
	handle_error(nc_put_var1_int(ncid, numpulsesm2id, &tmp_index, &tmp_int));
	tmp_short = (short int) header.number_lags;
	handle_error(nc_put_var1_short(ncid, numlagsid, &tmp_index, &tmp_short));
        tmp_short = (short int) header.gate_length_pts;
        handle_error(nc_put_var1_short(ncid, numptsid, &tmp_index, &tmp_short));
	handle_error(nc_put_var1_float(ncid, lensid, &tmp_index, &(spec.lens_diameter)));
	handle_error(nc_put_var1_float(ncid, divid, &tmp_index, &(spec.divergence)));
        if (spec.applyfocuscorrection)
          tmp_short = (short int) spec.focus;
        else
          tmp_short = (short int) header.focus_range_m1;
	handle_error(nc_put_var1_short(ncid, focusid, &tmp_index, &tmp_short));
	handle_error(nc_put_var1_float(ncid, energyid, &tmp_index, &(spec.beam_energy)));
	handle_error(nc_put_var1_float(ncid, waveid, &tmp_index, &(spec.wavelength)));
	handle_error(nc_put_var1_float(ncid, prfid, &tmp_index, &(spec.prf)));
	handle_error(nc_put_var1_float(ncid, pulselengthid, &tmp_index, &(spec.pulse_duration)));
	handle_error(nc_put_var1_float(ncid, bandid, &tmp_index, &(spec.bandwidth)));
	handle_error(nc_put_var1_float(ncid, noiseid, &tmp_index, &(spec.noise_threshold)));

        if (header.scantype < 3) {
          if (isverticalstare) {
            handle_error(nc_put_var1_float(ncid, elid, &tmp_index, &(spec.elevation)));
          }
  //        else {
  //          handle_error(nc_put_var1_float(ncid, elid, &tmp_index, &(profile.elevation[0])));
   //         handle_error(nc_put_var1_float(ncid, azid, &tmp_index, &(profile.azimuth[0])));
    //      }
 //       }
 //       else if (header.scantypereq == 1) {
  //        handle_error(nc_put_var1_float(ncid, elid, &tmp_index, &(profile.elevation[0])));
        }

	/* write vector values into netCDF variables */
	for (j = 0; j < header.number_gates_m1 * header.number_raw_samples_gate ; j++) 
	  {

	    tmp[j] = ((float) header.gate_length / (float) header.number_raw_samples_gate * (float) j * sin( spec.elevation / 180 * PI)) + ( (float) header.gate_length / (float) header.number_raw_samples_gate / 2);

	  }
	handle_error(nc_put_var_float(ncid, rangeid, tmp));
	for (j = 0; j < header.number_lags ; j++) 
	  {
	    tmp_short_var[j] = (short) j;
	  }

	handle_error(nc_put_var_short(ncid, lagid, tmp_short_var));

	newfile = 0;
      }


      /*      if (index == 0)
	{
	  for (j = 0; j < header.number_gates_m1 ; j++) 
	    printf("%d %f %f %f %f %f %e %e\n", j, acfprofile.signal[j], acfprofile.v_pp[j], acfprofile.sigma_v_pp[j],acfprofile.skew_v_pp[j], acfprofile.kurt_v_pp[j],acfprofile.acf_real[(j*3)+1],acfprofile.acf_imag[(j*3)+2]);
	    }*/
	  /* Add a new profile of data */  
	  /*      tmp_float = ((float) current_time->hour) + ((float) current_time->minute) / 60.0 + (((float) current_time->second)) / 3600.0;*/


      start[0] = index;
      start[1] = 0;
      start[2] = 0;

      count[0] = 1;
      count[1] = header.number_gates_m1 * header.number_raw_samples_gate;
      count[2] = header.number_lags;
      
      tmp_float = ((32.5000 / 6 ) * ((float) index + 0.5) * (float) spec.numrays)  / 3600;
      handle_error(nc_put_var1_float(ncid, timeid, start, &tmp_float));

      if (outputraw) {
	handle_error(nc_put_vara_float(ncid, arid, start,
				       count, acfprofile.acf_real));
	handle_error(nc_put_vara_float(ncid, aiid, start,
				       count, acfprofile.acf_imag));
      }
      if (spec.cleanrawrayflag)
	handle_error(nc_put_vara_float(ncid, siid, start,
				       count, acfprofile.signal));
	  


      handle_error(nc_put_vara_float(ncid, vid, start,
				     count, acfprofile.v_pp));

      if (averays || avetime) {
	handle_error(nc_put_vara_float(ncid, svid, start,
				       count, acfprofile.sigma_v_pp));
	handle_error(nc_put_vara_ushort(ncid, nrid, start, count, acfprofile.numrays));
      }

      if (highermoments && (averays || avetime)) {
	handle_error(nc_put_vara_float(ncid, skvid, start,
				       count, acfprofile.skew_v_pp));
	handle_error(nc_put_vara_float(ncid, kvid, start,
				       count, acfprofile.kurt_v_pp));
      }
      
      ++index;  
      
      /*tmpptr=current_time;     current points to current block */
      /*current_time=prev_time;  current points to prev block */
      /*prev_time=tmpptr;        prev points to current block which is now available for new block */
      
    } 

  
  if (!newfile) 
    {
      /* Close netCDF file */
      handle_error(nc_close(ncid));
    }
  
  fclose(fp_acf);
  exit(0); 
}


