/* doppler_read.c - functions for reading doppler lidar data*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netcdf.h>
#include <ctype.h>
#include <math.h> 

#include "dopplerlidar.h"
#include "nctools.h"

#define MAX_STRING_LENGTH 512
#define NUMLINES 65

/* Must cope with dodgy ray - in this case jump to next \n */
static
int 
SkipRay(FILE * fp)
{
  char c1;

  c1='\0';
  while ((c1=getc(fp)) != '\n' ) {
    if (c1 == EOF)
      return 0;
  }
  return -1;
}

void
doppler_lidar_check_date(DopplerLidarHeader *header, DopplerLidarTime *current_time)
{
  current_time->day = header->day;
  current_time->month = header->month;
  current_time->year = header->year + 2000;
}

int
doppler_lidar_open_hpl(DopplerLidarHeader *header, int argc, char **argv , char *hpldir)
{
  FILE *fp_hpl;
  char hplfilename[MAX_STRING_LENGTH];
  char *dot, *slash;

  /* First generate hpl filename from acf (raw) filename. */
  /* If -hpldir hpldirectory not given look in same directory as acf data */

  if (strlen(hpldir) == 1 && hpldir[0] == '.')
    strcpy(hplfilename,argv[header->iarg]);
  else {
    /* hpl files are in different directory */
    strncpy(hplfilename,hpldir,MAX_STRING_LENGTH);
    if (hplfilename[strlen(hplfilename)-1] == '/')
      hplfilename[strlen(hplfilename)-1] = '\0';

    slash = strrchr(argv[header->iarg], '/');
    if (!(slash == NULL))
      strcpy(hplfilename + strlen(hplfilename), slash);
    else
      return 1;
  }

  /* filename is similar to acf (raw) filename */
  dot = strrchr(hplfilename, '.');
  if (!(dot == NULL)) {
    strncpy(dot+1,"hpl",3);
  }
  else
    return 1;


  /* Should now have sensible hplfilename, assuming that file exists! */
  printf("%s \n",hplfilename );

  if ((fp_hpl = fopen(hplfilename, "r")) == NULL) {
    fprintf(stderr,"Can't open hpl file %s\n", hplfilename);
    exit(1);
  }

  if (doppler_lidar_read_hdr(fp_hpl, header)) {
    fprintf(stderr,"Error reading header for %s\n", hplfilename);
    exit(1);
  }

  return 0;

}

int
doppler_lidar_read_hdr(FILE *fp, DopplerLidarHeader *header)
{
  char * p;
  int i;

  /* Chilbolton header*/ 
  /*  Filename        c:\Lidar\Data\Proc\2007\200712\20071203\03120711.hpl
      Range gate length (m)   36.0
      Number of measurements  1
      Number of gates in M1   276
      Number of gates in M2   0
      Pulses/ray in M1        60000
      Pulses/ray in M2        0
      Number of Lags  3
      Gate length (pts)       6
      Focus range M1  65535
      Start time      11:00:18
      Resolution (m/s)        0.02*/

  /* Salford header */
  /*  Filename        c:\Lidar\Data\Proc\2007\200711\20071101\Stare_01110700.hpl
      Range gate length (m)   30.0
      Number of gates 66
      Number of Lags  3
      Pulses/ray      20000
      Integrations per loop   20000
      Scan type       Stare
      No. of rays in scan     1
      Gate length (pts)       6
      Focus range     801
      Start time      00:00:02
      Resolution (m/s)        0.02
      No. of RHIs     0*/

  /* Reading header */
  /* Filename:       C:\Lidar\Data\Proc\2010\201004\20100420\Stare_200410.hpl */
  /* Operating mode: N/A */
  /* Number of gates:        250 */ 
  /* Range gate length (m):  30.0 */
  /* Gate length (pts):      6 */
  /* Number of Lags: 3 */
  /* Pulses/ray:     20000 */
  /* No. of rays in file:    1 */
  /* Scan type:      Stare */
  /* Focus range:    65535 */
  /* Start time:     00:00:08 */
  /* Resolution (m/s):       0.02 */
  /* Altitude of measurement (center of gate) = (range gate + 0.5) * Gate length */
  /* Data line 1: Decimal time (hours)  Azimuth (degrees)  Elevation (degrees) */
  /* f7.4,1x,f6.2,1x,f6.2 */
  /* Data line 2: Range Gate  Doppler (m/s)  Intensity (SNR + 1)  Beta (m-1 sr-1) */
  /* i3,1x,f6.4,1x,f8.6,1x,e12.6 - repeat for no. gates */
  /* **** */

  /* New Header for streamline version 10 Mar 2014*/
  /* Filename now without directorY, otherwise similar */
  /* Filename:       Stare_34_20140401_00.hpl */

  /* New Header for streamline version XX Mar 2018 */
  /* Scan type now has name describing scan, so could be arbitrary */
  /* Appears to always include ' - stepped' though */
  /* Scan type:      S_vad15 - stepped */

  static char * datestampLine     = "Filename";
  static char * gatelengthLine    = "Range gate length (m)";
  static char * nummeasureLine    = "Number of measurements";
  static char * numgatesm1Line    = "Number of gates in M1";
  static char * numgatesm2Line    = "Number of gates in M2";
  static char * pulsesraym1Line   = "Pulses/ray in M1";
  static char * pulsesraym2Line   = "Pulses/ray in M2";
  static char * numlagsLine       = "Number of Lags";
  static char * gatelengthptsLine = "Gate length (pts)";
  static char * focusrangem1Line  = "Focus range M1";
  static char * focusrangem2Line  = "Focus range M2";
  static char * starttimeLine     = "Start time";
  static char * resolutionLine    = "Resolution (m/s)";

  static char * numgatesLine      = "Number of gates";
  static char * pulsesrayLine     = "Pulses/ray";
  static char * integrationsLine  = "Integrations per loop";
  static char * scantypeLine      = "Scan type";
  static char * numraysLine       = "No. of rays in scan";
  static char * focusrangeLine    = "Focus range";
  static char * numrhisLine       = "No. of RHIs";
  
  /* Files for later instruments may include system ID*/
  static char * sysidLine         = "System ID";
 
  /* scantype2 is tilt */
  static char * scantype1         = "Stare";
  static char * scantype3         = "RHI";
  static char * scantype4         = "VAD";
  static char * scantype5         = "Wind";
  static char * scantype6         = "User";
  static char * scantype6new      = "stepped";


  char line[MAX_STRING_LENGTH];
  line[0] = '\0';
  header->scantype = 0;
  header->filetype = 0;
  
  /* Count the number of header lines. */
  int numHdrLines = 0;
  int myint = -1;
  while ( (myint != 0) && numHdrLines < NUMLINES){
    fgets(line, MAX_STRING_LENGTH, fp); /* line of data */
    sscanf(&line[0], "%d", &myint);

    /* check for '*' */
    p = &line[0];  
    if (strncmp(p,"*",1) == 0) {
      header->filetype = 1;      
      myint = 0;
    }

    numHdrLines++;
  }

  numHdrLines--;
  rewind(fp);

  if (header->filetype == 1)
    numHdrLines++;

  for ( i = 0; i < numHdrLines; i++ )
    {
      fgets(line, MAX_STRING_LENGTH, fp); /* line of data */
      if ( strncmp(line, datestampLine, strlen(datestampLine)) == 0 )
	{
	  p = &line[strlen(datestampLine)];

	  /* check for ':' */
	  if (strncmp(p,":",1) == 0) {
	    p +=1;
	    header->filetype = 1;

	  }
	  /* Does line contain tilt? */
	  /* Move pointer to _tilt.hpl or NULL if end of line. */
	  if  (strstr(p, "_tilt.hpl") != NULL)
	    p = strstr(p, "_tilt.hpl");
	  else if (strstr(p, ".hpl") != NULL)
	    /* Else move pointer to .hpl or NULL if end of line. */
	    p = strstr(p, ".hpl");
	  else
	    /* Else file is probably not a .hpl file, return */
	    return -1;

	  /* Move pointer back to beginning of timestamp. */
	  if (header->filetype)
	    p = p-6;
	  else
	    p = p-8;

	  if  (strstr(p, "_tilt.hpl") != NULL) {
	    sscanf(p, "%2hd%2hd%2hd_tilt.hpl", &header->day, &header->month, &header->year);
	    header->scantype = 2;
	  }
	  else {
            /* Check whether filename includes directory */
            p = &line[0];
            /* If directory present, first line looks like this */
            /* Filename:       C:\Lidar\Data\Proc\2010\201004\20100420\Stare_200410.hpl */
            /* Safer to reverse search from end of line and find directory (which should be a date) */

            if (strrchr(p,'\\') != NULL) {
              p = strrchr(p,'\\');
              p = p-6;
              sscanf(p, "%2hd%2hd%2hd", &header->year, &header->month, &header->day);
            }
            else {
              /* Fist line looks like this */
              /* Filename:       Stare_34_20140401_00.hpl */
              p = strrchr(p,'_');
              p = p-6;
              sscanf(p, "%2hd%2hd%2hd", &header->year, &header->month, &header->day);
            }
	  }    
	}
      else if ( strncmp(line, gatelengthLine, strlen(gatelengthLine)) == 0 )
	{
	  p = &line[strlen(gatelengthLine+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%f", &header->gate_length);
	}      
      
      else if ( strncmp(line, nummeasureLine, strlen(nummeasureLine)) == 0 )
	{
	  p = &line[strlen(nummeasureLine+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%hu", &header->number_measurements);
	}     
      
      else if ( strncmp(line, numgatesLine, strlen(numgatesLine)) == 0 && !(strncmp(line, numgatesm1Line, strlen(numgatesm1Line)) == 0) && !(strncmp(line, numgatesm2Line, strlen(numgatesm2Line)) == 0) )
        {
          p = &line[strlen(numgatesLine+1)];
          /* Move to next space or NULL if end of line. */
          p = strpbrk(p, " \t");
          sscanf(p, "%u", &header->number_gates_m1);
        } 

      else if ( strncmp(line, numgatesm1Line, strlen(numgatesm1Line)) == 0 )
	{
	  p = &line[strlen(numgatesm1Line+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%u", &header->number_gates_m1);
      }     

      else if ( strncmp(line, numgatesm2Line, strlen(numgatesm2Line)) == 0 )
	{
	  p = &line[strlen(numgatesm2Line+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%u", &header->number_gates_m2);
	}     
      
      else if ( strncmp(line, pulsesrayLine, strlen(pulsesrayLine)) == 0 && !(strncmp(line, pulsesraym1Line, strlen(pulsesraym1Line)) == 0)  && !(strncmp(line, pulsesraym2Line, strlen(pulsesraym2Line)) == 0))
	{
	  p = &line[strlen(pulsesrayLine+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%u", &header->number_pulses_m1);
	}     
                                                                                    
      else if ( strncmp(line, pulsesraym1Line, strlen(pulsesraym1Line)) == 0 )
        {
          p = &line[strlen(pulsesraym1Line+1)];
          /* Move to next space or NULL if end of line. */
          p = strpbrk(p, " \t");
          sscanf(p, "%u", &header->number_pulses_m1);
        }
                                   
      else if ( strncmp(line, pulsesraym2Line, strlen(pulsesraym2Line)) == 0 )
	{
	  p = &line[strlen(pulsesraym2Line+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%u", &header->number_pulses_m2);
      } 

     else if ( strncmp(line, numlagsLine, strlen(numlagsLine)) == 0 )
	{
	  p = &line[strlen(numlagsLine+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%hu", &header->number_lags);
      }      

      else if ( strncmp(line, gatelengthptsLine, strlen(gatelengthptsLine)) == 0 )
	{
	  p = &line[strlen(gatelengthptsLine+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%hu", &header->gate_length_pts);
	}
      
      else if ( strncmp(line, focusrangeLine, strlen(focusrangeLine)) == 0 && !( strncmp(line, focusrangem1Line, strlen(focusrangem1Line)) == 0) && !( strncmp(line, focusrangem2Line, strlen(focusrangem2Line)) == 0))
	{
	  p = &line[strlen(focusrangeLine+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%u", &header->focus_range_m1);
      }  

      else if ( strncmp(line, focusrangem1Line, strlen(focusrangem1Line)) == 0 )
	{
          p = &line[strlen(focusrangem1Line+1)];
          /* Move to next space or NULL if end of line. */
          p = strpbrk(p, " \t");
          sscanf(p, "%u", &header->focus_range_m1);
        }       
      
      else if ( strncmp(line, focusrangem2Line, strlen(focusrangem2Line)) == 0 )
	{
	  p = &line[strlen(focusrangem2Line+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%u", &header->focus_range_m2);
      }        

      else if ( strncmp(line, starttimeLine, strlen(starttimeLine)) == 0 )
	{
	  p = &line[strlen(starttimeLine+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%2hd:%2hd:%2hd", &header->hour, &header->minute, &header->second);
	}    
      
      else if ( strncmp(line, resolutionLine, strlen(resolutionLine)) == 0 )
	{
	  p = &line[strlen(resolutionLine+1)];
	  /* Move to next space or NULL if end of line. */
	  p = strpbrk(p, " \t");
	  sscanf(p, "%f", &header->v_resolution);
	}
      
      
      else if ( strncmp(line, integrationsLine, strlen(integrationsLine)) == 0 )
        {
          p = &line[strlen(integrationsLine+1)];
          /* Move to next space or NULL if end of line. */
          p = strpbrk(p, " \t");
          sscanf(p, "%hu", &header->num_integrations);
        }        
      
      else if ( strncmp(line, numrhisLine, strlen(numrhisLine)) == 0 )
        {
          p = &line[strlen(numrhisLine+1)];
          /* Move to next space or NULL if end of line. */
          p = strpbrk(p, " \t");
          sscanf(p, "%hu", &header->num_rhis_in_scan);
        }        
      
      else if ( strncmp(line, numraysLine, strlen(numraysLine)) == 0 )
        {
          p = &line[strlen(numraysLine+1)];
          /* Move to next space or NULL if end of line. */
          p = strpbrk(p, " \t");
          sscanf(p, "%hu", &header->num_rays);
        }

      else if ( strncmp(line, sysidLine, strlen(sysidLine)) == 0 )
        {
          p = &line[strlen(sysidLine+1)];
          /* Move to next space or NULL if end of line. */
          p = strpbrk(p, " \t");
          sscanf(p, "%hu", &header->system_id);
        }

      /* Remember that scantype2 is tilt */
      else if ( strncmp(line, scantypeLine, strlen(scantypeLine)) == 0 )
        {
	  if (strstr(line, scantype1) > 0)
	    header->scantype = 1;
	  else if (strstr(line, scantype3) > 0)
	    header->scantype = 3;
	  else if (strstr(line, scantype4) > 0)
	    header->scantype = 4;
	  else if (strstr(line, scantype5) > 0)
	    header->scantype = 5;
	  else if (strstr(line, scantype6) > 0)
	    header->scantype = 6;
          else if (strstr(line, scantype6new) > 0)
            header->scantype = 6;
	  else 
	    header->scantype = 0;
        }
    }


  return 0;
}

int
doppler_lidar_read_profile(FILE *fp, DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarTime *profiletime)
{

  char line[MAX_STRING_LENGTH];
  int i;
  int tmp;
  float tmptime, tmpelev, tmpazim, tmppitch, tmproll;
  char * p = NULL;
  line[0] = '\0';

  if (feof(fp))
    return -1;

  /* For new filetypes time is in first line before each ray */
  /* Version upgrade (late 2017) includes pitch and roll */
  if (header->filetype == 1) {

    /* first line contains time, azim, elev (maybe ptich and roll) */
    fgets(line, MAX_STRING_LENGTH, fp); /* line of data */	
    if (strlen(line) < 10)
      return -1;
	
    p = &line[0];
    if (sscanf(p,"%f %f %f ", &tmptime, &tmpazim, &tmpelev) == 0 )
      return SkipRay(fp);

    if (header->scantypereq == 1 && tmpelev < 84)
      return SkipRay(fp);

    profiletime->hour = floor(tmptime);
    profiletime->minute = floor((tmptime - floor(tmptime)) * 60);
    profiletime->second = (tmptime - ((float) profiletime->hour + (float) profiletime->minute / 60)) * 3600;
    profile->azimuth[0] = (float) tmpazim;
    profile->elevation[0] = tmpelev;

    /* Include pitch and roll if they exist in the file */
    if (sscanf(p,"%f %f %f %f %f ", &tmptime, &tmpazim, &tmpelev, &tmppitch, &tmproll) == 5) {
      header->is_pitch_and_roll = 1;
      profile->pitch = tmppitch;
      profile->roll = tmproll;
    }

    line[0] = '\0';
    /* Read data */
    for (i=0;i<header->number_gates_m1;i++) {
      fgets(line, MAX_STRING_LENGTH, fp); /* line of data */
      
      if (strlen(line) < 10)
	return -1;
	
      p = &line[0];
	
      if (sscanf(p,"%u %f %f %f", &tmp, &(profile->velocity_raw[i]), &(profile->signal[i]), &(profile->beta_raw[i])) == 0 )
	return SkipRay(fp);
    }
    line[0] = '\0';
  }

  /* otherwise time is in every line of every ray */
  else
    {

      /* Read data */
      for (i=0;i<header->number_gates_m1;i++) {
	fgets(line, MAX_STRING_LENGTH, fp); /* line of data */
	
	if (strlen(line) < 10)
	  return -1;
	
	p = &line[0];
	
	if (header->scantype==0) {
	  if (sscanf(p,"%u %f %f %f %2hd:%2hd:%2hd", &tmp, &(profile->velocity_raw[i]), &(profile->signal[i]), &(profile->beta_raw[i]), &(profiletime->hour), &(profiletime->minute), &(profiletime->second)) == 0 )
	    return SkipRay(fp);
	}
	else {
	  if (sscanf(p,"%u %f %f %f %2hd:%2hd:%2hd %f %f", &tmp, &(profile->velocity_raw[i]), &(profile->signal[i]), &(profile->beta_raw[i]), &(profiletime->hour), &(profiletime->minute), &(profiletime->second), &(profile->azimuth[i]), &(profile->elevation[i])) == 0 )
	    return SkipRay(fp);
	}
	/*check tmp to make sure that num_gates_m1 is correct for RHI! */
	
	
	line[0] = '\0';
	
      }
    }
  if ( p == NULL && tmp !=  header->number_gates_m1)
    return -1; 
  
  return 0;
}

int
doppler_lidar_read_raw_profile(FILE *fp, DopplerLidarRawData *rawprofile, DopplerLidarACFData *acfprofile, DopplerLidarHeader *header, int argc, char **argv, char *hpldir)
{

  if (fread(rawprofile, sizeof(double) * 6 * header->number_gates_m1 * header->number_raw_samples_gate,1, fp)) {
    int i;

    for (i=0; i < header->number_gates_m1 * header->number_lags * header->number_raw_samples_gate; i++) {
//fprintf(stdout,"%6.0f %6.0f ",rawprofile->acf_raw[i*2],rawprofile->acf_raw[i*2+1]);

      acfprofile->acf_real[i] = (float) rawprofile->acf_raw[i*2];
      acfprofile->acf_imag[i] = (float) rawprofile->acf_raw[i*2+1];
    }
//fprintf(stdout," \n \n ");

    /*    acfprofile->numrays = 1;*/
    return 0;
  }
  else
    if (feof(fp) && header->iarg < argc-1) {
      /* Open next file if it exists - hopefully list has been sorted! */
      fclose(fp);
      ++header->iarg;

      if ((fp = fopen(argv[header->iarg], "r")) == NULL) {
        fprintf(stderr,"can't open input file %s\n",argv[header->iarg]);
        return 1;
        /*exit(1);*/
      }

      if (doppler_lidar_open_hpl(header, argc, argv, hpldir)) {
        fprintf(stderr,"Error reading header for %s\n", argv[header->iarg]);
        exit(1);
      }
      return doppler_lidar_read_raw_profile(fp, rawprofile, acfprofile, header, argc, argv, hpldir);
    }
    else
      return -1;
}

int
doppler_lidar_single_raw_profile(FILE *fp, DopplerLidarRawData *rawprofile, DopplerLidarACFData *acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec, int argc, char **argv, char *hpldir)
{
  if (doppler_lidar_read_raw_profile(fp, rawprofile, acfprofile, header, argc, argv, hpldir) == 0) {
    doppler_lidar_calculate_pulsepair_v(acfprofile, header, spec);

    if (spec->cleanrawrayflag == 1)
      doppler_lidar_simple_clean_raw_ray(acfprofile, header, spec);

    return 0;
  }
  else
    return -1;
}

int
doppler_lidar_average_raw_profile(FILE *fp, DopplerLidarRawData *rawprofile, DopplerLidarACFData *rawacfprofile, DopplerLidarACFData *acfprofile, DopplerLidarHeader *header, DopplerLidarSpec *spec, int argc, char **argv, char *hpldir)
{
  int i,j;

  /* First set all summations to zero */
  for (j=0; j < header->number_gates_m1 * header->number_lags * header->number_raw_samples_gate; j++) {
    acfprofile->acf_real[j] = 0;
    acfprofile->acf_imag[j] = 0;
  }

  for (j=0; j < header->number_gates_m1 * header->number_raw_samples_gate; j++) {
    acfprofile->signal[j]     = 0;
    acfprofile->v_pp[j]       = 0;
    acfprofile->sigma_v_pp[j] = 0;
    acfprofile->skew_v_pp[j]  = 0;
    acfprofile->kurt_v_pp[j]  = 0;
    acfprofile->numrays[j]    = 0;
  }

  /* Now perform summations for meteorological targets only and keep a record of how many summed */
  for (i =0; i < spec->numrays; i++ ) {
    if (doppler_lidar_read_raw_profile(fp, rawprofile, rawacfprofile, header, argc, argv, hpldir) == 0) {
      doppler_lidar_calculate_pulsepair_v(rawacfprofile, header, spec);
      doppler_lidar_simple_clean_raw_ray(rawacfprofile, header, spec);

      for (j=0; j < header->number_gates_m1 * header->number_lags * header->number_raw_samples_gate; j++) {
        if (rawacfprofile->signal[j] > 0) {
          acfprofile->acf_real[j] += rawacfprofile->acf_real[j];
          acfprofile->acf_imag[j] += rawacfprofile->acf_imag[j];
        }
      }
      for (j=0; j < header->number_gates_m1 * header->number_raw_samples_gate; j++) {
        if (rawacfprofile->signal[j] > 0) {
          acfprofile->v_pp[j] += rawacfprofile->v_pp[j];
          acfprofile->sigma_v_pp[j] += (rawacfprofile->v_pp[j] * rawacfprofile->v_pp[j]);
          acfprofile->skew_v_pp[j]  += (rawacfprofile->v_pp[j] * rawacfprofile->v_pp[j] * rawacfprofile->v_pp[j]);
          acfprofile->kurt_v_pp[j]  += (rawacfprofile->v_pp[j] * rawacfprofile->v_pp[j] * rawacfprofile->v_pp[j] * rawacfprofile->v_pp[j]);
          acfprofile->numrays[j]++;
          acfprofile->signal[j] += rawacfprofile->signal[j];
        }
      }
    }
    else
      return -1;
  }

  /* Use this value for calculating epsilon - only sensible for less than about 30 samples*/
  if (spec->calculate_epsilon)
    doppler_lidar_calculate_pulsepair_v(acfprofile, header, spec);
  else {
    for (j=0; j < header->number_gates_m1 * header->number_raw_samples_gate; j++) {
      acfprofile->v_pp[j] = acfprofile->v_pp[j] / acfprofile->numrays[j];
    }
  }
  /* sigma_v_pp, skew_v_pp, kurt_v_pp currently point to the 2nd, 3rd, 4th moments */
  /* Remember to calculate higher moments first before they are overwritten! */
  /* These moments are only calculated from non-noise targets */

  for (j=0; j < header->number_gates_m1 * header->number_raw_samples_gate; j++) {
    if (acfprofile->numrays[j] > spec->numrays / 4) {
      acfprofile->kurt_v_pp[j] = ((acfprofile->kurt_v_pp[j] / acfprofile->numrays[j] -
                                   (4 * acfprofile->v_pp[j] * acfprofile->skew_v_pp[j] / acfprofile->numrays[j]) +
                                   (6 * acfprofile->v_pp[j] * acfprofile->v_pp[j] * acfprofile->sigma_v_pp[j] / acfprofile->numrays[j]) -
                                   (3 * acfprofile->v_pp[j] * acfprofile->v_pp[j] * acfprofile->v_pp[j] * acfprofile->v_pp[j])) /
                                  pow(((acfprofile->sigma_v_pp[j] / acfprofile->numrays[j])
                                       - (acfprofile->v_pp[j] * acfprofile->v_pp[j])),2))-3;

      acfprofile->skew_v_pp[j] = (acfprofile->skew_v_pp[j] / acfprofile->numrays[j] -
                                  (3 * acfprofile->v_pp[j] * acfprofile->sigma_v_pp[j] / acfprofile->numrays[j]) +
                                  (2 * acfprofile->v_pp[j] * acfprofile->v_pp[j] * acfprofile->v_pp[j])) /
        pow(((acfprofile->sigma_v_pp[j] / acfprofile->numrays[j])
             - (acfprofile->v_pp[j] * acfprofile->v_pp[j])),1.5);

      acfprofile->sigma_v_pp[j] = sqrt((acfprofile->sigma_v_pp[j] / acfprofile->numrays[j])
                                       - (acfprofile->v_pp[j] * acfprofile->v_pp[j]));

      acfprofile->signal[j] = acfprofile->signal[j] / acfprofile->numrays[j];
    }
    else {
      acfprofile->signal[j]     = spec->missing_value;
      acfprofile->v_pp[j]       = spec->missing_value;
      acfprofile->sigma_v_pp[j] = spec->missing_value;
      acfprofile->skew_v_pp[j]  = spec->missing_value;
      acfprofile->kurt_v_pp[j]  = spec->missing_value;
      /*      acfprofile->numrays[j]    = spec->missing_value;*/
    }
  }
  return 0;
}



int doppler_lidar_simple_clean_ray(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value)
{
  int i;

  for (i=0; i < header->number_gates_m1; i++)
    {
      /* Remove data below threshold in spec */
      if (profile->signal[i] < spec->noise_threshold || profile->beta_raw[i] < 1e-8 )
	{
	  profile->beta[i] = missing_value;
	  profile->velocity[i] = missing_value;
	}
      /* otherwise copy data to clean variables */
      else
	{
	  profile->beta[i] = profile->beta_raw[i] * spec->calfactor;
	  profile->velocity[i] = profile->velocity_raw[i];
	}
    }
  
  /* remove remaining isolated pixels, whatever their value... */
  for (i = 1; i < (header->number_gates_m1) - 1; ++i) {
    if (profile->beta[i-1] < 1e-8 && profile->beta[i+1] < 1e-8) {
      profile->beta[i] = missing_value;
      profile->velocity[i] = missing_value;
    }
  }
  
  return 0;
}

int doppler_lidar_clean_ray(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value)
{
  int i, j, n = 5;
  float sumsq = 0;

  profile->mean_noise = 0;
  /*  Normalise the profile from the mean of the noise from the top  
      n gates. We assume it's cloud free - could be a mistake? */
  for (i=header->number_gates_m1-n; i < header->number_gates_m1; i++)
    profile->mean_noise += profile->signal[i];
  
  profile->mean_noise /= (float) n;
  
  for (i=0; i < header->number_gates_m1; i++)
    {
      profile->beta[i] = profile->beta_raw[i];
      profile->velocity[i] = profile->velocity_raw[i];
    }

  profile->std_noise = 0;
  /* Now calculate the standard deviation of the noise from the top
     n gates.*/
  /*     for (i=num_gates-n;i < num_gates;i++)
  sumsq += profile->beta[i] * profile->beta[i])
  ;*/
  for (i=header->number_gates_m1-n; i < header->number_gates_m1; i++)
    sumsq += profile->beta[i] * profile->beta[i];

  profile->std_noise = sqrt( sumsq / (float) n );

  /*  thresh = ((*stdn) * noisefactor);*/

  /* Remove data below threshold */

  for (i=0; i < header->number_gates_m1; i++)
    {
      if (profile->beta[i] < spec->noise_threshold )
	{
	  profile->beta[i] = 0;
	  profile->velocity[i] = missing_value;
	}

      if (profile->beta[i] < 1e-8 )
	{
	  profile->beta[i] = missing_value;
	  profile->velocity[i] = missing_value;
	}
    }

  /* remove remaining isolated pixels, whatever their value... */
  for (j = 1; j < header->number_gates_m1-1; ++j) {
    if (profile->beta[j-1] < 1e-8 && profile->beta[j+1] < 1e-8) {
      profile->beta[j] = missing_value;
      profile->velocity[j] = missing_value;
    }
  }

  return 0;
}

int doppler_lidar_clean_velocity(DopplerLidarData *profile, DopplerLidarHeader *header, DopplerLidarSpec *spec, float missing_value)
{
  int i; 
  
  /*      printf(" %f %f ",spec->cleanv_threshold, profile->beta[15]);*/
  for (i = 1; i < header->number_gates_m1-1; ++i) {
    if (profile->beta[i] < spec->cleanv_threshold) {
      profile->velocity[i] = missing_value;
      /*      printf(" %f %f\n",profile->beta[i], profile->velocity[i]);*/
    }
  }
  return 0;
}

/* hnu =  h.*c./1.55e-6 wavelength */


