/* 
Generate wind vector, wind speed/direction from Doppler Lidar netCDF data

History
-------
2011-11-10 eoc started

Author - Ewan O'Connor
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netcdf.h>
#include <ctype.h>
#include <math.h> 

#include "dopplerlidar.h"
#include "nctools.h"

#define MAX_STRING_LENGTH 512
#define MAX_LINE_LEN 2048

//#define NCFILEREAD 0
//#define NCFILEWRITE 1

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s (%s)\n", \
	    __LINE__, nc_strerror(ncstatus), s); exit(-1); \
} 

static
char *
seek_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  if (nc_inq_atttype(ncid, varid, name, &type) != NC_NOERR
      || type != NC_CHAR) {
    return NULL;
  }

  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = malloc((length+1) * sizeof(char));
  if (!value) {
    fprintf(stderr, "Out of memory\n");
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}


void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.nc  [file2.nc ...]\n"
	  "Options:\n"
	  "     -convention d        Use wind vector azimuth (0) or meteorological convention (1)\n"
	  "                            Meteorological convention is the default and is defined as\n"
	  "                            the true wind vector azimuth +180 degrees\n"
	  " -correctazimuth f        add f degrees to azimuth in file before calculating wind vectors\n"
	  " -noisethreshold d        use signal noise threshold to clean data (default = 1.018)\n"
	  "           -site s        override site name (default is to use location given in file)\n"
	  "         -prefix dir      write the files in specified directory\n"
          "    -concatenate          concatenate multiple scans into daily file\n"
          "  -noconcatenate          one file per scan, don't concatenate (default)\n"
	  "           -user s        name and email address of the user\n"
	  "        -verbose          write more\n",
	  argv[0]);
  exit(0);
}

int main(int argc, char **argv){
  char *filename = '\0';
  char *prefix = ".";
  DopplerLidarData profile[3];
  DopplerLidarData *DBSprofile = profile;

  DopplerLidarData ppiprofile[MAX_BEAMS];
  DopplerLidarData *VADprofile = ppiprofile;

  DopplerLidarWindData windprofile;

  char *location = NULL;
  char *llocation = NULL;
  char *user = NULL;
  DopplerLidarPlotRange plot_range = {{1e-7,    1e-4},
                                      { -2.0,    2.0},
                                      {  0.001,  1.0},
                                      {  0.001, 10.0},
                                      {  0.0,    0.5},
                                      { -1.0,    1.0},
                                      {  0.001,  1.0},
                                      {  0.001, 10.0},
                                      {  0.001,  1.0},
                                      { -2.0,    2.0},
                                      { -2.0,    2.0},
                                      {  0.0,    1.0},
                                      {-15.0,   15.0},
                                      {-15.0,   15.0},
                                      { -2.0,    2.0},
                                      {  0.0,   15.0},
                                      {  0.0,  360.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    1.0},
                                      {  0.0,    2.0},
                                      {  0.0,   30.0}};

  DopplerLidarTime time  = {0, 0, 0, 0, 0, 0};
  DopplerLidarTime *current_time = &time; /* initialise current */

  int   iarg = 0; 
  char  ncfilename[MAX_STRING_LENGTH];
  int   nrays_DBS = 3;

  int   ncid[2];                            /* netCDF ID */
  int   time_dim[2], range_dim[2];          /* dimension IDs */
  int   timeid[2], rangeid[2], azelid_read[4], azelid_write[2];     /* variable ID's */
  /*  int   signalid;*/ /* calculate error*/
  int   lag_dim, lagid; /* Dummy IDs */
  int   nlags = 0; /* dummy var*/

  int   heightid;
  int   uwindid, vwindid, wwindid, speedid, dirid;
  int   uerrid, verrid, werrid, speederrid, direrrid;
  int   meansnrid;
  int   bs_dimids[2];                       /* variable shape */
  size_t   nrays, ngates;
  float missing_value=-999;
  float convention = 180; /*Meteorological convention - direction wind is blowing from */

  /* flags */
  char  verbose = 0;
  char  location_override = 0;
  char  outputraw = 0;
  char  iswindscan = 0;
  char  isVAD = 0; /* Otherwise DBS */
  char  isconcatenate = 0;

  /* Set processing defaults */
  windprofile.applyazimuthalcorrection = 0;
  windprofile.azimuthal_correction = 0;
  windprofile.noise_threshold = 1.018;


  /* Process command-line arguments */
  if (argc < 2)
    {
      Usage(argv);
      exit(0);
    }

  while (++iarg < argc) {

    if (argv[iarg][0] != '-') {
      /* Found a lidar filename.*/
      filename = argv[iarg];
      break;
    }
    else if (strcmp(argv[iarg], "-convention") == 0) 
      {
	if (argc > iarg+1) 
	  {
	    convention = atoi(argv[++iarg]) * 180;
	  }
	else 
	  {
	    fprintf(stderr,"Convention not specified");
	    exit(1);
	  }
      }
    else if (strcmp(argv[iarg], "-correctazimuth") == 0) {
      if (argc > iarg+1) {
	windprofile.azimuthal_correction = atof(argv[++iarg]);
        if (windprofile.azimuthal_correction >= 360 || windprofile.azimuthal_correction < 0) {
          fprintf(stderr,"Azimuthal correction must be > 0 and < 360 degrees");
          exit(1);
        }
	windprofile.applyazimuthalcorrection = 1;
      }
      else {
	fprintf(stderr,"Azimuthal correction not specified");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg],"-noisethreshold") == 0) {
      if (argc > iarg+1) {
	windprofile.noise_threshold = atof(argv[++iarg]); 
      }
      else {
	fprintf(stderr,"Noise threshold not specified");
	exit(1);
      }
    }

    else if (strcmp(argv[iarg],"-location") == 0) {
      if (argc > iarg+1) {
	location = argv[++iarg];
	location_override = 1;
      }
      else {
	fprintf(stderr,"Location not specified");
      }
    }
    else if (strcmp(argv[iarg],"-prefix") == 0) {
      if (argc > iarg+1) {
	prefix = argv[++iarg];
      }
      else {
	fprintf(stderr,"Prefix not specified");
	exit(1);
      }
    } 
    else if (strcmp(argv[iarg], "-concatenate") == 0) {
      isconcatenate = 1;
    }
    else if (strcmp(argv[iarg], "-noconcatenate") == 0) {
      isconcatenate = 0;
    }
    else if (strcmp(argv[iarg], "-verbose") == 0) {
      verbose = 1;
    }   
    else if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	user = argv[++iarg];
      }
      else {
	fprintf(stderr,"Error: user not specified\n");
	exit(1);
      }
    }
    else {
      fprintf(stderr, "Error: `%s' not understood or not accompanied "
	      "by enough subsequent arguments\n", argv[iarg]);
      Usage(argv);
      exit(1);
    }
  }

  /* cycle through filenames in argv list */
  while (iarg < argc ) 
    {
      size_t tmp_index;
      size_t start[2], count[2];
      float tmp_float;
//      short j;
      int j, nelev, naz, nrays_per_scan = 0;

      float start_time = 0, end_time = 0;

      /* counters */
      int index = 0; /* Index of the current profile in the input file */

      /* Open file and get dimension sizes */
      handle_error(nc_open(argv[iarg], NC_NOWRITE, &ncid[NCFILEREAD]));
      handle_error(nc_inq_dimid(ncid[NCFILEREAD], "time", &time_dim[NCFILEREAD]));
      handle_error(nc_inq_dimlen(ncid[NCFILEREAD], time_dim[NCFILEREAD], &nrays));
      handle_error(nc_inq_dimid(ncid[NCFILEREAD], "range", &range_dim[NCFILEREAD]));
      handle_error(nc_inq_dimlen(ncid[NCFILEREAD], range_dim[NCFILEREAD], &ngates));
      
      /* Check file type, we only want to read wind file(s) */
      char *scantype = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "scantype");

      handle_error(nc_inq_varid(ncid[NCFILEREAD], "elevation", &azelid_read[0]));
      handle_error(nc_inq_varid(ncid[NCFILEREAD], "azimuth", &azelid_read[1]));
      handle_error(nc_inq_varndims(ncid[NCFILEREAD], azelid_read[0], &nelev)); 
      handle_error(nc_inq_varndims(ncid[NCFILEREAD], azelid_read[1], &naz));     


      /* Do we have a suitable scantype? */
      /* Suitable scantypes include:     */
      /*   WIND                          */
      /*   VAD                           */
      /*   USER (Need to check!)         */
      /*                                 */
      /* but not FIX                     */
      if  (strncmp(scantype,"WIND",4) == 0)
        {
          /* Should also check whether number of rays is divisible by three! */
          iswindscan = 1;
          isVAD = 0;
          nrays_per_scan = nrays_DBS;
        }
      else if (strncmp(scantype,"VAD",3) == 0) 
        {
          iswindscan = 1;
          isVAD = 1;
          nrays_per_scan = (int) nrays;
        }
      else if (strncmp(scantype,"USER",4) == 0)
        {
          /* Check elevation doesn't change (much !) */

/* some code checking elevations here*/          

          iswindscan = 1;
          isVAD = 1;
          nrays_per_scan = (int) nrays;

        }

      if (iswindscan)
	{
	  /* Populate time struct from ncfile*/
          doppler_lidar_populate_timestruct_nc(ncid[NCFILEREAD], timeid[NCFILEREAD], current_time);


	  if (!location_override)
	    {
	      location = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "location");
	      if (location == NULL)
		{
		  char *name = seek_char_attribute(ncid[NCFILEREAD], NC_GLOBAL, "source");
		  /* If first character is a space! */
		  if (strspn(name," "))
		    name++;
		  
		  location = strdup(name);
		  location[strcspn(name," ")]='\0';
		}
	    }

	  char *ch = strdup(location);
	  llocation = ch;
	  while (*ch) {
              if ( *ch == ' ' )
                *ch = '-';
              else 
	        *ch = tolower(*ch);
	    ++ch;
	  }


	  /* Open netCDF file for writing */		
          if ( isVAD ==0 && isconcatenate == 1 )
            sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd_%s_halo-doppler-lidar-wind.nc", prefix, current_time->year, current_time->month, current_time->day, llocation);	
          else if ( isVAD == 1 && isconcatenate == 1 )
            sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd_%s_halo-doppler-lidar-vad.nc", prefix, current_time->year, current_time->month, current_time->day, llocation);     
          else if ( isVAD == 0 && isconcatenate == 0 )
            sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd_%2.2hd%2.2hd%2.2hd_%s_halo-doppler-lidar-wind.nc", prefix, current_time->year, current_time->month, current_time->day, current_time->hour, current_time->minute, current_time->second, llocation);
          else if ( isVAD == 1 && isconcatenate == 0 )
            sprintf(ncfilename,"%s/%4hd%2.2hd%2.2hd_%2.2hd%2.2hd%2.2hd_%s_halo-doppler-lidar-vad.nc", prefix, current_time->year, current_time->month, current_time->day, current_time->hour, current_time->minute, current_time->second, llocation);
      
	  ncfilename[MAX_STRING_LENGTH-1] = '\0';      
	    
	  handle_error(nc_create(ncfilename, NC_CLOBBER, &ncid[NCFILEWRITE]));
	  fprintf(stderr,"Writing %s\n", ncfilename);

	  /* Copy global attributes */
	  doppler_lidar_copy_global_attributes(ncid, current_time, user, location);
	  /* Redefine scantype */
	  handle_error(nc_put_att_text(ncid[NCFILEWRITE], NC_GLOBAL, "scantype", 3, "FIX"));

	  /* Add a `command_line' attribute */
	  handle_error(nct_add_command_line(ncid[NCFILEWRITE], argc, argv));
	    
	  /* Copy some scalar variable attributes */
	  doppler_lidar_copy_scalar_varatts(ncid);

	  /* Add dummy elevation variable (no azimuth) attribute */
	  doppler_lidar_write_az_el_attributes(ncid[NCFILEWRITE], bs_dimids, azelid_write, 1, 0, 0);

	  /* Define dimensions and add coordinate variables */
	  handle_error(nc_inq_varid(ncid[NCFILEREAD], "time", &timeid[NCFILEREAD]));
	  doppler_lidar_write_coordvar_attributes(ncid[NCFILEWRITE], bs_dimids, &time_dim[NCFILEWRITE], &timeid[NCFILEWRITE], &range_dim[NCFILEWRITE], &rangeid[NCFILEWRITE], ngates, &lag_dim, &lagid, nlags, outputraw);

          /* Add height variable */
	  doppler_lidar_write_height_attributes(ncid[NCFILEWRITE], &range_dim[NCFILEWRITE], &heightid);

	  /* Add wind variables */ 
	  doppler_lidar_write_windvar_attributes(ncid[NCFILEWRITE], bs_dimids, &uwindid, &vwindid, &wwindid, &speedid, &dirid, convention, windprofile.azimuthal_correction, missing_value, &plot_range);

          /* Add wind error variables */
          doppler_lidar_write_winderrorvar_attributes(ncid[NCFILEWRITE], bs_dimids, &uerrid, &verrid, &werrid, &speederrid, &direrrid, missing_value, &plot_range);

          /* Add mean SNR variable */
          doppler_lidar_write_snr_attributes(ncid[NCFILEWRITE], bs_dimids, &meansnrid, missing_value);


	  /* End define mode */
	  handle_error(nc_enddef(ncid[NCFILEWRITE]));
	    
	  /* Add dummy elevation variable */
	  tmp_float = 90;
          tmp_index = 0;
          handle_error(nc_get_var1_float(ncid[NCFILEREAD], azelid_read[0], &tmp_index, &tmp_float));
	  handle_error(nc_put_var_float(ncid[NCFILEWRITE], azelid_write[0], &tmp_float));  

	  /* Copy range variable and populate range in struct */
	  doppler_lidar_copy_range_var_wind(ncid, &windprofile, rangeid, ngates);

//printf("%d %d %d %d\n",index, nrays_per_scan,nrays,(index + 1) * nrays_per_scan);

	  /* Read and process variables n profiles per scan at a time */
	  while (((index + 1) * nrays_per_scan) <= nrays)
	    {
	      /* Read new profiles of data*/
	      for (j=0; j < nrays_per_scan; j++) 
		{	      
		  tmp_index = (index * nrays_per_scan) + j;
		  /* Get time of profile */
		  handle_error(nc_get_var1_float(ncid[NCFILEREAD], timeid[NCFILEREAD], &tmp_index, &end_time));

		  if (j == 0) 
		    {
		      start_time = end_time; /* Set start time here */
		    }

                  if (!isVAD)
                    doppler_lidar_copy_profile(ncid[NCFILEREAD],&DBSprofile[j], ngates, (index*nrays_per_scan)+j);
                  else 
                    doppler_lidar_copy_profile(ncid[NCFILEREAD],&VADprofile[j], ngates, (index*nrays_per_scan)+j);


		}


              if (!isVAD)
  	        doppler_lidar_wind_profile(DBSprofile, &windprofile, ngates, convention, missing_value);
              else
                doppler_lidar_vad_profile_TL(VADprofile, &windprofile, ngates, nrays_per_scan, convention, missing_value);


/* full calculation */
/*                doppler_lidar_vad_profile_EP(VADprofile, &windprofile, ngates, nrays_per_scan, convention, missing_value); */


              /* write vector values into netCDF variables */

	      /* Write height variable once - could change this if processing multiple ppi scan elevations at once  */
              /* No difference for single scans */
              handle_error(nc_put_var_float(ncid[NCFILEWRITE], heightid, windprofile.height));

	      /* Write new profile of data */  
              /* Check for pathological case - scan starts on one day and finishes on next */
              /* Need to fix this by moving to unix_time! */
              if ( start_time > end_time)
                end_time = end_time + 24;
	      tmp_float = start_time + (end_time - start_time) / 2;

	      tmp_index = index;
	      handle_error(nc_put_var1_float(ncid[NCFILEWRITE], timeid[NCFILEWRITE], &tmp_index, &tmp_float));

	      start[0] = index;
	      start[1] = 0;
	      count[0] = 1;
	      count[1] = ngates;
	     
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], uwindid, start,
					     count, windprofile.uwind));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], vwindid, start,
					     count, windprofile.vwind));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], wwindid, start,
					     count, windprofile.wwind));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], speedid, start,
					     count, windprofile.speed));
	      handle_error(nc_put_vara_float(ncid[NCFILEWRITE], dirid, start,
					     count, windprofile.direction));

              handle_error(nc_put_vara_float(ncid[NCFILEWRITE], uerrid, start,
                                             count, windprofile.uwind_error));
              handle_error(nc_put_vara_float(ncid[NCFILEWRITE], verrid, start,
                                             count, windprofile.vwind_error));
              handle_error(nc_put_vara_float(ncid[NCFILEWRITE], werrid, start,
                                             count, windprofile.wwind_error));
              handle_error(nc_put_vara_float(ncid[NCFILEWRITE], speederrid, start,
                                             count, windprofile.speed_error));
              handle_error(nc_put_vara_float(ncid[NCFILEWRITE], direrrid, start,
                                             count, windprofile.direction_error));

              handle_error(nc_put_vara_float(ncid[NCFILEWRITE], meansnrid, start,
                                             count, windprofile.mean_snr));

	      
	      ++index;  
	    }
	  /* Close netCDF file */
	  handle_error(nc_close(ncid[NCFILEWRITE]));
	}
      /* Close netCDF file */
      handle_error(nc_close(ncid[NCFILEREAD]));
      ++iarg;
    }
  exit(0); 
}
