/* doppler_write.c - functions for writing doppler lidar data in netCDF */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netcdf.h>
#include <ctype.h>
#include <math.h> 

#include "dopplerlidar.h"
#include "nctools.h"

#define MAX_STRING_LENGTH 512
#define NUMLINES 20


int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
            nc_strerror(ncstatus)); \
    exit(-1); \
}

static
void
add_string_attribute(int ncid, int fieldid, char *attname, char *value)
{
  if (value) {
    /* The +1 ensures that \0 is included at the end of each attribute
       in the netcdf file - is this necessary? */
    handle_error(nc_put_att_text(ncid, fieldid, attname,
				 strlen(value), value));
  }
}

static
void
add_attributes(int ncid, int fieldid, char *units, char *units_html, char *long_name, char *standard_name)
{
  if (units) {
    handle_error(nc_put_att_text(ncid, fieldid, "units",
				 strlen(units), units));
  }
  if (units_html) {
    handle_error(nc_put_att_text(ncid, fieldid, "units_html",
				 strlen(units_html), units_html));
  }
  if (long_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "long_name",
				 strlen(long_name), long_name));
  }
  if (standard_name) {
    handle_error(nc_put_att_text(ncid, fieldid, "standard_name",
				 strlen(standard_name), standard_name));
  }

}

static
void
add_missing_value(int ncid, int fieldid, float value)
{
  handle_error(nc_put_att_float(ncid, fieldid,
				"missing_value",
				NC_FLOAT, 1, &value));

  handle_error(nc_put_att_float(ncid, fieldid,
				"_FillValue",
				NC_FLOAT, 1, &value));
}

static
void
add_missing_value_short(int ncid, int fieldid, short value)
{
  handle_error(nc_put_att_short(ncid, fieldid,
				"missing_value",
				NC_SHORT, 1, &value));
  handle_error(nc_put_att_short(ncid, fieldid,
				"_FillValue",
				NC_SHORT, 1, &value));
}

void
doppler_lidar_write_global_attributes(int ncid, DopplerLidarTime *current_time, char *location, char *user, char *lscantype, char *institution, short int system_id)
{
  
  char *tmp_name;
  char tmp_string[MAX_STRING_LENGTH];
  
  /* Add some global attributes */
  tmp_name = "Halo Photonics pulsed Doppler lidar";

  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "system", strlen(tmp_name), tmp_name));

  if (institution)
    snprintf(tmp_string, MAX_STRING_LENGTH, "%s 1.5 micron Doppler Lidar operating at %s", institution, location);
  else
    snprintf(tmp_string, MAX_STRING_LENGTH, "%s 1.5 micron Doppler Lidar", location);
  tmp_string[MAX_STRING_LENGTH-1] = '\0';	

  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "location", strlen(location), location));
  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "source", strlen(tmp_string), tmp_string));
  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "title", strlen(tmp_string), tmp_string));
  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "day", NC_SHORT, 1, &current_time->day));
  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "month", NC_SHORT, 1, &current_time->month));
  handle_error(nc_put_att_short(ncid, NC_GLOBAL, "year", NC_SHORT, 1, &current_time->year));
  handle_error(nc_put_att_text(ncid, NC_GLOBAL, "scantype", strlen(lscantype), lscantype));

  if (system_id > 0)
    handle_error(nc_put_att_short(ncid, NC_GLOBAL, "systemID", NC_SHORT, 1, &system_id));

  /* Add a `history' attribute */
  handle_error(nct_add_history(ncid, "NetCDF generated from original data", user));
}

void
doppler_lidar_write_instrument_attributes(int ncid, int bs_dimids[3], int *numpulsesm1id, int *numpulsesm2id, int *numlagsid, int *numptsid, int *lensid, int *divid, int *focusid, int *energyid, int *waveid, int *prfid, int *pulselengthid, int *noiseid, int *bandid)
{
  
  handle_error(nc_def_var(ncid, "num_pulses_m1", NC_INT, 0, bs_dimids, numpulsesm1id));
  add_attributes(ncid, (*numpulsesm1id), " ", NULL, "Number of pulses per ray in mode 1", NULL);
  
  handle_error(nc_def_var(ncid, "num_pulses_m2", NC_INT, 0, bs_dimids, numpulsesm2id));
  add_attributes(ncid, (*numpulsesm2id), " ", NULL, "Number of pulses per ray in mode 2", NULL);
  
  handle_error(nc_def_var(ncid, "num_lags", NC_SHORT, 0, bs_dimids, numlagsid));
  add_attributes(ncid, (*numlagsid), " ", NULL, "Number of lags used in evaluating the complex autocorrelation function", NULL);  

  handle_error(nc_def_var(ncid, "num_samples_gate", NC_SHORT, 0, bs_dimids, numptsid));
  add_attributes(ncid, (*numptsid), " ", NULL, "Number of samples (points) per gate", NULL);
 
  handle_error(nc_def_var(ncid, "lens_diameter", NC_FLOAT, 0, bs_dimids, lensid));
  add_attributes(ncid, (*lensid), "m", NULL, "diameter of lens", NULL);
  
  handle_error(nc_def_var(ncid, "divergence", NC_FLOAT, 0, bs_dimids, divid));
  add_attributes(ncid, (*divid), "rad", NULL, "divergence of beam", NULL);
  
  handle_error(nc_def_var(ncid, "focus", NC_SHORT, 0, bs_dimids, focusid));
  add_attributes(ncid, (*focusid), "m", NULL, "Focus range", NULL);
  
  handle_error(nc_def_var(ncid, "energy", NC_FLOAT, 0, bs_dimids, energyid));
  add_attributes(ncid, (*energyid), "J", NULL, "laser energy", NULL);
  
  handle_error(nc_def_var(ncid, "wavelength", NC_FLOAT, 0, bs_dimids, waveid));
  add_attributes(ncid, (*waveid), "m", NULL, "laser wavelength", NULL);
  
  handle_error(nc_def_var(ncid, "prf", NC_FLOAT, 0, bs_dimids, prfid));
  add_attributes(ncid, (*prfid), "Hz", NULL, "Pulse repetition frequency", NULL);
  
  handle_error(nc_def_var(ncid, "pulse_length", NC_FLOAT, 0, bs_dimids, pulselengthid));
  add_attributes(ncid, (*pulselengthid), "s", NULL, "Pulse length", NULL);

  handle_error(nc_def_var(ncid, "bandwidth", NC_FLOAT, 0, bs_dimids, bandid));
  add_attributes(ncid, (*bandid), "Hz", NULL, "Bandwidth", NULL);
  
  handle_error(nc_def_var(ncid, "noise_threshold", NC_FLOAT, 0, bs_dimids, noiseid));
  add_attributes(ncid, (*noiseid), " ", NULL, "noise threshold, in arbitrary units, used to remove background noise", NULL);

}

void
doppler_lidar_write_location_attributes(int ncid, int bs_dimids[3], int *latid, int *lonid, int *altid)
{
  handle_error(nc_def_var(ncid, "latitude", NC_FLOAT, 0, bs_dimids, latid));
    add_attributes(ncid, (*latid), "degrees_north", NULL, "Latitude of lidar", "latitude");
  
  handle_error(nc_def_var(ncid, "longitude", NC_FLOAT, 0, bs_dimids, lonid));
  add_attributes(ncid, (*lonid), "degrees_east", NULL, "Longitude of lidar", "longitude");
  
  handle_error(nc_def_var(ncid, "altitude", NC_FLOAT, 0, bs_dimids, altid));
  add_attributes(ncid, (*altid), "m", NULL,
		 "Height of instrument above mean sea level", NULL);
}

void
doppler_lidar_write_coordvar_attributes(int ncid, int bs_dimids[3], int *time_dim, int *timeid, int *range_dim, int *rangeid, int ngates, int *lag_dim, int *lagid, int nlags, char outputraw) 
{
  /* define dimensions */
  handle_error(nc_def_dim(ncid, "time", NC_UNLIMITED, time_dim));
  handle_error(nc_def_dim(ncid, "range", ngates, range_dim));

  /* Add time coordinate variable */ 
  bs_dimids[0] = (*time_dim);
  handle_error(nc_def_var(ncid, "time", NC_FLOAT, 1, bs_dimids, timeid));
  add_attributes(ncid, *timeid, "hours UTC", NULL, "Decimal hours UTC", NULL);
  add_string_attribute(ncid, *timeid, "axis", "T");
  
  /* Add range coordinate variable */
  bs_dimids[0] = (*range_dim);
  handle_error(nc_def_var(ncid, "range", NC_FLOAT, 1, bs_dimids, rangeid));
  add_attributes(ncid, *rangeid, "m", NULL, "Range from lidar", NULL);
  add_string_attribute(ncid, *rangeid, "axis", "Z");
  nct_append_attribute(ncid, *rangeid, "comments", "This variable is the radial distance from the lidar instrument",""); 

  if (outputraw) {
    /* define lagged dimension */
    handle_error(nc_def_dim(ncid, "lag", nlags, lag_dim));

    /* Add lag coordinate variable */
    bs_dimids[0] = (*lag_dim);
    handle_error(nc_def_var(ncid, "lag", NC_SHORT, 1, bs_dimids, lagid));
    add_attributes(ncid, *lagid, "", NULL, "number of lags", NULL);
    nct_append_attribute(ncid, *lagid, "comments", "These are the positive lags","");
  }
  
  /* define  variables */
  bs_dimids[0] = *time_dim;
  bs_dimids[1] = *range_dim;
  if (outputraw) {
    bs_dimids[2] = *lag_dim;
  }
}

void
doppler_lidar_write_height_attributes(int ncid, int *range_dim, int *heightid)
{

  handle_error(nc_def_var(ncid, "height", NC_FLOAT, 1, range_dim, heightid));
  add_attributes(ncid, *heightid, "m", NULL, "Height above ground", NULL);
  add_string_attribute(ncid, *heightid, "axis", "Z");
  nct_append_attribute(ncid, *heightid, "comments", "This variable is range from lidar * sin(elevation)","");

}

void
doppler_lidar_write_az_el_attributes(int ncid, int bs_dimids[3], int azelid[4], unsigned short int scantypereq, unsigned short int scantype, char is_pitch_and_roll)
{
  int numdim;

  /* If vertical-only data requested (for fixed Stare), azimuth and elevation only need to be scalar variables, */
  /* otherwise azimuth and elevation are vector variables. */
  /* Note that stare files are usually fixed at one az-el, but not always, therefore safer to write as vector. */
  /* Pitch and roll (if present) can vary every ray */
  /* azelid: [0] elevation, [1] azimuth, [2], pitch, [3] roll */


  if (scantype < 3 && scantypereq == 1)
    numdim = 0;
  else
    numdim = 1;
  //  else // This option shouldn't be possible
  //    numdim = 2;


  /* Unless only vertical stare requested, Stare files can also change in elevation (and azimuth) */
  handle_error(nc_def_var(ncid, "elevation", NC_FLOAT, numdim, bs_dimids, &azelid[0]));
  add_attributes(ncid, azelid[0], "degrees", NULL, "elevation from horizontal", NULL);

  /* Unless only vertical stare requested, include azimuth */
  if (scantypereq == 0) {
    handle_error(nc_def_var(ncid, "azimuth", NC_FLOAT, numdim, bs_dimids, &azelid[1]));
    add_attributes(ncid, azelid[1], "degrees", NULL, "azimuth from North", NULL);
  }

  /* Add Pitch and Roll, if they exist in the file */
 if (is_pitch_and_roll) {
    handle_error(nc_def_var(ncid, "pitch", NC_FLOAT, 1, bs_dimids, &azelid[2]));
    add_attributes(ncid, azelid[2], "degrees", NULL, "pitch angle", "platform_pitch_angle");

    handle_error(nc_def_var(ncid, "roll", NC_FLOAT, 1, bs_dimids, &azelid[3]));
    add_attributes(ncid, azelid[3], "degrees", NULL, "roll angle", "platform_roll_angle");
  }
}

void
doppler_lidar_write_standardvar_attributes(DopplerLidarHeader *header, DopplerLidarSpec *spec, int ncid, int bs_dimids[3], int *mnid, int *bsid, int *vid, int *rangeid, int *bsrawid, int *vrawid, int *sigid, float missing_value, DopplerLidarPlotRange *plot_range)
{
  /* Add mean_noise variable */ 
  handle_error(nc_def_var(ncid, "mean_noise", NC_FLOAT, 1, bs_dimids, mnid));
  add_attributes(ncid, *mnid,  "arbitrary noise", NULL,  "Mean noise level", NULL);
  nct_append_attribute(ncid, *mnid, "comments", "Mean of the noise in the last few gates of the profile",""); 
  add_missing_value(ncid, *mnid, missing_value);
  
  /* Add backscatter variable */ 
  handle_error(nc_def_var(ncid, "beta", NC_FLOAT, 2, bs_dimids, bsid));
  add_attributes(ncid, *bsid, "m-1 sr-1", "m<sup>-1</sup> sr<sup>-1</sup>", "Attenuated backscatter", NULL);
  handle_error(nc_put_att_float(ncid, *bsid, "plot_range", NC_FLOAT, 2, plot_range->beta)); 
  nct_append_attribute(ncid, *bsid, "plot_scale", "logarithmic",""); 
  nct_append_attribute(ncid, *bsid, "comments", "Signal normalised, background noise removed and correction factor applied.",""); 
  add_missing_value(ncid, *bsid, missing_value);

  /* If focus correction has been applied */
  if (spec->applyfocuscorrection) {
    nct_append_attribute(ncid, *bsid, "comments", "Focus correction has been applied.","\n"); 
    nct_append_attribute(ncid, NC_GLOBAL, "history", "Focus correction has been applied.",". ");
  }

  /* If calibration factor has been applied */
  if (spec->applycalfactor)
    {
      handle_error(nc_put_att_float(ncid, *bsid, "calibration_factor_applied", NC_FLOAT, 1, &(spec->calfactor)));
 
      nct_append_attribute(ncid, NC_GLOBAL, "comments", "If present, the calibration_factor_applied attribute of the beta variable attribute indicates the calibration factor that has been applied, which should be accurate to within 10% (O'Connor et al., 2004).","\n"); 
      nct_append_attribute(ncid, NC_GLOBAL, "history", "Calibration factor applied.",". "); 
    }
  
  /* Add doppler velocity variable */ 
  handle_error(nc_def_var(ncid, "v", NC_FLOAT, 2, bs_dimids, vid));
  add_attributes(ncid, *vid, "m s-1",  "m s<sup>-1</sup>", "Radial velocity (positive away from lidar)", NULL);
  add_string_attribute(ncid, *vid, "short_name", "Doppler velocity");
  handle_error(nc_put_att_float(ncid, *vid, "resolution", NC_FLOAT, 1, &(header->v_resolution)));	
  handle_error(nc_put_att_float(ncid, *vid, "plot_range", NC_FLOAT, 2, plot_range->velocity)); 
  nct_append_attribute(ncid, *vid, "plot_scale", "linear",""); 
  add_missing_value(ncid, *vid, missing_value);
  nct_append_attribute(ncid, *vid, "comments", "Background noise removed","");

  /* Add raw backscatter variable */ 
  handle_error(nc_def_var(ncid, "beta_raw", NC_FLOAT, 2, bs_dimids, bsrawid));
  add_attributes(ncid, *bsrawid, "m-1 sr-1", "m<sup>-1</sup> sr<sup>-1</sup>", "Attenuated backscatter (raw)", NULL);
  handle_error(nc_put_att_float(ncid, *bsrawid, "plot_range", NC_FLOAT, 2, plot_range->beta)); 
  nct_append_attribute(ncid, *bsrawid, "plot_scale", "logarithmic","");
  add_missing_value(ncid, *bsrawid, missing_value);
  nct_append_attribute(ncid, *bsrawid, "comments", "Signal normalised and correction factor applied. No noise removal.",""); 

  /* If focus correction has been applied */
  if (spec->applyfocuscorrection) {
    nct_append_attribute(ncid, *bsrawid, "comments", "Focus correction has been applied.","\n"); 
  }
  
  /* Add raw doppler velocity variable */ 
  handle_error(nc_def_var(ncid, "v_raw", NC_FLOAT, 2, bs_dimids, vrawid));
  add_attributes(ncid, *vrawid, "m s-1",  "m s<sup>-1</sup>", "Radial velocity (positive away from lidar)", NULL);
  add_string_attribute(ncid, *vrawid, "short_name", "Doppler velocity (raw)");
  handle_error(nc_put_att_float(ncid, *vrawid, "resolution", NC_FLOAT, 1, &(header->v_resolution)));
  handle_error(nc_put_att_float(ncid, *vrawid, "plot_range", NC_FLOAT, 2, plot_range->velocity)); 
  nct_append_attribute(ncid, *vrawid, "plot_scale", "linear",""); 
  add_missing_value(ncid, *vrawid, missing_value);
  nct_append_attribute(ncid, *vrawid, "comments", "Raw Doppler velocity, no noise removal","");
  
  /* Add raw signal variable */ 
  handle_error(nc_def_var(ncid, "signal", NC_FLOAT, 2, bs_dimids, sigid));
  add_attributes(ncid, *sigid, " ", NULL, "Signal (raw)", NULL);
  add_missing_value(ncid, *sigid, missing_value);
  nct_append_attribute(ncid, *sigid, "comments", "Raw signal in arbitrary units. No range correction or other corrections applied.","");
}


void
doppler_lidar_write_errorvar_attributes(DopplerLidarHeader *header, DopplerLidarSpec *spec, int ncid, int bs_dimids[3], int *bserrid, int *verrid, float missing_value, DopplerLidarPlotRange *plot_range)
{
  
  /* Add backscatter error variable */ 
  handle_error(nc_def_var(ncid, "beta_error", NC_FLOAT, 2, bs_dimids, bserrid));
  add_attributes(ncid, *bserrid, " ", NULL, "Fractional error in beta", NULL);
  handle_error(nc_put_att_float(ncid, *bserrid, "plot_range", NC_FLOAT, 2, plot_range->beta_error)); 
  nct_append_attribute(ncid, *bserrid, "plot_scale", "logarithmic",""); 
  nct_append_attribute(ncid, *bserrid, "comments", "Fractional standard deviation (error estimate) applies to both backscatter and signal.",""); 
  add_missing_value(ncid, *bserrid, missing_value);

  /* Add doppler velocity error variable */ 
  handle_error(nc_def_var(ncid, "v_error", NC_FLOAT, 2, bs_dimids, verrid));
  add_attributes(ncid, *verrid, "m s-1",  "m s<sup>-1</sup>", "Uncertainty in velocity estimate", NULL);
  handle_error(nc_put_att_float(ncid, *verrid, "plot_range", NC_FLOAT, 2, plot_range->velocity_error)); 
  nct_append_attribute(ncid, *verrid, "plot_scale", "logarithmic",""); 
  add_missing_value(ncid, *verrid, missing_value);
}

void
doppler_lidar_write_rawvar_attributes(DopplerLidarHeader *header, DopplerLidarSpec *spec, int ncid, int bs_dimids[3], int *arid, int *aiid, int *siid, int *vid, float missing_value, char outputraw, DopplerLidarPlotRange *plot_range)
{

  if (outputraw) { 
    /* Add real ACF variable */
    handle_error(nc_def_var(ncid, "acf_real", NC_FLOAT, 3, bs_dimids, arid));
    add_attributes(ncid, *arid,  "", NULL,  "Auto-correlation function, real", NULL);
    add_missing_value(ncid, *arid, missing_value);

    /* Add imaginary ACF variable */
    handle_error(nc_def_var(ncid, "acf_imag", NC_FLOAT, 3, bs_dimids, aiid));
    add_attributes(ncid, *aiid,  "", NULL,  "Auto-correlation function, imaginary", NULL);
    add_missing_value(ncid, *aiid, missing_value);
  }
  
  if (spec->cleanrawrayflag) {
    /* Add signal variable */
    handle_error(nc_def_var(ncid, "signal", NC_FLOAT, 2, bs_dimids, siid));
    add_attributes(ncid, *siid,  "", NULL,  "Signal from real part of the auto-correlation function at zeroth lag", NULL);
//    handle_error(nc_put_att_float(ncid, *siid, "plot_range", NC_FLOAT, 2, plot_range->signal));
    add_missing_value(ncid, *siid, missing_value);
  }

  /* Add v_pp variable */
  handle_error(nc_def_var(ncid, "v_pp", NC_FLOAT, 2, bs_dimids, vid));
  add_attributes(ncid, *vid,  "m s-1", "m s<sup>-1</sup>",  "Velocity from pulse-pair estimator", NULL);
  handle_error(nc_put_att_float(ncid, *vid, "folding_velocity", NC_FLOAT, 1, &(spec->nyquist)));
  handle_error(nc_put_att_float(ncid, *vid, "plot_range", NC_FLOAT, 2, plot_range->velocity));
  add_missing_value(ncid, *vid, missing_value);

}

void
doppler_lidar_write_averagedrawvar_attributes(int ncid, int bs_dimids[3], int *svid, int *nrid, float missing_value, DopplerLidarPlotRange *plot_range)
{
  /* Add sigma_v variable */
  handle_error(nc_def_var(ncid, "sigma_v_pp", NC_FLOAT, 2, bs_dimids, svid));
  add_attributes(ncid, *svid,  "m s-1", "m s<sup>-1</sup>",  "Std. of velocity from pulse-pair estimator", NULL);
  handle_error(nc_put_att_float(ncid, *svid, "plot_range", NC_FLOAT, 2, plot_range->sigma_v));
  add_string_attribute(ncid, *svid, "plot_scale", "logarithmic");
  add_missing_value(ncid, *svid, missing_value);

  /* Add numrays variable */
  handle_error(nc_def_var(ncid, "numrays", NC_SHORT, 2, bs_dimids, nrid));
  add_attributes(ncid, *nrid,  " ", NULL,  "Number of rays averaged", NULL);
  handle_error(nc_put_att_float(ncid, *nrid, "plot_range", NC_FLOAT, 2, plot_range->num));
  add_string_attribute(ncid, *nrid, "plot_scale", "linear");
  add_missing_value_short(ncid, *nrid, 0);
}


void
doppler_lidar_write_averagedvar_attributes(int ncid, int bs_dimids[3], int *mbetaid, int *mvid, int *svid, int *nrid, float missing_value, DopplerLidarPlotRange *plot_range)
{
  /* Add mean_beta variable */ 
  handle_error(nc_def_var(ncid, "mean_beta", NC_FLOAT, 2, bs_dimids, mbetaid));
  add_attributes(ncid, *mbetaid, "m-1 sr-1", "m<sup>-1</sup> sr<sup>-1</sup>", "Mean attenuated backscatter", NULL);
  handle_error(nc_put_att_float(ncid, *mbetaid, "plot_range", NC_FLOAT, 2, plot_range->beta));
  add_string_attribute(ncid, *mbetaid, "plot_scale", "logarithmic");
  add_missing_value(ncid, *mbetaid, missing_value);
  
  /* Add mean_v variable */ 
  handle_error(nc_def_var(ncid, "mean_v", NC_FLOAT, 2, bs_dimids, mvid));
  add_attributes(ncid, *mvid,  "m s-1", "m s<sup>-1</sup>",  "Mean velocity", NULL);
  handle_error(nc_put_att_float(ncid, *mvid, "plot_range", NC_FLOAT, 2, plot_range->velocity));
  add_string_attribute(ncid, *mvid, "plot_scale", "linear");
  add_missing_value(ncid, *mvid, missing_value);
  
  /* Add sigma_v variable */ 
  handle_error(nc_def_var(ncid, "sigma_v", NC_FLOAT, 2, bs_dimids, svid));
  add_attributes(ncid, *svid,  "m s-1", "m s<sup>-1</sup>",  "Std. of velocity", NULL);
  handle_error(nc_put_att_float(ncid, *svid, "plot_range", NC_FLOAT, 2, plot_range->sigma_v));
  add_string_attribute(ncid, *svid, "plot_scale", "logarithmic");
  add_missing_value(ncid, *svid, missing_value);
  
  /* Add numrays variable */ 
  handle_error(nc_def_var(ncid, "numrays", NC_SHORT, 2, bs_dimids, nrid));
  add_attributes(ncid, *nrid,  " ", NULL,  "Number of rays averaged", NULL);
  handle_error(nc_put_att_float(ncid, *nrid, "plot_range", NC_FLOAT, 2, plot_range->num));
  add_string_attribute(ncid, *nrid, "plot_scale", "linear");
  add_missing_value_short(ncid, *nrid, 0);
}

void
doppler_lidar_write_highermomentvar_attributes(int ncid, int bs_dimids[3], int *skvid, int *kvid, float missing_value, DopplerLidarPlotRange *plot_range, char ispulsepair)
{
  /* Add skew_v variable */ 
  if (ispulsepair) {
    handle_error(nc_def_var(ncid, "skew_v_pp", NC_FLOAT, 2, bs_dimids, skvid));
    add_attributes(ncid, *skvid,  "m s-1", "m s<sup>-1</sup>",  "Skewness of velocity from pulse-pair estimator", NULL);
  }
  else {
    handle_error(nc_def_var(ncid, "skew_v_pp", NC_FLOAT, 2, bs_dimids, skvid));
    add_attributes(ncid, *skvid,  "m s-1", "m s<sup>-1</sup>",  "Skewness of velocity", NULL);
  }
  handle_error(nc_put_att_float(ncid, *skvid, "plot_range", NC_FLOAT, 2, plot_range->skew_v));
  add_string_attribute(ncid, *skvid, "plot_scale", "linear");
  add_missing_value(ncid, *skvid, missing_value);
  
  /* Add kurt_v variable */ 
  if (ispulsepair) {
    handle_error(nc_def_var(ncid, "kurt_v_pp", NC_FLOAT, 2, bs_dimids, kvid));
    add_attributes(ncid, *kvid,  "m s-1", "m s<sup>-1</sup>",  "Kurtosis of velocity from pulse-pair estimator", NULL);
  }
  else {
    handle_error(nc_def_var(ncid, "kurt_v", NC_FLOAT, 2, bs_dimids, kvid));
    add_attributes(ncid, *kvid,  "m s-1", "m s<sup>-1</sup>",  "Kurtosis of velocity", NULL);
  }
  handle_error(nc_put_att_float(ncid, *kvid, "plot_range", NC_FLOAT, 2, plot_range->kurt_v));
  add_string_attribute(ncid, *kvid, "plot_scale", "linear");
  add_missing_value(ncid, *kvid, missing_value);
}


void
doppler_lidar_write_windvar_attributes(int ncid, int bs_dimids[3], int *uwindid, int *vwindid, int *wwindid, int *speedid, int *dirid, float convention, float azimuthalcorrection, float missing_value, DopplerLidarPlotRange *plot_range)
{
  /* Add uwind variable */ 
  handle_error(nc_def_var(ncid, "uwind", NC_FLOAT, 2, bs_dimids, uwindid));
  add_attributes(ncid, *uwindid, "m s-1", "m s<sup>-1</sup>", "Zonal wind", "eastward_wind");
  handle_error(nc_put_att_float(ncid, *uwindid, "plot_range", NC_FLOAT, 2, plot_range->uwind));
  add_string_attribute(ncid, *uwindid, "plot_scale", "linear");
  add_missing_value(ncid, *uwindid, missing_value);
  
  /* Add vwind variable */ 
  handle_error(nc_def_var(ncid, "vwind", NC_FLOAT, 2, bs_dimids, vwindid));
  add_attributes(ncid, *vwindid, "m s-1", "m s<sup>-1</sup>", "Meridional wind", "northward_wind");
  handle_error(nc_put_att_float(ncid, *vwindid, "plot_range", NC_FLOAT, 2, plot_range->vwind));
  add_string_attribute(ncid, *vwindid, "plot_scale", "linear");
  add_missing_value(ncid, *vwindid, missing_value);

  /* Add uwind variable */ 
  handle_error(nc_def_var(ncid, "wwind", NC_FLOAT, 2, bs_dimids, wwindid));
  add_attributes(ncid, *wwindid, "m s-1", "m s<sup>-1</sup>", "Vertical wind", "upward_wind");
  handle_error(nc_put_att_float(ncid, *wwindid, "plot_range", NC_FLOAT, 2, plot_range->wwind));
  add_string_attribute(ncid, *wwindid, "plot_scale", "linear");
  add_missing_value(ncid, *wwindid, missing_value);

  /* Add wind speed variable */ 
  handle_error(nc_def_var(ncid, "wind_speed", NC_FLOAT, 2, bs_dimids, speedid));
  add_attributes(ncid, *speedid, "m s-1", "m s<sup>-1</sup>", "Wind Speed", NULL);
  handle_error(nc_put_att_float(ncid, *speedid, "plot_range", NC_FLOAT, 2, plot_range->speed));
  add_string_attribute(ncid, *speedid, "plot_scale", "linear");
  add_missing_value(ncid, *speedid, missing_value);

  /* Add wind direction variable */ 
  handle_error(nc_def_var(ncid, "wind_direction", NC_FLOAT, 2, bs_dimids, dirid));
  add_attributes(ncid, *dirid, "degrees", "degrees", "Wind Direction", NULL);
  handle_error(nc_put_att_float(ncid, *dirid, "plot_range", NC_FLOAT, 2, plot_range->dir));
  add_string_attribute(ncid, *dirid, "plot_scale", "linear");
  add_missing_value(ncid, *dirid, missing_value);
  if (convention == 180)
    nct_append_attribute(ncid, *dirid, "comments", "Meteorological convention, direction wind is blowing from.","");
  else
    nct_append_attribute(ncid, *dirid, "comments", "Wind vector azimuth convention, direction wind is blowing to.","");


  /* If azimuthal correction has been applied */
  if (azimuthalcorrection > 0.1)
    {
      handle_error(nc_put_att_float(ncid, *dirid, "azimuthal_correction_applied", NC_FLOAT, 1, &(azimuthalcorrection)));

      nct_append_attribute(ncid, NC_GLOBAL, "comments", "An azimuthal correction factor has been applied, given in the azimuthal_correction_applied attribute of the wind_direction variable. This correction has also been applied to the variables uwind and vwind.","\n");
      nct_append_attribute(ncid, NC_GLOBAL, "history", "Azimuthal correction applied.",". ");
    }

}

void
doppler_lidar_write_winderrorvar_attributes(int ncid, int bs_dimids[3], int *uerrid, int *verrid, int *werrid, int *speederrid, int *direrrid, float missing_value, DopplerLidarPlotRange *plot_range)
{
  /* Add uwind variable */
  handle_error(nc_def_var(ncid, "uwind_error", NC_FLOAT, 2, bs_dimids, uerrid));
  add_attributes(ncid, *uerrid, "m s-1", "m s<sup>-1</sup>", "Uncertainty in zonal wind", NULL);
  handle_error(nc_put_att_float(ncid, *uerrid, "plot_range", NC_FLOAT, 2, plot_range->uwind_error));
  add_string_attribute(ncid, *uerrid, "plot_scale", "linear");
  add_missing_value(ncid, *uerrid, missing_value);

  /* Add vwind variable */
  handle_error(nc_def_var(ncid, "vwind_error", NC_FLOAT, 2, bs_dimids, verrid));
  add_attributes(ncid, *verrid, "m s-1", "m s<sup>-1</sup>", "Uncertainty in meridional wind", NULL);
  handle_error(nc_put_att_float(ncid, *verrid, "plot_range", NC_FLOAT, 2, plot_range->vwind_error));
  add_string_attribute(ncid, *verrid, "plot_scale", "linear");
  add_missing_value(ncid, *verrid, missing_value);

  /* Add uwind variable */
  handle_error(nc_def_var(ncid, "wwind_error", NC_FLOAT, 2, bs_dimids, werrid));
  add_attributes(ncid, *werrid, "m s-1", "m s<sup>-1</sup>", "Uncertainty in vertical wind", NULL);
  handle_error(nc_put_att_float(ncid, *werrid, "plot_range", NC_FLOAT, 2, plot_range->wwind_error));
  add_string_attribute(ncid, *werrid, "plot_scale", "linear");
  add_missing_value(ncid, *werrid, missing_value);

  /* Add wind speed variable */
  handle_error(nc_def_var(ncid, "wind_speed_error", NC_FLOAT, 2, bs_dimids, speederrid));
  add_attributes(ncid, *speederrid, "m s-1", "m s<sup>-1</sup>", "Uncertainty in wind speed", NULL);
  handle_error(nc_put_att_float(ncid, *speederrid, "plot_range", NC_FLOAT, 2, plot_range->speed_error));
  add_string_attribute(ncid, *speederrid, "plot_scale", "linear");
  add_missing_value(ncid, *speederrid, missing_value);


  /* Add wind direction variable */
  handle_error(nc_def_var(ncid, "wind_direction_error", NC_FLOAT, 2, bs_dimids, direrrid));
  add_attributes(ncid, *direrrid, "degrees", "degrees", "Uncertainty in wind direction", NULL);
  handle_error(nc_put_att_float(ncid, *direrrid, "plot_range", NC_FLOAT, 2, plot_range->dir_error));
  add_string_attribute(ncid, *direrrid, "plot_scale", "linear");
  add_missing_value(ncid, *direrrid, missing_value);
}

void
doppler_lidar_write_snr_attributes(int ncid, int bs_dimids[3], int *meansnrid, float missing_value)
{

  handle_error(nc_def_var(ncid, "snr_mean", NC_FLOAT, 2, bs_dimids, meansnrid));
  add_attributes(ncid, *meansnrid, "dB", "dB", "Mean SNR", NULL);
  nct_append_attribute(ncid, *meansnrid, "comments", "This variable is the profile of the scan-averaged mean of SNR","");
  add_missing_value(ncid, *meansnrid, missing_value);

}


void
doppler_lidar_write_depolvar_attributes(int ncid, int bs_dimids[3], int paramid[6], float missing_value, DopplerLidarPlotRange *plot_range)
{

  /* Add attenuated backscatter variable */
  handle_error(nc_def_var(ncid, "beta", NC_FLOAT, 2, bs_dimids, &paramid[0]));
  add_attributes(ncid, paramid[0], "m-1 sr-1", "m<sup>-1</sup> sr<sup>-1</sup>", "Attenuated backscatter", NULL);
  handle_error(nc_put_att_float(ncid, paramid[0], "plot_range", NC_FLOAT, 2, plot_range->beta));
  nct_append_attribute(ncid, paramid[0], "plot_scale", "logarithmic","");
  add_missing_value(ncid, paramid[0], missing_value);
  nct_append_attribute(ncid, paramid[0], "comments", "This parameter is the co-polar attenuated backscatter","");

  /* Add Doppler velocity variable */ 
  handle_error(nc_def_var(ncid, "v", NC_FLOAT, 2, bs_dimids, &paramid[1]));
  add_attributes(ncid, paramid[1], "m s-1",  "m s<sup>-1</sup>", "Doppler velocity", NULL);
  handle_error(nc_put_att_float(ncid, paramid[1], "plot_range", NC_FLOAT, 2, plot_range->velocity));
  nct_append_attribute(ncid, paramid[1], "plot_scale", "linear","");
  add_missing_value(ncid, paramid[1], missing_value);
  nct_append_attribute(ncid, paramid[1], "comments", "This parameter is the co-polar radial velocity (positive away from lidar)","");

  /* Add depolarization variable */
  handle_error(nc_def_var(ncid, "depolarization", NC_FLOAT, 2, bs_dimids, &paramid[2]));
  add_attributes(ncid, paramid[2], " ", NULL, "Depolarization ratio", NULL);
  handle_error(nc_put_att_float(ncid, paramid[2], "plot_range", NC_FLOAT, 2, plot_range->depolarization));
  nct_append_attribute(ncid, paramid[2], "plot_scale", "linear","");
  add_missing_value(ncid, paramid[2], missing_value);

  /* Add depolarization error variable */
  handle_error(nc_def_var(ncid, "depolarization_error", NC_FLOAT, 2, bs_dimids, &paramid[3]));
  add_attributes(ncid, paramid[3], " ", NULL, "Fractional uncertainty in depolarisation estimate", NULL);
  handle_error(nc_put_att_float(ncid, paramid[3], "plot_range", NC_FLOAT, 2, plot_range->depolarization_error));
  nct_append_attribute(ncid, paramid[3], "plot_scale", "logarithmic","");
  add_missing_value(ncid, paramid[3], missing_value);

  /* Add differential Doppler velocity variable */
  handle_error(nc_def_var(ncid, "differential_v", NC_FLOAT, 2, bs_dimids, &paramid[4]));
  add_attributes(ncid, paramid[4], "m s-1",  "m s<sup>-1</sup>", "Differential Doppler velocity", NULL);
  handle_error(nc_put_att_float(ncid, paramid[4], "plot_range", NC_FLOAT, 2, plot_range->differential_v));
  nct_append_attribute(ncid, paramid[4], "plot_scale", "linear","");
  add_missing_value(ncid, paramid[4], missing_value);
  nct_append_attribute(ncid, paramid[4], "comments", "This parameter is calculated as cross-polar - co-polar radial velocities (positive away from lidar)","");

  /* Add differential Doppler velocity error variable */
  handle_error(nc_def_var(ncid, "differential_v_error", NC_FLOAT, 2, bs_dimids, &paramid[5]));
  add_attributes(ncid, paramid[5], "m s-1",  "m s<sup>-1</sup>", "Uncertainty in differential Doppler velocity estimate", NULL);
  handle_error(nc_put_att_float(ncid, paramid[5], "plot_range", NC_FLOAT, 2, plot_range->differential_v_error));
  nct_append_attribute(ncid, paramid[5], "plot_scale", "logarithmic","");
  add_missing_value(ncid, paramid[5], missing_value);
}


